(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Proofs_DeBruijnSIndices.v *)
(* Description: Properties of De Bruijn's indices (Appendix 6) *)

(** * Proofs about De Bruijn's Indices *)

Require Import Calculus.
Require Import Arith.
Require Import MatchCompNat.
Require Import Util.
Require Import List.
Require Import ListSet.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_Progress.v *)
Require Proofs_Progress.
Module My_Proofs_Progress := Proofs_Progress.SetProgram(My_Program).

Import My_Proofs_Progress.
Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

(** Lifting a closed term or type **)
Lemma lemmaA30_0:
  forall (p: Path) (n: nat),
    (path_is_var_closed p) ->
    (lift_in_path p n) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
rename v into i.
rename H into H'.
assert (path_is_var_closed' (Path_var i)).
apply path_is_var_closed_eq_r. trivial.
unfold path_is_var_closed' in H.
unfold var_in_path in H.
induction (H i). trivial.
simpl. trivial.
Qed.

(** Lifting a closed term or type **)
Lemma lemmaA30_2:
  forall (T: type) (n: nat),
    (type_is_var_closed T) ->
    (lift_in_type T n) = T
.

Proof.
intros.
induction T.
simpl.
unfold type_is_var_closed in H.
simpl in H.
fold (lift_in_path p n).
rewrite lemmaA30_0. trivial. exact H.
simpl. trivial.
Qed.

(** Lifting a closed path **)
Lemma lemmaA30_0':
  forall (p: Path) (n k: nat),
    (fv_at_depth_in_path p k) = empty_varset ->
    (lift_at_depth_in_path p n k) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
rename v into i.
unfold lift_at_depth_in_path.
rewrite match_comp_nat2_lt. trivial.
apply fv_pathvar_empty. trivial.
simpl. trivial.
Qed.

(** Lifting a closed type **)
Lemma lemmaA30_2':
  forall (T: type) (n k: nat),
    (fv_at_depth_in_type T k) = empty_varset ->
    (lift_at_depth_in_type T n k) = T
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
simpl in H.
apply lemmaA30_0'. trivial. trivial.
simpl. auto.
Qed.

Definition lemmaA30_1_prop1 (t: Term): Prop :=
  forall (n k: nat),
    (fv_at_depth_in_term t k) = empty_varset ->
    (lift_at_depth_in_term t n k) = t
.

Definition lemmaA30_1_prop2 (fts: FieldTermList): Prop :=
  forall (n k: nat),
    (fv_at_depth_in_fieldTermList fts k) = empty_varset ->
    (lift_at_depth_in_fieldTermList fts n k) = fts
.

Definition lemmaA30_1_prop3 (xts: ParamTermList): Prop :=
  forall (n k: nat),
    (fv_at_depth_in_paramTermList xts k) = empty_varset ->
    (lift_at_depth_in_paramTermList xts n k) = xts
.

(** Lifting a closed term **)
Lemma lemmaA30_1':
  forall (t: Term) (n k: nat),
    (fv_at_depth_in_term t k) = empty_varset ->
    (lift_at_depth_in_term t n k) = t
.

Proof.
(* Prelude *)
cut (forall (t: Term), (lemmaA30_1_prop1 t)).
intros. apply H. trivial.
apply (Term_prop
  lemmaA30_1_prop1
  lemmaA30_1_prop2
  lemmaA30_1_prop3
).

(* This *)
unfold lemmaA30_1_prop1.
simpl. trivial.
(* Param *)
unfold lemmaA30_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA30_1_prop1.
intros.
unfold lift_at_depth_in_term.
rewrite match_comp_nat2_lt. trivial.
apply fv_termvar_empty. trivial.
(* Select *)
unfold lemmaA30_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA30_0'.
unfold fv_at_depth_in_term in H. trivial. trivial.
(* New *)
unfold lemmaA30_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA30_1_prop2 in H. apply H. trivial.
(* Call *)
unfold lemmaA30_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA30_0'.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H0.
trivial.
apply H.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H0.
(* Let *)
unfold lemmaA30_1_prop1.
intros.
simpl in H1.
simpl.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite lemmaA30_2'. rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H2.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H2.
(* Nil (fields) *)
unfold lemmaA30_1_prop2.
simpl. trivial.
(* Cons (fields) *)
unfold lemmaA30_1_prop2.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
(* Nil (params) *)
unfold lemmaA30_1_prop3.
simpl. trivial.
(* Cons (params) *)
unfold lemmaA30_1_prop3.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
Qed.

(** Lifting a closed term **)
Lemma lemmaA30_1:
  forall (t: Term) (n: nat),
    (term_is_var_closed t) ->
    (lift_in_term t n) = t
.

Proof.
intros.
unfold term_is_var_closed in H.
unfold lift_in_term.
apply lemmaA30_1'. apply H.
Qed.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

(** Dropping a closed term or type **)
Lemma lemmaA31_0:
  forall (p: Path),
    (path_is_var_closed p) ->
    (drop_in_path p) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
rename v into i.
rename H into H'.
assert (path_is_var_closed' (Path_var i)).
apply path_is_var_closed_eq_r. trivial.
unfold path_is_var_closed' in H.
unfold var_in_path in H.
induction (H i). trivial.
simpl. trivial.
Qed.

(** Dropping a closed term or type **)
Lemma lemmaA31_2:
  forall (T: type),
    (type_is_var_closed T) ->
    (drop_in_type T) = T
.

Proof.
intros.
induction T.
simpl.
unfold type_is_var_closed in H.
simpl in H.
fold (drop_in_path p).
rewrite lemmaA31_0. trivial. exact H.
simpl. trivial.
Qed.

(** Dropping a closed path **)
Lemma lemmaA31_0':
  forall (p: Path) (k: nat),
    (fv_at_depth_in_path p k) = empty_varset ->
    (drop_at_depth_in_path p k) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
rename v into i.
unfold drop_at_depth_in_path.
rewrite match_comp_nat2_lt. trivial.
apply fv_pathvar_empty. trivial.
simpl. trivial.
Qed.

(** Dropping a closed type **)
Lemma lemmaA31_2':
  forall (T: type) (k: nat),
    (fv_at_depth_in_type T k) = empty_varset ->
    (drop_at_depth_in_type T k) = T
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
simpl in H.
apply lemmaA31_0'. trivial. trivial.
simpl. auto.
Qed.

Definition lemmaA31_1_prop1 (t: Term): Prop :=
  forall (k: nat),
    (fv_at_depth_in_term t k) = empty_varset ->
    (drop_at_depth_in_term t k) = t
.

Definition lemmaA31_1_prop2 (fts: FieldTermList): Prop :=
  forall (k: nat),
    (fv_at_depth_in_fieldTermList fts k) = empty_varset ->
    (drop_at_depth_in_fieldTermList fts k) = fts
.

Definition lemmaA31_1_prop3 (xts: ParamTermList): Prop :=
  forall (k: nat),
    (fv_at_depth_in_paramTermList xts k) = empty_varset ->
    (drop_at_depth_in_paramTermList xts k) = xts
.

(** Dropping a closed term **)
Lemma lemmaA31_1':
  forall (t: Term) (k: nat),
    (fv_at_depth_in_term t k) = empty_varset ->
    (drop_at_depth_in_term t k) = t
.

Proof.
(* Prelude *)
cut (forall (t: Term), (lemmaA31_1_prop1 t)).
intros. apply H. trivial.
apply (Term_prop
  lemmaA31_1_prop1
  lemmaA31_1_prop2
  lemmaA31_1_prop3
).

(* This *)
unfold lemmaA31_1_prop1.
simpl. trivial.
(* Param *)
unfold lemmaA31_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA31_1_prop1.
intros.
unfold drop_at_depth_in_term.
rewrite match_comp_nat2_lt. trivial.
apply fv_termvar_empty. trivial.
(* Select *)
unfold lemmaA31_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA31_0'.
unfold fv_at_depth_in_term in H. trivial. trivial.
(* New *)
unfold lemmaA31_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA31_1_prop2 in H. apply H. trivial.
(* Call *)
unfold lemmaA31_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA31_0'.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H0.
trivial.
apply H.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H0.
(* Let *)
unfold lemmaA31_1_prop1.
intros.
simpl in H1.
simpl.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite lemmaA31_2'. rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H2.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H2.
(* Nil (fields) *)
unfold lemmaA31_1_prop2.
simpl. trivial.
(* Cons (fields) *)
unfold lemmaA31_1_prop2.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
(* Nil (params) *)
unfold lemmaA31_1_prop3.
simpl. trivial.
(* Cons (params) *)
unfold lemmaA31_1_prop3.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
Qed.

(** Dropping a closed term **)
Lemma lemmaA31_1:
  forall (t: Term),
    (term_is_var_closed t) ->
    (drop_in_term t) = t
.

Proof.
intros.
unfold term_is_var_closed in H.
unfold drop_in_term.
apply lemmaA31_1'. apply H.
Qed.

(******************************************************************)
(******************************************************************)
(* Values are invariant by lifting *)
(******************************************************************)
(******************************************************************)

Definition lemmaA32_prop1 (v: Value): Prop :=
  forall (n k: nat),
    (lift_at_depth_in_term (value_to_term v) n k) = (value_to_term v)
.

Definition lemmaA32_prop2 (fvs: FieldValueList): Prop :=
  forall (n k: nat),
    (lift_at_depth_in_fieldTermList (fieldValueList_to_fieldTermList fvs) n k) =
    (fieldValueList_to_fieldTermList fvs)
.

(** Values are invariant by lifting **)
Lemma lemmaA32:
  forall (v: Value) (n k: nat),
    (lift_at_depth_in_term (value_to_term v) n k) = (value_to_term v)
.

Proof.
cut (forall (v: Value), (lemmaA32_prop1 v)).
intros. apply H.
apply (Value_prop
  lemmaA32_prop1
  lemmaA32_prop2
).

unfold lemmaA32_prop1.
unfold lemmaA32_prop2.
intros.
simpl. 
apply f_equal2 with (f := Term_new). trivial.
apply H.
unfold lemmaA32_prop2.
simpl. trivial.
intros.
unfold lemmaA32_prop1.
unfold lemmaA32_prop2.
intros.
simpl.
apply f_equal3 with (f := FieldTerm_cons). trivial.
apply H. apply H0.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting lifting and self substitution **)
(******************************************************************)
(******************************************************************)

(** Permuting lifting and self substitution **)
Lemma lemmaA33_0:
  forall (p: Path) (n: VarSym) (v: Value),
    (lift_in_path (thisSubst_in_path p (Path_value v)) n) =
    (thisSubst_in_path (lift_in_path p n) (Path_value v))
. 

Proof.
intros.
unfold lift_in_path.
generalize 0.
induction p.
intros. simpl. trivial.
intros. simpl. trivial.
intros. simpl. rename n0 into k. rename v0 into i.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting lifting and self substitution **)
Lemma lemmaA33_2:
  forall (T: type) (n: VarSym) (v: Value),
    (lift_in_type (thisSubst_in_type T (Path_value v)) n) =
    (thisSubst_in_type (lift_in_type T n) (Path_value v))
. 

Proof.
intros.
induction T.
simpl.
fold (lift_in_path (thisSubst_in_path p (Path_value v)) n).
fold (lift_in_path p n).
apply f_equal2 with (f := type_virtual).
apply lemmaA33_0. trivial.
simpl. trivial.
Qed.

(** Permuting lifting and self substitution **)
Lemma lemmaA33_0':
  forall (p: Path) (n k: nat) (v: Value),
    (lift_at_depth_in_path (thisSubst_in_path p (Path_value v)) n k) =
    (thisSubst_in_path (lift_at_depth_in_path p n k) (Path_value v))
. 

Proof.
intros.
generalize k. clear k.
induction p.
intros. simpl. trivial.
intros. simpl. trivial.
intros. simpl. rename v0 into i.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting lifting and self substitution **)
Lemma lemmaA33_2':
  forall (T: type) (n k: nat) (v: Value),
    (lift_at_depth_in_type (thisSubst_in_type T (Path_value v)) n k) =
    (thisSubst_in_type (lift_at_depth_in_type T n k) (Path_value v))
. 

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
apply lemmaA33_0'. trivial.
simpl. trivial.
Qed.

Definition lemmaA33_1_prop1 (t: Term): Prop :=
  forall (n k: nat) (v: Value),
    (lift_at_depth_in_term (thisSubst_in_term t (Path_value v)) n k) =
    (thisSubst_in_term (lift_at_depth_in_term t n k) (Path_value v))
.

Definition lemmaA33_1_prop2 (fts: FieldTermList): Prop :=
  forall (n k: nat) (v: Value),
    (lift_at_depth_in_fieldTermList (thisSubst_in_fieldTermList fts (Path_value v)) n k) =
    (thisSubst_in_fieldTermList (lift_at_depth_in_fieldTermList fts n k) (Path_value v))
.

Definition lemmaA33_1_prop3 (xts: ParamTermList): Prop :=
  forall (n k: nat) (v: Value),
    (lift_at_depth_in_paramTermList (thisSubst_in_paramTermList xts (Path_value v)) n k) =
    (thisSubst_in_paramTermList (lift_at_depth_in_paramTermList xts n k) (Path_value v))
.

Lemma lemmaA33_1':
  forall (t: Term) (n k: nat) (v: Value),
    (lift_at_depth_in_term (thisSubst_in_term t (Path_value v)) n k) =
    (thisSubst_in_term (lift_at_depth_in_term t n k) (Path_value v))
.

Proof.
(* Prelude *)
cut (forall (t: Term), (lemmaA33_1_prop1 t)).
intros. apply H.
apply (Term_prop
  lemmaA33_1_prop1
  lemmaA33_1_prop2
  lemmaA33_1_prop3
).

(* This *)
unfold lemmaA33_1_prop1.
intros.
simpl.
rewrite lemmaA32. trivial.
(* Param *)
unfold lemmaA33_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA33_1_prop1.
intros.
rename v into i.
simpl.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
(* Select *)
unfold lemmaA33_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA33_0'. trivial.
(* New *)
unfold lemmaA33_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA33_1_prop2 in H. apply H.
(* Call *)
unfold lemmaA33_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA33_0'. trivial.
apply H.
(* Let *)
intros.
unfold lemmaA33_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_loc).
apply lemmaA33_2'.
apply H. apply H0.
(* Nil (fields) *)
unfold lemmaA33_1_prop2.
simpl. trivial.
(* Cons (fields) *)
intros.
unfold lemmaA33_1_prop2.
intros. simpl.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H.
apply H0.
(* Nil (params) *)
unfold lemmaA33_1_prop3.
simpl. trivial.
(* Cons (params) *)
intros.
unfold lemmaA33_1_prop3.
intros. simpl.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H.
apply H0.
Qed.

(** Permuting lifting and self substitutions **)
Lemma lemmaA33_1:
  forall (t: Term) (n: nat) (v: Value),
    (lift_in_term (thisSubst_in_term t (Path_value v)) n) =
    (thisSubst_in_term (lift_in_term t n) (Path_value v))
.

Proof.
intros.
unfold lift_in_term.
apply lemmaA33_1'.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting lifting and variable substitution **)
(******************************************************************)
(******************************************************************)

Lemma lemmaA35_0':
  forall (p: Path) (n k k': VarSym) (v: Value),
    (k' <= k) ->
    (lift_at_depth_in_path (varSubst_in_path p k v) n k') =
    (varSubst_in_path (lift_at_depth_in_path p n k') (k + n) v)
.

Proof.
intros.
generalize k k' H. clear k k' H.
induction p.

(* Case This *)
intros. simpl. trivial.

(* Case Param *)
intros. simpl. trivial.

(* Case Var *)
rename v0 into i.
intros.
induction (lt_eq_lt_dec i k).
induction a.

(* Case i < k *)
simpl.
rewrite match_comp_nat3_lt.
induction (lt_eq_lt_dec i k').
induction a0.

(* i < k' *)
rewrite match_comp_nat2_lt. simpl.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
trivial. trivial.

(* i = k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_path.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
omega.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_path.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
omega. trivial.

(* Case i = k *)
simpl.
rewrite match_comp_nat3_eq.

induction (lt_eq_lt_dec i k').
induction a.

(* i < k' *)
absurd (k' <= k).
omega. trivial.

(* i = k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_path.
rewrite match_comp_nat3_eq. trivial.
auto with arith.
omega.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_path.
rewrite match_comp_nat3_eq. trivial.
auto with arith.
omega. trivial.

(* Case i > k *)
simpl.
rewrite match_comp_nat3_gt.
induction (lt_eq_lt_dec i k').
induction a.

(* i < k' *)
absurd (k' <= k).
omega. trivial.

(* i = k' *)
absurd (k' <= k).
omega. trivial.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
unfold varSubst_in_path.
rewrite match_comp_nat3_gt.
assert (
  forall (n m: nat),
    (n > 0) ->
    (pred n) + m = pred (n + m)
).
intros.
induction n0.
omega.
simpl. trivial.
rewrite H0. trivial.
omega. omega. omega. omega. omega.

(* Case Value *)
intros.
simpl. trivial.
Qed.

(** Permuting lifting and variable substitution **)
Lemma lemmaA35_0:
  forall (p: Path) (n k: VarSym) (v: Value),
    (lift_in_path (varSubst_in_path p k v) n) =
    (varSubst_in_path (lift_in_path p n) (k + n) v)
.

Proof.
intros.
unfold lift_in_path.
apply lemmaA35_0'. auto with arith.
Qed.

Lemma lemmaA35_2':
  forall (T: type) (n k k': nat) (v: Value),
    (k' <= k) ->
    (lift_at_depth_in_type (varSubst_in_type T k v) n k') =
    (varSubst_in_type (lift_at_depth_in_type T n k') (k + n) v)
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
apply lemmaA35_0'. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting lifting and variable substitution **)
Lemma lemmaA35_2:
  forall (T: type) (n k: VarSym) (v: Value),
    (lift_in_type (varSubst_in_type T k v) n) =
    (varSubst_in_type (lift_in_type T n) (k + n) v)
.

Proof.
intros.
unfold lift_in_type.
apply lemmaA35_2'.
auto with arith.
Qed.

Definition lemmaA35_1_prop1 (n: nat) (v: Value) (t: Term): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    (lift_at_depth_in_term (varSubst_in_term t k v) n k') =
    (varSubst_in_term (lift_at_depth_in_term t n k') (k + n) v)
.

Definition lemmaA35_1_prop2 (n: nat) (v: Value) (fts: FieldTermList): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    (lift_at_depth_in_fieldTermList (varSubst_in_fieldTermList fts k v) n k') =
    (varSubst_in_fieldTermList (lift_at_depth_in_fieldTermList fts n k') (k + n) v)
.

Definition lemmaA35_1_prop3 (n: nat) (v: Value) (xts: ParamTermList): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    (lift_at_depth_in_paramTermList (varSubst_in_paramTermList xts k v) n k') =
    (varSubst_in_paramTermList (lift_at_depth_in_paramTermList xts n k') (k + n) v)
.

Lemma lemmaA35_1':
  forall (t: Term) (n k k': nat) (v: Value),
    (k' <= k) ->
    (lift_at_depth_in_term (varSubst_in_term t k v) n k') =
    (varSubst_in_term (lift_at_depth_in_term t n k') (k + n) v)
.

Proof.
(* Prelude *)
cut (forall (n: nat) (v: Value) (t: Term), (lemmaA35_1_prop1 n v t)).
intros. apply H. trivial.
intros n v.
apply (Term_prop
  (lemmaA35_1_prop1 n v)
  (lemmaA35_1_prop2 n v)
  (lemmaA35_1_prop3 n v)
).

(* This *)
unfold lemmaA35_1_prop1.
intros.
simpl.
trivial.
(* Param *)
unfold lemmaA35_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA35_1_prop1.
intros.
rename v0 into i.
(*************** Start: copy-and-paste *******************)
induction (lt_eq_lt_dec i k).
induction a.

(* Case i < k *)
simpl.
rewrite match_comp_nat3_lt.
induction (lt_eq_lt_dec i k').
induction a0.

(* i < k' *)
rewrite match_comp_nat2_lt. simpl.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
trivial. trivial.

(* i = k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_term.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
omega.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_term.
rewrite match_comp_nat3_lt. trivial.
auto with arith.
omega. trivial.

(* Case i = k *)
simpl.
rewrite match_comp_nat3_eq.

induction (lt_eq_lt_dec i k').
induction a.

(* i < k' *)
absurd (k' <= k).
omega. trivial.

(* i = k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_term.
rewrite match_comp_nat3_eq.
rewrite lemmaA32.
trivial.
auto with arith.
omega.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
unfold varSubst_in_term.
rewrite match_comp_nat3_eq.
rewrite lemmaA32.
trivial.
auto with arith.
omega. trivial.

(* Case i > k *)
simpl.
rewrite match_comp_nat3_gt.
induction (lt_eq_lt_dec i k').
induction a.

(* i < k' *)
absurd (k' <= k).
omega. trivial.

(* i = k' *)
absurd (k' <= k).
omega. trivial.

(* i > k' *)
simpl.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge.
unfold varSubst_in_term.
rewrite match_comp_nat3_gt.
assert (
  forall (n m: nat),
    (n > 0) ->
    (pred n) + m = pred (n + m)
).
intros.
induction n0.
omega.
simpl. trivial.
rewrite H0. trivial.
omega. omega. omega. omega. omega.
(*************** End: copy-and-paste *******************)
(* Select *)
unfold lemmaA35_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA35_0'. trivial. trivial.
(* New *)
unfold lemmaA35_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA35_1_prop2 in H. apply H. trivial.
(* Call *)
unfold lemmaA35_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA35_0'. trivial. trivial.
apply H. trivial.
(* Let *)
intros.
unfold lemmaA35_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_loc).
apply lemmaA35_2'. trivial.
apply H. trivial.
unfold lemmaA35_1_prop1 in H0.
assert ((S k') <= (S k)). auto with arith.
assert (S (k + n) = ((S k) + n)). auto with arith.
rewrite H3.
apply H0. trivial.
(* Nil (fields) *)
unfold lemmaA35_1_prop2.
simpl. trivial.
(* Cons (fields) *)
intros.
unfold lemmaA35_1_prop2.
intros. simpl.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H. trivial.
apply H0. trivial.
(* Nil (params) *)
unfold lemmaA35_1_prop3.
simpl. trivial.
(* Cons (params) *)
intros.
unfold lemmaA35_1_prop3.
intros. simpl.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H. trivial.
apply H0. trivial.
Qed.

(** Permuting lifting and variable substitutions **)
Lemma lemmaA35_1:
  forall (t: Term) (n k: VarSym) (v: Value),
    (lift_in_term (varSubst_in_term t k v) n) =
    (varSubst_in_term (lift_in_term t n) (k + n) v)
.

Proof.
intros.
unfold lift_in_term.
apply lemmaA35_1'.
auto with arith.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting lifting and parameter substitution **)
(******************************************************************)
(******************************************************************)

(** Permuting lifting and parameter substitution **)
Lemma lemmaA34_0':
  forall (p: Path) (n k: VarSym) (xvs: list (ParamSym * Value)),
    (lift_at_depth_in_path (map_paramSubst_in_path p xvs) n k) =
    (map_paramSubst_in_path (lift_at_depth_in_path p n k) xvs)
.

Proof.
intros.
generalize k. clear k.
induction p.
(* Case This *)
simpl.
rewrite polySubst_in_pathThis. simpl. trivial.
(* Case Param *)
rename p into x.
destruct (polySubst_in_pathParam_corollary xvs x).
simpl.
rewrite H. simpl. trivial.
destruct H.
simpl. rewrite H. simpl. trivial.
(* Case Var *)
intro.
rename v into i.
rewrite polySubst_in_pathVar.
unfold lift_at_depth_in_path.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt.
rewrite polySubst_in_pathVar. trivial. trivial.
rewrite match_comp_nat2_ge.
rewrite polySubst_in_pathVar. trivial. trivial.
(* Case Value *)
simpl.
rewrite polySubst_in_pathValue. simpl. trivial.
Qed.

(** Permuting lifting and parameter substitution **)
Lemma lemmaA34_0:
  forall (p: Path) (n: VarSym) (xvs: list (ParamSym * Value)),
    (lift_in_path (map_paramSubst_in_path p xvs) n) =
    (map_paramSubst_in_path (lift_in_path p n) xvs)
.

Proof.
intros.
unfold lift_in_path.
apply lemmaA34_0'.
Qed.

(** Permuting lifting and parameter substitution **)
Lemma lemmaA34_2':
  forall (T: type) (n k: VarSym) (xvs: list (ParamSym * Value)),
    (lift_at_depth_in_type (map_paramSubst_in_type T xvs) n k) =
    (map_paramSubst_in_type (lift_at_depth_in_type T n k) xvs)
.

Proof.
intros.
induction T.
rewrite polySubst_in_typeVirtual.
simpl.
rewrite polySubst_in_typeVirtual.
apply f_equal2 with (f := type_virtual).
apply lemmaA34_0'. trivial.
simpl.
rewrite polySubst_in_typeClass.
simpl. trivial.
Qed.

(** Permuting lifting and parameter substitution **)
Lemma lemmaA34_2:
  forall (T: type) (n: VarSym) (xvs: list (ParamSym * Value)),
    (lift_in_type (map_paramSubst_in_type T xvs) n) =
    (map_paramSubst_in_type (lift_in_type T n) xvs)
.

Proof.
intros.
unfold lift_in_type.
apply lemmaA34_2'.
Qed.

Definition lemmaA34_1_prop1 (n: nat) (xvs: list (ParamSym * Value)) (t: Term): Prop :=
  forall (k: nat) ,
    (lift_at_depth_in_term (map_paramSubst_in_term t xvs) n k) =
    (map_paramSubst_in_term (lift_at_depth_in_term t n k) xvs)
.

Definition lemmaA34_1_prop2 (n: nat) (xvs: list (ParamSym * Value)) (fts: FieldTermList): Prop :=
  forall (k: nat) ,
    (lift_at_depth_in_fieldTermList (map_paramSubst_in_fieldTermList fts xvs) n k) =
    (map_paramSubst_in_fieldTermList (lift_at_depth_in_fieldTermList fts n k) xvs)
.

Definition lemmaA34_1_prop3 (n: nat) (xvs: list (ParamSym * Value)) (xts: ParamTermList): Prop :=
  forall (k: nat) ,
    (lift_at_depth_in_paramTermList (map_paramSubst_in_paramTermList xts xvs) n k) =
    (map_paramSubst_in_paramTermList (lift_at_depth_in_paramTermList xts n k) xvs)
.

Lemma lemmaA34_1':
  forall (t: Term) (n k: nat) (xvs: (list (ParamSym * Value))),
    (lift_at_depth_in_term (map_paramSubst_in_term t xvs) n k) =
    (map_paramSubst_in_term (lift_at_depth_in_term t n k) xvs)
.

Proof.
(* Prelude *)
intros.
cut (forall (t: Term), (lemmaA34_1_prop1 n xvs t)).
intros. apply H. clear t k.
apply (Term_prop
  (lemmaA34_1_prop1 n xvs)
  (lemmaA34_1_prop2 n xvs)
  (lemmaA34_1_prop3 n xvs)
).

(* This *)
unfold lemmaA34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termThis.
simpl. trivial.
(* Case Param *)
intros.
rename p into x.
unfold lemmaA34_1_prop1.
intros.
destruct (polySubst_in_termParam_corollary xvs x).
simpl.
rewrite H. simpl. trivial.
destruct H.
simpl. rewrite H.
apply lemmaA32.
(* Case Var *)
unfold lemmaA34_1_prop1.
intros.
rename v into i.
rewrite polySubst_in_termVar.
unfold lift_at_depth_in_term.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt.
rewrite polySubst_in_termVar. trivial. trivial.
rewrite match_comp_nat2_ge.
rewrite polySubst_in_termVar. trivial. trivial.
(* Select *)
unfold lemmaA34_1_prop1.
intros.
rewrite polySubst_in_termSelect.
simpl.
rewrite polySubst_in_termSelect.
apply f_equal2 with (f := Term_select).
apply lemmaA34_0'. trivial.
(* New *)
unfold lemmaA34_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
rewrite polySubst_in_termNew. simpl.
rewrite polySubst_in_termNew.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA34_1_prop2 in H. apply H.
(* Call *)
unfold lemmaA34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termCall. simpl.
rewrite polySubst_in_termCall.
apply f_equal3 with (f := Term_call).
apply lemmaA34_0'. trivial.
apply H.
(* Let *)
intros.
unfold lemmaA34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termLet. simpl.
rewrite polySubst_in_termLet.
apply f_equal3 with (f := Term_loc).
apply lemmaA34_2'.
apply H. apply H0.
(* Nil (fields) *)
unfold lemmaA34_1_prop2.
intros.
simpl.
rewrite polySubst_in_fieldTermNil.
simpl.
trivial.
(* Cons (fields) *)
intros.
unfold lemmaA34_1_prop2.
intros. simpl.
rewrite polySubst_in_fieldTermCons. simpl.
rewrite polySubst_in_fieldTermCons.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H.
apply H0.
(* Nil (params) *)
unfold lemmaA34_1_prop3.
intros.
simpl.
rewrite polySubst_in_paramTermNil.
simpl.
trivial.
(* Cons (params) *)
intros.
unfold lemmaA34_1_prop3.
intros. simpl.
rewrite polySubst_in_paramTermCons. simpl.
rewrite polySubst_in_paramTermCons.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H.
apply H0.
Qed.

(** Permuting lifting and self substitutions **)
Lemma lemmaA34_1:
  forall (t: Term) (n: nat) (xvs: (list (ParamSym * Value))),
    (lift_in_term (map_paramSubst_in_term t xvs) n) =
    (map_paramSubst_in_term (lift_in_term t n) xvs)
.

Proof.
intros.
unfold lift_in_term.
apply lemmaA34_1'.
Qed.

(******************************************************************)
(******************************************************************)
(** Facts about variable substitution **)
(******************************************************************)
(******************************************************************)

(** Facts about variable substitution **)
Lemma lemmaA37_0:
  forall (p: Path) (k: nat) (v: Value),
    (path_is_var_closed p) ->
    (varSubst_in_path p k v) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
cut (False). tauto.
rename H into H'.
assert (path_is_var_closed' (Path_var v0)).
apply path_is_var_closed_eq_r. trivial.
unfold path_is_var_closed' in H.
apply (H v0).
simpl. trivial.
simpl. trivial.
Qed.

(** Facts about variable substitution **)
Lemma lemmaA37_2:
  forall (T: type) (k: nat) (v: Value),
    (type_is_var_closed T) ->
    (varSubst_in_type T k v) = T
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
apply lemmaA37_0.
unfold path_is_var_closed.
unfold type_is_var_closed in H. simpl in H.
trivial.
trivial.
simpl. trivial.
Qed.

Lemma lemmaA37_0':
  forall (p: Path) (k: nat) (v: Value),
    (fv_at_depth_in_path p k) = empty_varset ->
    (varSubst_in_path p k v) = p
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
rename v0 into i.
unfold varSubst_in_path.
rewrite match_comp_nat3_lt. trivial.
apply fv_pathvar_empty. trivial.
simpl. trivial.
Qed.

Lemma lemmaA37_2':
  forall (T: type) (k: nat) (v: Value),
    (fv_at_depth_in_type T k) = empty_varset ->
    (varSubst_in_type T k v) = T
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
simpl in H.
apply lemmaA37_0'. trivial. trivial.
simpl. auto.
Qed.

Definition lemmaA37_1_prop1 (t: Term): Prop :=
  forall (k: nat) (v: Value),
    (fv_at_depth_in_term t k) = empty_varset ->
    (varSubst_in_term t k v) = t
.

Definition lemmaA37_1_prop2 (fts: FieldTermList): Prop :=
  forall (k: nat) (v: Value),
    (fv_at_depth_in_fieldTermList fts k) = empty_varset ->
    (varSubst_in_fieldTermList fts k v) = fts
.

Definition lemmaA37_1_prop3 (xts: ParamTermList): Prop :=
  forall (k: nat) (v: Value),
    (fv_at_depth_in_paramTermList xts k) = empty_varset ->
    (varSubst_in_paramTermList xts k v) = xts
.

(** Lifting a closed term **)
Lemma lemmaA37_1':
  forall (t: Term) (k: nat) (v: Value),
    (fv_at_depth_in_term t k) = empty_varset ->
    (varSubst_in_term t k v) = t
.

Proof.
(* Prelude *)
cut (forall (t: Term), (lemmaA37_1_prop1 t)).
intros. apply H. trivial.
apply (Term_prop
  lemmaA37_1_prop1
  lemmaA37_1_prop2
  lemmaA37_1_prop3
).

(* This *)
unfold lemmaA37_1_prop1.
simpl. trivial.
(* Param *)
unfold lemmaA37_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA37_1_prop1.
intros.
unfold varSubst_in_term.
rewrite match_comp_nat3_lt. trivial.
apply fv_termvar_empty. trivial.
(* Select *)
unfold lemmaA37_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA37_0'.
unfold fv_at_depth_in_term in H. trivial. trivial.
(* New *)
unfold lemmaA37_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA37_1_prop2 in H. apply H. trivial.
(* Call *)
unfold lemmaA37_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA37_0'.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H0.
trivial.
apply H.
simpl in H0.
unfold varset_union in H0. unfold empty_varset in H0.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H0.
(* Let *)
unfold lemmaA37_1_prop1.
intros.
simpl in H1.
simpl.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite lemmaA37_2'. rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H2.
assert ((set_union VarSym_eqdec (fv_at_depth_in_type t k)
           (fv_at_depth_in_term t0 k)) = (empty_set VarSym)).
eapply empty_union_left. apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H2.
(* Nil (fields) *)
unfold lemmaA37_1_prop2.
simpl. trivial.
(* Cons (fields) *)
unfold lemmaA37_1_prop2.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
(* Nil (params) *)
unfold lemmaA37_1_prop3.
simpl. trivial.
(* Cons (params) *)
unfold lemmaA37_1_prop3.
simpl. intros.
unfold varset_union in H1. unfold empty_varset in H1.
rewrite H. rewrite H0. trivial.
eapply empty_union_right with (eqdec := VarSym_eqdec). apply H1.
eapply empty_union_left with (eqdec := VarSym_eqdec). apply H1.
Qed.

(** Facts about variable substitution -- term  **)
Lemma lemmaA37_1:
  forall (t: Term) (k: nat) (v: Value),
    (term_is_var_closed t) ->
    (varSubst_in_term t 0 v) = t
.

Proof.
intros.
unfold term_is_var_closed in H.
apply lemmaA37_1'. apply H.
Qed.

(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)

(******************************************************************)
(******************************************************************)
(* Values are invariant by dropping *)
(******************************************************************)
(******************************************************************)

Definition lemmaA36_32_prop1 (v: Value): Prop :=
  forall (k: nat),
    (drop_at_depth_in_term (value_to_term v) k) = (value_to_term v)
.

Definition lemmaA36_32_prop2 (fvs: FieldValueList): Prop :=
  forall (k: nat),
    (drop_at_depth_in_fieldTermList (fieldValueList_to_fieldTermList fvs) k) =
    (fieldValueList_to_fieldTermList fvs)
.

(** Values are invariant by dropping **)
Lemma lemmaA36_32:
  forall (v: Value) (k: nat),
    (drop_at_depth_in_term (value_to_term v) k) = (value_to_term v)
.

Proof.
cut (forall (v: Value), (lemmaA36_32_prop1 v)).
intros. apply H.
apply (Value_prop
  lemmaA36_32_prop1
  lemmaA36_32_prop2
).

unfold lemmaA36_32_prop1.
unfold lemmaA36_32_prop2.
intros.
simpl. 
apply f_equal2 with (f := Term_new). trivial.
apply H.
unfold lemmaA36_32_prop2.
simpl. trivial.
intros.
unfold lemmaA36_32_prop1.
unfold lemmaA36_32_prop2.
intros.
simpl.
apply f_equal3 with (f := FieldTerm_cons). trivial.
apply H. apply H0.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting dropping and self substitution **)
(******************************************************************)
(******************************************************************)

(** Permuting dropping and self substitution **)
Lemma lemmaA36_33_0:
  forall (p: Path) (v: Value),
    (drop_in_path (thisSubst_in_path p (Path_value v))) =
    (thisSubst_in_path (drop_in_path p) (Path_value v))
. 

Proof.
intros.
unfold drop_in_path.
generalize 0.
induction p.
intros. simpl. trivial.
intros. simpl. trivial.
intros. simpl. rename n into k. rename v0 into i.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting dropping and self substitution **)
Lemma lemmaA36_33_2:
  forall (T: type) (v: Value),
    (drop_in_type (thisSubst_in_type T (Path_value v))) =
    (thisSubst_in_type (drop_in_type T) (Path_value v))
.

Proof.
intros.
induction T.
simpl.
fold (drop_in_path (thisSubst_in_path p (Path_value v))).
fold (drop_in_path p).
apply f_equal2 with (f := type_virtual).
apply lemmaA36_33_0. trivial.
simpl. trivial.
Qed.

(** Permuting dropping and self substitution **)
Lemma lemmaA36_33_0':
  forall (p: Path) (k: nat) (v: Value),
    (drop_at_depth_in_path (thisSubst_in_path p (Path_value v)) k) =
    (thisSubst_in_path (drop_at_depth_in_path p k) (Path_value v))
.

Proof.
intros.
generalize k. clear k.
induction p.
intros. simpl. trivial.
intros. simpl. trivial.
intros. simpl. rename v0 into i.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting dropping and self substitution **)
Lemma lemmaA36_33_2':
  forall (T: type) (k: nat) (v: Value),
    (drop_at_depth_in_type (thisSubst_in_type T (Path_value v)) k) =
    (thisSubst_in_type (drop_at_depth_in_type T k) (Path_value v))
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
apply lemmaA36_33_0'. trivial.
simpl. trivial.
Qed.

Definition lemmaA36_33_1_prop1 (t: Term): Prop :=
  forall (k: nat) (v: Value),
    (drop_at_depth_in_term (thisSubst_in_term t (Path_value v)) k) =
    (thisSubst_in_term (drop_at_depth_in_term t k) (Path_value v))
.

Definition lemmaA36_33_1_prop2 (fts: FieldTermList): Prop :=
  forall (k: nat) (v: Value),
    (drop_at_depth_in_fieldTermList (thisSubst_in_fieldTermList fts (Path_value v)) k) =
    (thisSubst_in_fieldTermList (drop_at_depth_in_fieldTermList fts k) (Path_value v))
.

Definition lemmaA36_33_1_prop3 (xts: ParamTermList): Prop :=
  forall (k: nat) (v: Value),
    (drop_at_depth_in_paramTermList (thisSubst_in_paramTermList xts (Path_value v)) k) =
    (thisSubst_in_paramTermList (drop_at_depth_in_paramTermList xts k) (Path_value v))
.

Lemma lemmaA36_33_1':
  forall (t: Term) (k: nat) (v: Value),
    (drop_at_depth_in_term (thisSubst_in_term t (Path_value v)) k) =
    (thisSubst_in_term (drop_at_depth_in_term t k) (Path_value v))
.

Proof.
(* Prelude *)
cut (forall (t: Term), (lemmaA36_33_1_prop1 t)).
intros. apply H.
apply (Term_prop
  lemmaA36_33_1_prop1
  lemmaA36_33_1_prop2
  lemmaA36_33_1_prop3
).

(* This *)
unfold lemmaA36_33_1_prop1.
intros.
simpl.
rewrite lemmaA36_32. trivial.
(* Param *)
unfold lemmaA36_33_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA36_33_1_prop1.
intros.
rename v into i.
simpl.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
(* Select *)
unfold lemmaA36_33_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA36_33_0'. trivial.
(* New *)
unfold lemmaA36_33_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA36_33_1_prop2 in H. apply H.
(* Call *)
unfold lemmaA36_33_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA36_33_0'. trivial.
apply H.
(* Let *)
intros.
unfold lemmaA36_33_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_loc).
apply lemmaA36_33_2'.
apply H. apply H0.
(* Nil (fields) *)
unfold lemmaA36_33_1_prop2.
simpl. trivial.
(* Cons (fields) *)
intros.
unfold lemmaA36_33_1_prop2.
intros. simpl.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H.
apply H0.
(* Nil (params) *)
unfold lemmaA36_33_1_prop3.
simpl. trivial.
(* Cons (params) *)
intros.
unfold lemmaA36_33_1_prop3.
intros. simpl.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H.
apply H0.
Qed.

(** Permuting dropping and self substitutions **)
Lemma lemmaA36_33_1:
  forall (t: Term) (v: Value),
    (drop_in_term (thisSubst_in_term t (Path_value v))) =
    (thisSubst_in_term (drop_in_term t) (Path_value v))
.

Proof.
intros.
unfold drop_in_term.
apply lemmaA36_33_1'.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting dropping and variable substitution **)
(******************************************************************)
(******************************************************************)

Lemma lemmaA36_35_0':
  forall (p: Path) (k k': VarSym) (v: Value),
    k' <= k ->
    ((k' = 0) -> ~(varset_In 0 (fv_in_path p))) ->
    (drop_at_depth_in_path (varSubst_in_path p (S k) v) k') =
    (varSubst_in_path (drop_at_depth_in_path p k') k v)
.

Proof.
intros.
generalize k k' H H0. clear k k' H H0.
induction p.

(* Case This *)
intros. simpl. trivial.

(* Case Param *)
intros. simpl. trivial.

(* Case Var *)
rename v0 into i.
intros.
destruct (lt_eq_lt_dec i (S k)).
destruct s.

(* Case i < k + 1*)
simpl.
rewrite match_comp_nat3_lt.

destruct (match_comp_nat2_aux_0 i k').

(* i < k' *)
rewrite match_comp_nat2_lt.
simpl.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial. omega. trivial. trivial.

(* i >= k' *)
simpl.
rewrite match_comp_nat2_ge.
assert (i <> 0).
intro.
assert (k' = 0). rewrite H1 in g. auto with arith.
apply H0. trivial.
simpl. rewrite match_comp_nat2_ge. simpl.
rewrite H1. simpl. tauto. auto with arith.
assert (0 < i). destruct i. destruct H1. trivial. auto with arith.
simpl. rewrite match_comp_nat3_lt. trivial.
omega. trivial. trivial.

(* simpl. *)
(* rewrite match_comp_nat3_lt. trivial. *)
(* assert (i <> 0). *)
(* intro. apply H0. *)
(* rewrite H1. *)
(* simpl. rewrite match_comp_nat2_ge. *)
(* simpl. tauto. omega. *)
(* assert (0 < i). destruct i. destruct H1. trivial. auto with arith. *)
(* omega. trivial. trivial. *)

(* Case i = k + 1 *)
simpl.
rewrite match_comp_nat3_eq. simpl.
rewrite match_comp_nat2_ge. simpl.
rewrite match_comp_nat3_eq. trivial.
rewrite e. simpl. trivial.
omega. trivial.

(* Case i > k + 1 *)
simpl.
rewrite match_comp_nat3_gt. simpl.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge. simpl.
rewrite match_comp_nat3_gt. trivial.
omega. omega. omega. omega.

(* Case Value *)
intros.
simpl. trivial.
Qed.

(** Permuting dropping and variable substitution **)
Lemma lemmaA36_35_0:
  forall (p: Path) (k: VarSym) (v: Value),
    ~(varset_In 0 (fv_in_path p)) ->
    (drop_in_path (varSubst_in_path p (S k) v)) =
    (varSubst_in_path (drop_in_path p) k v)
.

Proof.
intros.
unfold drop_in_path.
apply lemmaA36_35_0'. auto with arith. intro. trivial.
Qed.

Lemma lemmaA36_35_2':
  forall (T: type) (k k': nat) (v: Value),
    k' <= k ->
    ((k' = 0) -> ~(varset_In 0 (fv_in_type T))) ->
    (drop_at_depth_in_type (varSubst_in_type T (S k) v) k') =
    (varSubst_in_type (drop_at_depth_in_type T k') k v)
.

Proof.
intros.
induction T.
simpl.
apply f_equal2 with (f := type_virtual).
apply lemmaA36_35_0'. trivial.
simpl in H0. trivial. trivial.
simpl. trivial.
Qed.

(** Permuting dropping and variable substitution **)
Lemma lemmaA36_35_2:
  forall (T: type) (k: VarSym) (v: Value),
    ~(varset_In 0 (fv_in_type T)) ->
    (drop_in_type (varSubst_in_type T (S k) v)) =
    (varSubst_in_type (drop_in_type T) k v)
.

Proof.
intros.
unfold drop_in_type.
apply lemmaA36_35_2'.
auto with arith. intro. trivial.
Qed.

Definition lemmaA36_35_1_prop1 (v: Value) (t: Term): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    ((k' = 0) -> ~(varset_In 0 (fv_in_term t))) ->
    (drop_at_depth_in_term (varSubst_in_term t (S k) v) k') =
    (varSubst_in_term (drop_at_depth_in_term t k') k v)
.

Definition lemmaA36_35_1_prop2 (v: Value) (fts: FieldTermList): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    ((k' = 0) -> ~(varset_In 0 (fv_at_depth_in_fieldTermList fts 0))) ->
    (drop_at_depth_in_fieldTermList (varSubst_in_fieldTermList fts (S k) v) k') =
    (varSubst_in_fieldTermList (drop_at_depth_in_fieldTermList fts k') k v)
.

Definition lemmaA36_35_1_prop3 (v: Value) (xts: ParamTermList): Prop :=
  forall (k k': VarSym),
    (k' <= k) ->
    ((k' = 0) -> ~(varset_In 0 (fv_at_depth_in_paramTermList xts 0))) ->
    (drop_at_depth_in_paramTermList (varSubst_in_paramTermList xts (S k) v) k') =
    (varSubst_in_paramTermList (drop_at_depth_in_paramTermList xts k') k v)
.

Lemma lemmaA36_35_1':
  forall (t: Term) (k k': nat) (v: Value),
    (k' <= k) ->
    ((k' = 0) -> ~(varset_In 0 (fv_in_term t))) ->
    (drop_at_depth_in_term (varSubst_in_term t (S k) v) k') =
    (varSubst_in_term (drop_at_depth_in_term t k') k v)
.

Proof.
(* Prelude *)
cut (forall (v: Value) (t: Term), (lemmaA36_35_1_prop1 v t)).
intros. apply H. trivial. trivial.
intros v.
apply (Term_prop
  (lemmaA36_35_1_prop1 v)
  (lemmaA36_35_1_prop2 v)
  (lemmaA36_35_1_prop3 v)
).

(* This *)
unfold lemmaA36_35_1_prop1.
intros.
simpl.
trivial.
(* Param *)
unfold lemmaA36_35_1_prop1.
simpl. trivial.
(* Var *)
unfold lemmaA36_35_1_prop1.
intros.
rename v0 into i.
(* start: copy and paste*)
destruct (lt_eq_lt_dec i (S k)).
destruct s.

(* Case i < k + 1*)
simpl.
rewrite match_comp_nat3_lt.

destruct (match_comp_nat2_aux_0 i k').

(* i < k' *)
rewrite match_comp_nat2_lt.
simpl.
rewrite match_comp_nat2_lt.
rewrite match_comp_nat3_lt. trivial. omega. trivial. trivial.

(* i >= k' *)
simpl.
rewrite match_comp_nat2_ge.
assert (i <> 0).
intro.
assert (k' = 0). rewrite H1 in g. auto with arith.
apply H0. trivial.
unfold fv_in_term. simpl. rewrite match_comp_nat2_ge. simpl.
rewrite H1. simpl. tauto. auto with arith.
assert (0 < i). destruct i. destruct H1. trivial. auto with arith.
simpl. rewrite match_comp_nat3_lt. trivial.
omega. trivial. trivial.

(* simpl. *)
(* rewrite match_comp_nat2_ge. *)
(* simpl. *)
(* rewrite match_comp_nat3_lt. trivial. *)
(* assert (i <> 0). *)
(* intro. apply H0. *)
(* rewrite H1. *)
(* unfold fv_in_term. simpl. *)
(* rewrite match_comp_nat2_ge. *)
(* simpl. tauto. omega. *)
(* assert (0 < i). destruct i. destruct H1. trivial. auto with arith. *)
(* omega. trivial. trivial. *)

(* Case i = k + 1 *)
simpl.
rewrite match_comp_nat3_eq.
rewrite lemmaA36_32.
rewrite match_comp_nat2_ge. simpl.
rewrite match_comp_nat3_eq. trivial.
rewrite e. simpl. trivial.
omega. trivial.

(* Case i > k + 1 *)
simpl.
rewrite match_comp_nat3_gt. simpl.
rewrite match_comp_nat2_ge.
rewrite match_comp_nat2_ge. simpl.
rewrite match_comp_nat3_gt. trivial.
omega. omega. omega. omega.
(* end: copy and paste*)

(* Select *)
unfold lemmaA36_35_1_prop1.
intros.
simpl.
apply f_equal2 with (f := Term_select).
apply lemmaA36_35_0'. trivial. trivial. trivial.
(* New *)
unfold lemmaA36_35_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA36_35_1_prop2 in H. apply H. trivial. trivial.
(* Call *)
unfold lemmaA36_35_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_call).
apply lemmaA36_35_0'. trivial.
intro. intro.
apply H1. trivial.
unfold fv_in_term. simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro1. trivial. trivial.
apply H. trivial.
intro. intro. apply H1. trivial.
unfold fv_in_term. simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
(* Let *)
intros.
unfold lemmaA36_35_1_prop1.
intros.
simpl.
apply f_equal3 with (f := Term_loc).
apply lemmaA36_35_2'. trivial.
intro. intro. apply H2. trivial.
unfold fv_in_term. simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro1.
apply set_union_intro1. trivial.
apply H. trivial. intro.
intro. apply H2. trivial.
unfold fv_in_term. simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro1.
apply set_union_intro2. trivial.
unfold lemmaA36_35_1_prop1 in H0.
assert ((S k') <= (S k)). auto with arith.
apply H0. trivial.
intro.
absurd ((S k') = 0). auto with arith. trivial.

(* Nil (fields) *)
unfold lemmaA36_35_1_prop2.
simpl. trivial.
(* Cons (fields) *)
intros.
unfold lemmaA36_35_1_prop2.
intros. simpl.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H. trivial.
intro. intro. apply H2. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro1. trivial.
apply H0. trivial.
intro. intro. apply H2. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.

(* Nil (params) *)
unfold lemmaA36_35_1_prop3.
simpl. trivial.
(* Cons (params) *)
intros.
unfold lemmaA36_35_1_prop3.
intros. simpl.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H. trivial.
intro. intro. apply H2. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro1. trivial.
apply H0. trivial.
intro. intro. apply H2. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
Qed.

(** Permuting dropping and variable substitutions **)
Lemma lemmaA36_35_1:
  forall (t: Term) (k: VarSym) (v: Value),
    ~(varset_In 0 (fv_in_term t)) ->
    (drop_in_term (varSubst_in_term t (S k) v)) =
    (varSubst_in_term (drop_in_term t) k v)
.

Proof.
intros.
unfold drop_in_term.
apply lemmaA36_35_1'.
auto with arith. intro. trivial.
Qed.

(******************************************************************)
(******************************************************************)
(** Permuting dropping and parameter substitution **)
(******************************************************************)
(******************************************************************)

(** Permuting dropping and parameter substitution **)
Lemma lemmaA36_34_0':
  forall (p: Path) (k: VarSym) (xvs: list (ParamSym * Value)),
    (drop_at_depth_in_path (map_paramSubst_in_path p xvs) k) =
    (map_paramSubst_in_path (drop_at_depth_in_path p k) xvs)
.

Proof.
intros.
generalize k. clear k.
induction p.
(* Case This *)
simpl.
rewrite polySubst_in_pathThis. simpl. trivial.
(* Case Param *)
rename p into x.
destruct (polySubst_in_pathParam_corollary xvs x).
simpl.
rewrite H. simpl. trivial.
destruct H.
simpl. rewrite H. simpl. trivial.
(* Case Var *)
intro.
rename v into i.
rewrite polySubst_in_pathVar.
unfold drop_at_depth_in_path.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt.
rewrite polySubst_in_pathVar. trivial. trivial.
rewrite match_comp_nat2_ge.
rewrite polySubst_in_pathVar. trivial. trivial.
(* Case Value *)
simpl.
rewrite polySubst_in_pathValue. simpl. trivial.
Qed.

(** Permuting dropping and parameter substitution **)
Lemma lemmaA36_34_0:
  forall (p: Path) (xvs: list (ParamSym * Value)),
    (drop_in_path (map_paramSubst_in_path p xvs)) =
    (map_paramSubst_in_path (drop_in_path p) xvs)
.

Proof.
intros.
unfold drop_in_path.
apply lemmaA36_34_0'.
Qed.

(** Permuting dropping and parameter substitution **)
Lemma lemmaA36_34_2':
  forall (T: type) (k: VarSym) (xvs: list (ParamSym * Value)),
    (drop_at_depth_in_type (map_paramSubst_in_type T xvs) k) =
    (map_paramSubst_in_type (drop_at_depth_in_type T k) xvs)
.

Proof.
intros.
induction T.
rewrite polySubst_in_typeVirtual.
simpl.
rewrite polySubst_in_typeVirtual.
apply f_equal2 with (f := type_virtual).
apply lemmaA36_34_0'. trivial.
simpl.
rewrite polySubst_in_typeClass.
simpl. trivial.
Qed.

(** Permuting dropping and parameter substitution **)
Lemma lemmaA36_34_2:
  forall (T: type) (xvs: list (ParamSym * Value)),
    (drop_in_type (map_paramSubst_in_type T xvs)) =
    (map_paramSubst_in_type (drop_in_type T) xvs)
.

Proof.
intros.
unfold drop_in_type.
apply lemmaA36_34_2'.
Qed.

Definition lemmaA36_34_1_prop1 (xvs: list (ParamSym * Value)) (t: Term): Prop :=
  forall (k: nat) ,
    (drop_at_depth_in_term (map_paramSubst_in_term t xvs) k) =
    (map_paramSubst_in_term (drop_at_depth_in_term t k) xvs)
.

Definition lemmaA36_34_1_prop2 (xvs: list (ParamSym * Value)) (fts: FieldTermList): Prop :=
  forall (k: nat) ,
    (drop_at_depth_in_fieldTermList (map_paramSubst_in_fieldTermList fts xvs) k) =
    (map_paramSubst_in_fieldTermList (drop_at_depth_in_fieldTermList fts k) xvs)
.

Definition lemmaA36_34_1_prop3 (xvs: list (ParamSym * Value)) (xts: ParamTermList): Prop :=
  forall (k: nat) ,
    (drop_at_depth_in_paramTermList (map_paramSubst_in_paramTermList xts xvs) k) =
    (map_paramSubst_in_paramTermList (drop_at_depth_in_paramTermList xts k) xvs)
.

Lemma lemmaA36_34_1':
  forall (t: Term) (k: nat) (xvs: (list (ParamSym * Value))),
    (drop_at_depth_in_term (map_paramSubst_in_term t xvs) k) =
    (map_paramSubst_in_term (drop_at_depth_in_term t k) xvs)
.

Proof.
(* Prelude *)
intros.
cut (forall (t: Term), (lemmaA36_34_1_prop1 xvs t)).
intros. apply H. clear t k.
apply (Term_prop
  (lemmaA36_34_1_prop1 xvs)
  (lemmaA36_34_1_prop2 xvs)
  (lemmaA36_34_1_prop3 xvs)
).

(* This *)
unfold lemmaA36_34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termThis.
simpl. trivial.
(* Case Param *)
intros.
rename p into x.
unfold lemmaA36_34_1_prop1.
intros.
destruct (polySubst_in_termParam_corollary xvs x).
simpl.
rewrite H. simpl. trivial.
destruct H.
simpl. rewrite H.
apply lemmaA36_32.
(* Case Var *)
unfold lemmaA36_34_1_prop1.
intros.
rename v into i.
rewrite polySubst_in_termVar.
unfold drop_at_depth_in_term.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt.
rewrite polySubst_in_termVar. trivial. trivial.
rewrite match_comp_nat2_ge.
rewrite polySubst_in_termVar. trivial. trivial.
(* Select *)
unfold lemmaA36_34_1_prop1.
intros.
rewrite polySubst_in_termSelect.
simpl.
rewrite polySubst_in_termSelect.
apply f_equal2 with (f := Term_select).
apply lemmaA36_34_0'. trivial.
(* New *)
unfold lemmaA36_34_1_prop1.
intros.
rename c into C. rename f into fvs.
simpl.
rewrite polySubst_in_termNew. simpl.
rewrite polySubst_in_termNew.
apply f_equal2 with (f := Term_new). trivial.
unfold lemmaA36_34_1_prop2 in H. apply H.
(* Call *)
unfold lemmaA36_34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termCall. simpl.
rewrite polySubst_in_termCall.
apply f_equal3 with (f := Term_call).
apply lemmaA36_34_0'. trivial.
apply H.
(* Let *)
intros.
unfold lemmaA36_34_1_prop1.
intros.
simpl.
rewrite polySubst_in_termLet. simpl.
rewrite polySubst_in_termLet.
apply f_equal3 with (f := Term_loc).
apply lemmaA36_34_2'.
apply H. apply H0.
(* Nil (fields) *)
unfold lemmaA36_34_1_prop2.
intros.
simpl.
rewrite polySubst_in_fieldTermNil.
simpl.
trivial.
(* Cons (fields) *)
intros.
unfold lemmaA36_34_1_prop2.
intros. simpl.
rewrite polySubst_in_fieldTermCons. simpl.
rewrite polySubst_in_fieldTermCons.
apply f_equal3 with (f := FieldTerm_cons).
trivial.
apply H.
apply H0.
(* Nil (params) *)
unfold lemmaA36_34_1_prop3.
intros.
simpl.
rewrite polySubst_in_paramTermNil.
simpl.
trivial.
(* Cons (params) *)
intros.
unfold lemmaA36_34_1_prop3.
intros. simpl.
rewrite polySubst_in_paramTermCons. simpl.
rewrite polySubst_in_paramTermCons.
apply f_equal3 with (f := ParamTerm_cons).
trivial.
apply H.
apply H0.
Qed.

(** Permuting dropping and self substitutions **)
Lemma lemmaA36_34_1:
  forall (t: Term) (xvs: (list (ParamSym * Value))),
    (drop_in_term (map_paramSubst_in_term t xvs)) =
    (map_paramSubst_in_term (drop_in_term t) xvs)
.

Proof.
intros.
unfold drop_in_term.
apply lemmaA36_34_1'.
Qed.

End SetProgram.
