(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Miscellaneous.v *)
(* Description: Miscellaneous properties (Appendix 1) *)

(** * Miscellaneous properties *)

Require Import Calculus.
Require Import List.
Require Import ListSet.
Require Import MatchCompNat.
Require Import Util.
Require Import Omega.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of WellFormedness.v *)
Require WellFormedness.
Module My_WellFormedness := WellFormedness.SetProgram(My_Program).

Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(**************************************************************)
(* The program is well-formed *)
(**************************************************************)

(** The program is well-formed **)
Axiom program_is_well_formed_proof: WF_program.

(**************************************************************)
(* Short-cuts for well-formedness statements *)
(**************************************************************)

Lemma symRel_is_well_founded_proof: symRel_is_well_founded.

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

Lemma classes_smaller_than_types_proof: classes_smaller_than_types.

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

Lemma classes_are_well_formed_proof: classes_are_well_formed.

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

Lemma no_method_overriding_proof: no_method_overriding.

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

Lemma no_type_overriding_proof: no_type_overriding.

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

Lemma main_term_is_well_typed: (exists T: type, (Typing rootEnv main T)).

Proof.
induction program_is_well_formed_proof. trivial.
Qed.

(**************************************************************)
(* Some properties of well-formed members *)
(**************************************************************)

Lemma typeVal_subtype_of_upperBound:
  forall (L: TypeSym) (C: ClassSym) (S T: type),
    (typeUpperBound L) = T ->
    (typeValue L C) = (Some S) ->
    (Subtyping (classEnv C) S T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_typeVal C L S).
apply H4. trivial.
destruct H7. rewrite H in H11.
rewrite <- H11 in H12. trivial.
Qed.

Lemma typeVal_supertype_of_lowerBound:
  forall (L: TypeSym) (C: ClassSym) (S T: type),
    (typeLowerBound L) = (Some T) ->
    (typeValue L C) = (Some S) ->
    (Subtyping (classEnv C) T S)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_typeVal C L S).
apply H4. trivial.
destruct H7.
apply H13. trivial.
Qed.

Lemma methodBody_of_type_methodType:
  forall (m: MethodSym) (C: ClassSym) (t: Term) (T: type),
    (methodValue m C) = (Some t) ->
    (methodType m) = T ->
    (Typing (Env_mk (Owner_class C) (method_parameters m) nil) t T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_methodVal C m t).
apply H6. trivial.
destruct H7.
rewrite <- H9. rewrite <- H0. trivial.
Qed.

Lemma param_types_are_well_formed:
  forall (x: ParamSym) (T: type) (m: MethodSym) (C: ClassSym),
    (paramType x) = T ->
    (paramOwner x) = m ->
    (methodOwner m) = C ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H2.
assert (WF_methodDef C m). apply H6. trivial.
destruct H8.
eapply H10. apply H0. apply H.
Qed.

Lemma methodTypes_are_well_formed:
  forall (m: MethodSym) (C: ClassSym) (T: type),
    (methodOwner m) = C ->
    (methodType m) = T ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_methodDef C m). apply H5. trivial.
destruct H7.
rewrite H0 in H7.
rewrite H7. trivial.
Qed.

Lemma fieldTypes_are_well_formed:
  forall (f: FieldSym) (C: ClassSym) (T: type),
    (fieldOwner f) = C ->
    (fieldType f) = T ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_fieldDef C f). apply H2. trivial.
destruct H7.
rewrite H0 in H7.
rewrite H7. trivial.
Qed.

Lemma typeUpperBounds_are_well_formed:
  forall (L: TypeSym) (C: ClassSym) (T: type),
    (typeOwner L) = C ->
    (typeUpperBound L) = T ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_typeDef C L). apply H3. trivial.
destruct H7.
rewrite H0 in H7.
rewrite H7. trivial.
Qed.

Lemma typeLowerBounds_are_well_formed:
  forall (L: TypeSym) (C: ClassSym) (T: type),
    (typeOwner L) = C ->
    (typeLowerBound L) = (Some T) ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H1.
assert (WF_typeDef C L). apply H3. trivial.
destruct H7.
destruct (H10 T H0).
trivial.
Qed.

Lemma typeValues_are_well_formed:
  forall (L: TypeSym) (C: ClassSym) (T: type),
    (typeValue L C) = (Some T) ->
    (Kinding (classEnv C) T)
.

Proof.
intros.
assert (WF_class C). apply classes_are_well_formed_proof.
destruct H0.
assert (WF_typeVal C L T). apply H3. trivial.
destruct H6. trivial.
Qed.

(**************************************************************)
(* Admissibility of subsumption *)
(**************************************************************)

(** Admissibility of subsumption **)
Lemma lemmaA1:
  forall (G: Env) (t: Term) (T U: type),
    (Typing G t T) ->
    (Gen_Subtyping G T U) ->
    (Typing G t U)
.

Proof.
intros.
destruct H.
(* This *)
eapply T_This. apply H.
eapply Gen_S_Trans. apply H1. trivial.
unfold NoConstraint.applyTo. tauto.
(* Param *)
eapply T_Param. apply H. apply H1.
eapply Gen_S_Trans. apply H2. trivial.
unfold NoConstraint.applyTo. tauto.
(* Var *)
eapply T_Var. apply H. apply H1.
eapply Gen_S_Trans. apply H2. trivial.
unfold NoConstraint.applyTo. tauto.
(* Select *)
eapply T_Select. apply H. trivial. apply H2.
eapply Gen_S_Trans. apply H3. trivial.
unfold NoConstraint.applyTo. tauto.
(* Call *)
eapply T_Call. apply H. trivial. apply H2. trivial. trivial.
eapply Gen_S_Trans. apply H5. trivial.
unfold NoConstraint.applyTo. tauto.
(* New *)
eapply T_New. apply H. trivial. trivial.
eapply Gen_S_Trans. apply H3. trivial.
unfold NoConstraint.applyTo. tauto.
(* Let *)
eapply T_Let. apply H. trivial. apply H2. trivial.
eapply Gen_S_Trans. apply H4. trivial.
unfold NoConstraint.applyTo. tauto.
Qed.

(**************************************************************)
(* Self in self substitution *)
(**************************************************************)

(** Self in self substitution **)
Lemma lemmaA2_1:
  forall (p q: Path),
    (this_in_path (thisSubst_in_path q p)) ->
    (this_in_path p)
.

Proof.
intro p.
induction q.
simpl. tauto.
simpl. tauto.
simpl. tauto.
simpl. tauto.
Qed.

(** Self in self substitution **)
Lemma lemmaA2_2:
  forall (p: Path) (T: type),
    (this_in_type (thisSubst_in_type T p)) ->
    (this_in_path p)
.

Proof.
intro p.
induction T. rename t into L.
simpl.
apply lemmaA2_1.
simpl. tauto.
Qed.

Lemma lemmaA24_aux:
  forall (p: Path) (T: type),
    ~(this_in_path p) ->
    ~(this_in_type (thisSubst_in_type T p))
.

Proof.
unfold not.
intros.
assert (this_in_path p).
eapply lemmaA2_2. apply H0.
tauto.
Qed.

(*****************************************************************)
(** ??? **)
(*****************************************************************)

Lemma free_parameters_and_lifting_aux_0:
  forall (p: Path) (n k: nat),
    (fp_in_path (lift_at_depth_in_path p n k)) = (fp_in_path p)
.

Proof.
intros.
induction p.
simpl.  trivial.
simpl. trivial.
simpl. rename v into i.
destruct (match_comp_nat2_aux_0 i k).
rewrite match_comp_nat2_lt. simpl. trivial. trivial.
rewrite match_comp_nat2_ge. simpl. trivial. trivial.
simpl. trivial.
Qed.

Lemma free_parameters_and_lifting_aux_2:
  forall (T: type) (n k: nat),
    (fp_in_type (lift_at_depth_in_type T n k)) = (fp_in_type T)
.

Proof.
intros.
induction T.
simpl. apply free_parameters_and_lifting_aux_0.
simpl. trivial.
Qed.

(* Lemma free_parameters_and_lifting_aux_1: *)
(*   forall (t: Term) (n k: nat), *)
(*     (fp_in_term (lift_at_depth_in_term t n k)) = (fp_in_term t) *)
(* . *)

(* Proof. *)
(* Admitted. *)

Lemma free_parameters_and_lifting_0:
  forall (p: Path) (n: nat),
    (fp_in_path (lift_in_path p n)) = (fp_in_path p)
.

Proof.
intros.
unfold lift_in_path.
apply free_parameters_and_lifting_aux_0.
Qed.

Lemma free_parameters_and_lifting_2:
  forall (T: type) (n: nat),
    (fp_in_type (lift_in_type T n)) = (fp_in_type T)
.

Proof.
intros.
unfold lift_in_type.
apply free_parameters_and_lifting_aux_2.
Qed.

(* Lemma free_parameters_and_lifting_1: *)
(*   forall (t: Term) (n: nat), *)
(*     (fp_in_term (lift_in_term t n)) = (fp_in_term t) *)
(* . *)

(* Proof. *)
(* Admitted. *)

(*****************************************************************)
(** Free parameters are in the context **)
(*****************************************************************)

(** Function: first projection of a list of typed parameters **)
Definition paramTypeList_to_paramList (xTs: ParamTypeList): (list ParamSym) :=
  (map (fun a => match a with (x, T) => x end) xTs)
.

(************************************************************)

Lemma lemmaA3_aux_1:
  forall (x: ParamSym) (T: type) (xTs: ParamTypeList),
    (In (x, T) xTs) ->
    (In x (paramTypeList_to_paramList xTs))
.

Proof.
intros.
induction xTs.
simpl in H. tauto.
simpl. destruct a. simpl in H. destruct H.
left. injection H; intros. trivial.
right. apply IHxTs. trivial.
Qed.

(************************************************************)

Definition lemmaA3_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (x: ParamSym),
      (paramset_In x (paramset_union (fp_in_path p) (fp_in_type T))) ->
      (In x (paramTypeList_to_paramList xTs))
.

Definition lemmaA3_prop2 (G: Env) (T U: type) (H: (Gen_Subtyping G T U)): Prop :=
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (x: ParamSym),
      (paramset_In x (paramset_union (fp_in_type T) (fp_in_type U))) ->
      (In x (paramTypeList_to_paramList xTs))
.

(** Free parameters are in the context -- subtyping **)
Lemma lemmaA3_2:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T U: type),
    (Gen_Subtyping (Env_mk O xTs Us) T U) ->
    forall (x: ParamSym),
      (paramset_In x (paramset_union (fp_in_type T) (fp_in_type U))) ->
      (In x (paramTypeList_to_paramList xTs))
.

Proof.
cut (
  forall (G: Env) (T U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA3_prop2 G T U H)
).
unfold lemmaA3_prop2. intros.
eapply H with (O := O) (Us := Us). apply H0. trivial. trivial.
apply (Gen_Subtyping_prop3
  lemmaA3_prop1
  lemmaA3_prop2
).

(* P-This *)
intros. unfold lemmaA3_prop1.
intros. simpl in H1.
unfold lemmaA3_prop2 in H.
rewrite e in H0. injection H0; intros. rewrite <- H3.
eapply H. apply e.
simpl.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1.
(* P-Param *)
intros. unfold lemmaA3_prop1.
intros. simpl in H1. rename x0 into y.
unfold lemmaA3_prop2 in H.
rewrite e in H0. injection H0; intros. rewrite <- H3.
assert ((set_In y (x :: nil)) \/ (set_In y (fp_in_type S))).
eapply set_union_elim. apply H1. destruct H5.
simpl in H5. destruct H5. rewrite <- H5.
eapply lemmaA3_aux_1. apply i. tauto.
eapply H. apply e.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2. trivial.
(* P-Var*)
intros. unfold lemmaA3_prop1.
intros. simpl in H1.
unfold lemmaA3_prop2 in H.
rewrite e in H0. injection H0; intros. rewrite <- H3.
eapply H. apply e.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1.
(* P-Value *)
intros. unfold lemmaA3_prop1.
intros. simpl in H1.
unfold lemmaA3_prop2 in H.
eapply H. apply H0.
simpl. trivial.
(* S-Class *)
intros. unfold lemmaA3_prop2. intros.
simpl in H0. tauto.
(* S-Extends *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop2 in H.
simpl in H1.
eapply H. apply H0.
simpl.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1.
(* S-Up *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop1 in H.
unfold lemmaA3_prop2 in H0.
simpl in H2.
assert ((set_In x (fp_in_path p)) \/ (set_In x (fp_in_type S))).
eapply set_union_elim. apply H2.
destruct H3.
eapply H. apply H1.
simpl. trivial.
eapply H0. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2. trivial.
(* S-Down *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop1 in H.
unfold lemmaA3_prop2 in H0.
simpl in H2.
assert ((set_In x (fp_in_type S)) \/ (set_In x (fp_in_path p))).
eapply set_union_elim. apply H2.
destruct H3.
eapply H0. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro1.  trivial.
eapply H. apply H1.
simpl. trivial.
(* S-Alias-Left *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop1 in H.
unfold lemmaA3_prop2 in H0.
simpl in H2.
assert ((set_In x (fp_in_path p)) \/ (set_In x (fp_in_type S))).
eapply set_union_elim. apply H2.
destruct H3.
eapply H. apply H1.
simpl. trivial.
eapply H0. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2. trivial.
(* S-Alias-Right *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop1 in H.
unfold lemmaA3_prop2 in H0.
simpl in H2.
assert ((set_In x (fp_in_type S)) \/ (set_In x (fp_in_path p))).
eapply set_union_elim. apply H2.
destruct H3.
eapply H0. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro1.  trivial.
eapply H. apply H1.
simpl. trivial.
(* S-Virtual *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop1 in H.
simpl in H1.
assert ((set_In x (fp_in_path p)) \/ (set_In x (fp_in_path p))).
eapply set_union_elim. apply H1.
destruct H2.
eapply H. apply H0.
simpl. trivial.
eapply H. apply H0.
simpl. trivial.
(* S-Trans *)
intros. unfold lemmaA3_prop2. intros.
unfold lemmaA3_prop2 in H.
unfold lemmaA3_prop2 in H0.
assert ((set_In x (fp_in_type T)) \/ (set_In x (fp_in_type U))).
eapply set_union_elim. apply H2. destruct H3.
eapply H. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro1.  trivial.
eapply H0. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.  trivial.
Qed.

(** Free parameters are in the context -- path typing **)
Lemma lemmaA3_0:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (p: Path) (T: type),
    (Gen_PathTyping (Env_mk O xTs Us) p T) ->
    forall (x: ParamSym),
      (paramset_In x (paramset_union (fp_in_path p) (fp_in_type T))) ->
      (In x (paramTypeList_to_paramList xTs))
.

Proof.
cut (
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA3_prop1 G p T H)
).
unfold lemmaA3_prop1. intros.
eapply H with (O := O) (Us := Us). apply H0. trivial. trivial.

intros. unfold lemmaA3_prop1. induction H.
(* P-This *)
intros.
rewrite H in H1. injection H1; intros. rewrite <- H4.
eapply lemmaA3_2. rewrite H in H0. apply H0.
simpl.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H2.
(* P-Param *)
intros.
simpl in H3. rename x0 into y.
rewrite H in H2. injection H2; intros. rewrite <- H5.
assert ((set_In y (x :: nil)) \/ (set_In y (fp_in_type S))).
eapply set_union_elim. apply H3. destruct H7.
simpl in H7. destruct H7. rewrite <- H7.
eapply lemmaA3_aux_1. apply H0. tauto.
eapply lemmaA3_2. rewrite H in H1. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2. trivial.
(* P-Var*)
intros.
simpl in H3.
rewrite H in H2. injection H2; intros. rewrite <- H5.
eapply lemmaA3_2.  rewrite H in H1. apply H1.
unfold paramset_In. unfold paramset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H3.
(* P-Value *)
intros.
simpl in H4.
rewrite H3 in H2.
eapply lemmaA3_2. apply H2.
simpl. trivial.
Qed.

(* (\** Free parameters are in the context -- term typing **\) *)
(* Lemma lemmaA3_1: *)
(*   forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (t: Term) (T: type), *)
(*     (Typing (Env_mk O xTs Us) t T) -> *)
(*     forall (x: ParamSym), *)
(*       (paramset_In x (paramset_union (fp_in_term t) (fp_in_type T))) -> *)
(*       (In x (paramTypeList_to_paramList xTs)) *)
(* . *)

(* Proof. *)
(* Admitted. *)

(** Free parameters are in the context -- type well-formedness **)
Lemma lemmaA3_3:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T: type),
    (Kinding (Env_mk O xTs Us) T) ->
    forall (x: ParamSym),
      (paramset_In x (fp_in_type T)) ->      
      (In x (paramTypeList_to_paramList xTs))
.

Proof.
intros.
induction T.
inversion_clear H.
simpl in H0.
eapply lemmaA3_0. apply H1. simpl. trivial.
inversion_clear H.
simpl in H0. tauto.
Qed.

Lemma lemmaA3_3_corollary:
  forall (C: ClassSym) (T: type),
    (Kinding (classEnv C) T) ->
    forall (x: ParamSym), ~(paramset_In x (fp_in_type T))
.

Proof.
intros. intro.
cut (In x (paramTypeList_to_paramList nil)).
simpl. tauto.
eapply lemmaA3_3.
unfold classEnv in H. apply H. trivial.
Qed.

(*****************************************************************)
(** Free variables are in the context **)
(*****************************************************************)

Definition lemmaA4_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (varset_union (fv_at_depth_in_path p d) (fv_at_depth_in_type T d))) ->
        (k < (length Us) - d)
.

Definition lemmaA4_prop2 (G: Env) (T U: type) (H: (Gen_Subtyping G T U)): Prop :=
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (varset_union (fv_at_depth_in_type T d) (fv_at_depth_in_type U d))) ->
        (k < (length Us) - d) 
.

(** Free variables are in the context -- subtyping **)
Lemma lemmaA4_2':
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T U: type),
    (Gen_Subtyping (Env_mk O xTs Us) T U) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (varset_union (fv_at_depth_in_type T d) (fv_at_depth_in_type U d))) ->
        (k < (length Us) - d) 
.

Proof.
cut (
  forall (G: Env) (T U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA4_prop2 G T U H)
).
unfold lemmaA4_prop2. intros.
eapply H with (O := O) (xTs := xTs). apply H0. trivial. trivial. trivial.
apply (Gen_Subtyping_prop3
  lemmaA4_prop1
  lemmaA4_prop2
).

(* P-This *)
intros. unfold lemmaA4_prop1.
intros. simpl in H2.
unfold lemmaA4_prop2 in H.
eapply H. apply H0. trivial. simpl. trivial.
(* P-Param *)
intros. unfold lemmaA4_prop1.
intros. simpl in H2.
unfold lemmaA4_prop2 in H.
eapply H. apply H0. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H2.
(* P-Var*)
intros. unfold lemmaA4_prop1.
intros. simpl in H2.
destruct (match_comp_nat2_aux_0 n d).
rewrite (match_comp_nat2_lt n d) in H2.
unfold lemmaA4_prop2 in H.
eapply H. apply H0. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H2. trivial.
rewrite (match_comp_nat2_ge n d) in H2.
unfold lemmaA4_prop2 in H.
assert ((set_In k (n - d :: nil)) \/ (set_In k (fv_at_depth_in_type S d))).
eapply set_union_elim. apply H2. destruct H3.
simpl in H3. destruct H3.
rewrite <- H3.
cut (n < (length Us)). intros. omega.
rewrite e in H0. injection H0; intros. rewrite H4 in a.
tauto.
tauto.
eapply H. apply H0. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
trivial.
(* P-Value *)
intros. unfold lemmaA4_prop1.
intros. simpl in H2.
unfold lemmaA4_prop2 in H.
eapply H. apply H0. trivial.
simpl. trivial.
(* S-Class *)
intros. unfold lemmaA4_prop2. intros.
simpl in H1. tauto.
(* S-Extends *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop2 in H.
simpl in H2.
eapply H. apply H0. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H2.
(* S-Up *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop1 in H.
unfold lemmaA4_prop2 in H0.
simpl in H3.
assert ((set_In k (fv_at_depth_in_path p d)) \/ (set_In k (fv_at_depth_in_type S d))).
eapply set_union_elim. apply H3.
destruct H4.
eapply H. apply H1. trivial.
simpl. trivial.
eapply H0. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
(* S-Down *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop1 in H.
unfold lemmaA4_prop2 in H0.
simpl in H3.
assert ((set_In k (fv_at_depth_in_type S d)) \/ (set_In k (fv_at_depth_in_path p d))).
eapply set_union_elim. apply H3.
destruct H4.
eapply H0. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro1.  trivial.
eapply H. apply H1. trivial.
simpl. trivial.
(* S-Alias-Left *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop1 in H.
unfold lemmaA4_prop2 in H0.
simpl in H3.
assert ((set_In k (fv_at_depth_in_path p d)) \/ (set_In k (fv_at_depth_in_type S d))).
eapply set_union_elim. apply H3.
destruct H4.
eapply H. apply H1. trivial.
simpl. trivial.
eapply H0. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
(* S-Alias-Right *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop1 in H.
unfold lemmaA4_prop2 in H0.
simpl in H3.
assert ((set_In k (fv_at_depth_in_type S d)) \/ (set_In k (fv_at_depth_in_path p d))).
eapply set_union_elim. apply H3.
destruct H4.
eapply H0. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro1.  trivial.
eapply H. apply H1. trivial.
simpl. trivial.
(* S-Virtual *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop1 in H.
simpl in H2.
assert ((set_In k (fv_at_depth_in_path p d)) \/ (set_In k (fv_at_depth_in_path p d))).
eapply set_union_elim. apply H2.
destruct H3.
eapply H. apply H0. trivial.
simpl. trivial.
eapply H. apply H0. trivial.
simpl. trivial.
(* S-Trans *)
intros. unfold lemmaA4_prop2. intros.
unfold lemmaA4_prop2 in H.
unfold lemmaA4_prop2 in H0.
assert ((set_In k (fv_at_depth_in_type T d)) \/ (set_In k (fv_at_depth_in_type U d))).
eapply set_union_elim. apply H3. destruct H4.
eapply H. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro1.  trivial.
eapply H0. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.  trivial.
Qed.

(** Free variables are in the context -- path typing **)
Lemma lemmaA4_0':
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (p: Path) (T: type),
    (Gen_PathTyping (Env_mk O xTs Us) p T) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (varset_union (fv_at_depth_in_path p d) (fv_at_depth_in_type T d))) ->
        (k < (length Us) - d)
.

Proof.
cut (
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA4_prop1 G p T H)
).
unfold lemmaA4_prop1. intros.
eapply H with (O := O) (xTs := xTs). apply H0. trivial. trivial. trivial.

intros. unfold lemmaA4_prop1. induction H.
(* P-This *)
intros.
rewrite H in H1. injection H1; intros. rewrite <- H4.
eapply lemmaA4_2'. rewrite H in H0. apply H0. rewrite H4. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H3.
(* P-Param *)
intros.
simpl in H4.
rewrite H2 in H1.
eapply lemmaA4_2'. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H4.
(* P-Var*)
intros. simpl in H4.
destruct (match_comp_nat2_aux_0 n d).
rewrite (match_comp_nat2_lt n d) in H4.
rewrite H2 in H1.
eapply lemmaA4_2'. apply H1. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H4. trivial.
rewrite (match_comp_nat2_ge n d) in H4.
assert ((set_In k (n - d :: nil)) \/ (set_In k (fv_at_depth_in_type S d))).
eapply set_union_elim. apply H4. destruct H5.
simpl in H5. destruct H5.
rewrite <- H5.
cut (n < (length Us)). intros. omega.
rewrite H in H2. injection H2; intros. rewrite H6 in H0.
tauto.
tauto.
rewrite H2 in H1.
eapply lemmaA4_2'. apply H1. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
trivial.
(* P-Value *)
intros. simpl in H5.
rewrite H3 in H2.
eapply lemmaA4_2'. apply H2. trivial.
simpl. trivial.
Qed.

(** Free variables are in the context -- type well-formedness **)
Lemma lemmaA4_3':
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T: type),
    (Kinding (Env_mk O xTs Us) T) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (fv_at_depth_in_type T d)) ->
        (k < (length Us) - d)
.

Proof.
intros.
induction T.
inversion_clear H.
simpl in H1.
eapply lemmaA4_0'. apply H2. trivial. simpl. trivial.
inversion_clear H.
simpl in H1. tauto.
Qed.

(********************************************************************)

Lemma lemmaA4_0:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (p: Path) (T: type),
    (Gen_PathTyping (Env_mk O xTs Us) p T) ->
    forall (n: VarSym),
      (varset_In n (varset_union (fv_in_path p) (fv_in_type T))) ->
      (n < (length Us))
.

Proof.
intros.
cut (n < (length Us - 0)). rewrite <- minus_n_O. trivial.
eapply lemmaA4_0'. apply H. auto with arith. trivial.
Qed.

Lemma lemmaA4_2:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T U: type),
    (Gen_Subtyping (Env_mk O xTs Us) T U) ->
    forall (n: VarSym),
      (varset_In n (varset_union (fv_in_type T) (fv_in_type U))) ->
      (n < (length Us))
.

Proof.
intros.
cut (n < (length Us - 0)). rewrite <- minus_n_O. trivial.
eapply lemmaA4_2'. apply H. auto with arith. trivial.
Qed.

Lemma lemmaA4_3:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (T: type),
    (Kinding (Env_mk O xTs Us) T) ->
    forall (n: VarSym),
      (varset_In n (fv_in_type T)) ->
      (n < (length Us))
.

Proof.
intros.
cut (n < (length Us - 0)). rewrite <- minus_n_O. trivial.
eapply lemmaA4_3'. apply H. auto with arith. trivial.
Qed.

Lemma lemmaA4_3_corollary:
  forall (C: ClassSym) (T: type),
    (Kinding (classEnv C) T) ->
    (forall (n: nat), ~(varset_In n (fv_in_type T)))
.

Proof.
intros. intro.
cut (n < (length (nil (A := type)))).
simpl. intro. absurd (n < 0). auto with arith. trivial.
eapply lemmaA4_3.
unfold classEnv in H. apply H. trivial.
Qed.

Lemma lemmaA4_0_corollary:
  forall (v: Value) (T: type),
    (Gen_PathTyping rootEnv (Path_value v) T) ->
    forall (n: VarSym), ~(varset_In n (fv_in_type T))
.

Proof.
intros. intro.
absurd (n < (length (nil (A := type)))).
simpl. auto with arith.
eapply lemmaA4_0.
unfold rootEnv in H. apply H.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
Qed.

(*****************************************************************)

Definition lemmaA4_1_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop := 
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (varset_union (fv_at_depth_in_term t d) (fv_at_depth_in_type T d))) ->
        (k < (length Us) - d)
.

Definition lemmaA4_1_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList)
  (H: (ParamTermListTyping G p m xts)): Prop := 
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (d: nat),
      d <= (length Us) ->
      forall (k: VarSym),
        (varset_In k (fv_at_depth_in_paramTermList xts d)) ->
        (k < (length Us) - d)
.

Definition lemmaA4_1_prop3 (G: Env) (C: ClassSym) (fts: FieldTermList)
  (H: (FieldTermListTyping G C fts)): Prop := 
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk O xTs Us) ->
    forall (d: nat),
      d <= (length Us) ->
    forall (k: VarSym),
      (varset_In k (fv_at_depth_in_fieldTermList fts d)) ->
      (k < (length Us) - d)
.

Lemma lemmaA4_1:
  forall (O: Owner) (xTs: ParamTypeList) (Us: TypeList) (t: Term) (T: type),
    (Typing (Env_mk O xTs Us) t T) ->
    forall (n: VarSym),
      (varset_In n (varset_union (fv_in_term t) (fv_in_type T))) ->
      (n < (length Us))
.

Proof.
cut (
  forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
    (lemmaA4_1_prop1 G t T H)
).
intros.
unfold lemmaA4_1_prop1 in H.
pattern (length Us). rewrite minus_n_O.
eapply H with
  (O := O)
  (xTs := xTs)
  (Us := Us)
.
apply H0. trivial. auto with arith. trivial.
apply (Typing_prop
  lemmaA4_1_prop1
  lemmaA4_1_prop2
  lemmaA4_1_prop3
).

(* T-This *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite e in H. injection H; intros.  clear H.
clear xTs0 H3. clear O H4. rewrite H2 in e. clear Ts H2.
rewrite e in g.
eapply lemmaA4_2'. apply g. trivial.
simpl.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1.
(* T-Param *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite e in H. injection H; intros.  clear H.
clear xTs0 H3. clear O0 H4. rewrite H2 in e. clear Ts H2.
rewrite e in g.
eapply lemmaA4_2'. apply g. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1.
(* T-Var*)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite e in H. injection H; intros.  clear H.
clear xTs0 H3. clear O0 H4. rewrite H2 in e.
rewrite H2 in a. clear Ts H2.
rewrite e in g. clear e G.
destruct (match_comp_nat2_aux_0 n d).
  (* n < d *)
rewrite (match_comp_nat2_lt n d) in H1.
eapply lemmaA4_2'. apply g. trivial.
unfold varset_In. unfold varset_union.
apply set_union_intro2.
eapply set_union_emptyL. apply H1. trivial.
  (* n >= d *)
rewrite (match_comp_nat2_ge n d) in H1.
assert ((set_In k (n - d :: nil)) \/ (set_In k (fv_at_depth_in_type S d))).
eapply set_union_elim. apply H1. destruct H.
simpl in H. destruct H.
rewrite <- H.
cut (n < (length Us)). intros. omega. tauto. tauto.
eapply lemmaA4_2'. apply g. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
trivial.
(* T-Select *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite H in g. rewrite H in g0.
assert (
  (set_In k (fv_at_depth_in_path p d)) \/
  (set_In k (fv_at_depth_in_type S d))
).
eapply set_union_elim. apply H1. destruct H2.
eapply lemmaA4_0'. apply g. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
eapply lemmaA4_2'. apply g0. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
(* T-Call *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite H0 in g. rewrite H0 in g0.
assert (
  (set_In k (varset_union
    (fv_at_depth_in_path p d)
    (fv_at_depth_in_paramTermList xts d))) \/
  (set_In k (fv_at_depth_in_type S d))
).
eapply set_union_elim. apply H2. clear H2. destruct H3.
assert (
  (set_In k (fv_at_depth_in_path p d)) \/
  (set_In k (fv_at_depth_in_paramTermList xts d))
).
eapply set_union_elim. apply H2. clear H2. destruct H3.
eapply lemmaA4_0'. apply g. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
eapply H. apply H0. trivial. trivial.
eapply lemmaA4_2'. apply g0. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
(* T-New *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
rewrite H0 in g.
assert (
  (set_In k (fv_at_depth_in_fieldTermList fts d)) \/
  (set_In k (fv_at_depth_in_type S d))
).
eapply set_union_elim. apply H2. clear H2. destruct H3.
eapply H. apply H0. trivial. trivial.
eapply lemmaA4_2'. apply g. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
(* T-Let *)
intros. unfold lemmaA4_1_prop1. simpl. intros.
assert (
  (set_In k (varset_union
    (varset_union (fv_at_depth_in_type T d) (fv_at_depth_in_term t d))
    (fv_at_depth_in_term u (Datatypes.S d)))) \/
  (set_In k (fv_at_depth_in_type S d))
).
eapply set_union_elim. apply H3. clear H3. destruct H4.
assert (
  (set_In k (varset_union (fv_at_depth_in_type T d) (fv_at_depth_in_term t d))) \/
  (set_In k (fv_at_depth_in_term u (Datatypes.S d)))
).
eapply set_union_elim. apply H3. clear H3. destruct H4.
  (* for t and T *)
eapply H. apply H1. trivial.
assert (
  (set_In k (fv_at_depth_in_type T d)) \/
  (set_In k (fv_at_depth_in_term t d))
).
eapply set_union_elim. apply H3. clear H3. destruct H4.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
  (* for u *)
unfold lemmaA4_1_prop1 in H0.
rewrite e in H1. injection H1; intros. clear H1.
clear O0 H6. clear xTs0 H5.
rewrite <- H4 in H2. rewrite <- H4. clear Us H4.
    (* k < |Ts ++ (T :: nil)| - (d + 1) *)
assert ( k < (length (Ts ++ (T :: nil))) - (Datatypes.S d)).
eapply H0 with (O0 := O) (xTs0 := xTs). trivial.
rewrite length_of_append. simpl.
rewrite plus_comm. simpl. auto with arith.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
rewrite length_of_append in H1. simpl in H1.
rewrite plus_comm in H1. simpl in H1. trivial.
  (* for S *)
rewrite H1 in g.
eapply lemmaA4_2'. apply g. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro2. trivial.
(* T-Nil (param/term) *)
intros.  unfold lemmaA4_1_prop2. intros.
simpl in H1. tauto.
(* T-Cons (param/term) *)
intros.  unfold lemmaA4_1_prop2. simpl. intros.
assert (
  (set_In k (fv_at_depth_in_term t d)) \/
  (set_In k (fv_at_depth_in_paramTermList xts d))
).
eapply set_union_elim. apply H3. clear H3. destruct H4.
eapply H. apply H1. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
eapply H0. apply H1. trivial. trivial.
(* T-Nil (field/term) *)
intros.  unfold lemmaA4_1_prop3. intros.
simpl in H1. tauto.
(* T-Cons (field/term) *)
intros.  unfold lemmaA4_1_prop3. simpl. intros.
assert (
  (set_In k (fv_at_depth_in_term t d)) \/
  (set_In k (fv_at_depth_in_fieldTermList fts d))
).
eapply set_union_elim. apply H3. clear H3. destruct H4.
eapply H. apply H1. trivial.
unfold varset_In.  unfold varset_union.
apply set_union_intro1. trivial.
eapply H0. apply H1. trivial. trivial.
Qed.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

Lemma param_in_classEnv:
  forall (C: ClassSym) (x: ParamSym) (T: type),
    ~(PathTyping (classEnv C) (Path_param x) T)
.

Proof.
unfold not.
unfold classEnv.
intros.
inversion H.
destruct xTs. trivial.
discriminate H1.
Qed.

Lemma var_in_classEnv:
  forall (C: ClassSym ) (n: VarSym) (T: type),
    ~(PathTyping (classEnv C) (Path_var n) T)
.

Proof.
unfold not.
unfold classEnv.
intros.
inversion H.
destruct Ts.
destruct H2.
inversion H6.
discriminate H1.
Qed.

Lemma path_in_classEnv:
  forall (C: ClassSym) (p: Path) (T: type),
    (PathTyping (classEnv C) p T) ->
    (p = Path_this) \/
    exists v: Value, p = (Path_value v)      
.

Proof.
induction p.
intros. left.  trivial.
intros. induction (param_in_classEnv C p T). trivial.
intros. induction (var_in_classEnv C v T). trivial.
intros. right. exists v. trivial.
Qed.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

(** Chaining self substitutions **)
Lemma lemmaA8_1:
  forall (p1 p2 q: Path),
    (thisSubst_in_path (thisSubst_in_path q p1) p2) =
    (thisSubst_in_path q (thisSubst_in_path p1 p2))
.

Proof.
intros p1 p2. induction q.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
Qed.

(** Chaining self substitutions **)
Lemma lemmaA8_2:
  forall (p1 p2: Path) (T: type),
    (thisSubst_in_type (thisSubst_in_type T p1) p2) =
    (thisSubst_in_type T (thisSubst_in_path p1 p2))
.

Proof.
intros p1 p2. induction T.
simpl. rewrite lemmaA8_1. trivial.
simpl. trivial.
Qed.

(*****************************************************************)
(*****************************************************************)
(* Values are irreductible (not used) *)
(*****************************************************************)
(*****************************************************************)

Definition lemmaA5_P1 (t: Term): Prop :=
  forall (v: Value) (t': Term),
    (value_to_term v) = t ->
    ~(Red_term t t').

Definition lemmaA5_P2 (fts: FieldTermList): Prop :=
  forall (fvs: FieldValueList) (fts': FieldTermList),
    (fieldValueList_to_fieldTermList fvs) = fts ->
    ~(Red_fieldTermList fts fts').

Lemma lemmaA5: forall (t : Term), (lemmaA5_P1 t).

Proof.
apply (Term_prop lemmaA5_P1 lemmaA5_P2 (fun (xts: ParamTermList) => True)).

Proof.
unfold lemmaA5_P1.
unfold not.
intros.
induction v; simpl in H; discriminate H.

unfold lemmaA5_P1.
unfold not.
intros.
induction v; simpl in H; discriminate H.

unfold lemmaA5_P1.
unfold not.
intros.
induction v0; simpl in H; discriminate H.

unfold lemmaA5_P1.
unfold not.
intros.
induction v; simpl in H; discriminate H.

unfold lemmaA5_P1.
unfold not.
intros.
induction v.
simpl in H0.
inversion H0. clear H0.
unfold lemmaA5_P2 in H.
unfold not in H.
inversion H1.
apply (H f0 fts'). trivial. trivial.

unfold lemmaA5_P1.
unfold not.
intros.
induction v. simpl in H0. discriminate H0.

unfold lemmaA5_P1.
unfold not.
intros.
induction v; simpl in H1; discriminate H1.

unfold lemmaA5_P2.
unfold not.
intros.
inversion H0.

intros f t H fts H'.
unfold lemmaA5_P2.
unfold not.
intros.
induction fvs. simpl in H0.  discriminate H0.

simpl in H0.
inversion H0. clear H0.

inversion H1.
unfold lemmaA5_P1 in H.
unfold not in H.
apply (H v u). trivial. trivial.

clear H.
unfold lemmaA5_P2 in H'.
unfold not in H'.
apply (H' fvs fts'0). trivial. trivial.

trivial.
trivial.
Qed.

(** Some terms are not typable in empty context - 1 **)
Lemma lemmaA6_1:
  forall (T: type),
    ~(PathTyping rootEnv Path_this T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
discriminate H0.
Qed.

(** Some terms are not typable in empty context - 1 **)
Lemma lemmaA6_1':
  forall (T: type),
    ~(Gen_PathTyping rootEnv Path_this T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
discriminate H0.
Qed.

(** Some terms are not typable in empty context - 2 **)
Lemma lemmaA6_2:
  forall (x: ParamSym) (T: type),
    ~(PathTyping rootEnv (Path_param x) T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
destruct xTs.
trivial.
discriminate H1.
Qed.

(** Some terms are not typable in empty context - 2 **)
Lemma lemmaA6_2':
  forall (x: ParamSym) (T: type),
    ~(Gen_PathTyping rootEnv (Path_param x) T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
destruct xTs.
trivial.
discriminate H1.
Qed.

(** Some terms are not typable in empty context - 3 **)
Lemma lemmaA6_3:
  forall (n: VarSym) (T: type),
    ~(PathTyping rootEnv (Path_var n) T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
destruct H2.
inversion H6.
rewrite <- H7 in H1.
discriminate H1.
rewrite <- H8 in H1.
discriminate H1.
Qed.

(** Some terms are not typable in empty context - 3 **)
Lemma lemmaA6_3':
  forall (n: VarSym) (T: type),
    ~(Gen_PathTyping rootEnv (Path_var n) T)
.

Proof.
unfold not.
unfold rootEnv.
intros.
inversion H.
destruct H2.
inversion H6.
rewrite <- H7 in H1.
discriminate H1.
rewrite <- H8 in H1.
discriminate H1.
Qed.

(** Paths typable in empty context are values *)
Lemma lemmaA7:
  forall (p: Path) (T: type),
    (PathTyping rootEnv p T) ->
    exists v: Value, p = (Path_value v)
.

Proof.
induction p.
intros. induction (lemmaA6_1 T). trivial.
intros. induction (lemmaA6_2 p T). trivial.
intros. induction (lemmaA6_3 v T). trivial.
intros. exists v. trivial.
Qed.

(** Paths typable in empty context are values *)
Lemma lemmaA7':
  forall (p: Path) (T: type),
    (Gen_PathTyping rootEnv p T) ->
    exists v: Value, p = (Path_value v)
.

Proof.
induction p.
intros. induction (lemmaA6_1' T). trivial.
intros. induction (lemmaA6_2' p T). trivial.
intros. induction (lemmaA6_3' v T). trivial.
intros. exists v. trivial.
Qed.

(** Decidability of equality for values **)

Definition Value_eqdec' (v: Value): Set := forall (v' : Value), { v = v' }+{ v <> v' }.
Definition FieldValueList_eqdec' (fvs: FieldValueList): Set := forall (fvs': FieldValueList), { fvs = fvs' }+{ fvs <> fvs' }.

Lemma Value_eqdec: forall (v: Value), (Value_eqdec' v).

Proof.
apply (Value_set Value_eqdec' FieldValueList_eqdec').
intros C fvs H.
unfold FieldValueList_eqdec' in H.
unfold Value_eqdec'.
intros v.
induction v. rename f into fvs'. rename c into C'.
induction (ClassSym_eq C C').
induction (H fvs').
left.
rewrite a. rewrite a0. trivial.
right.
unfold not.
intro.
inversion H0.
apply (b H3).
right.
unfold not. intro. inversion H0. apply (b H2).

unfold FieldValueList_eqdec'.
induction fvs'.
left. trivial.
right.
unfold not. intro. inversion H.

unfold FieldValueList_eqdec'.
unfold Value_eqdec'.
intros.
induction fvs'.
right.
unfold not. intro. inversion H1.
induction (FieldSym_eq f f1).
induction (H v0).
induction (H0 fvs').
rewrite a. rewrite a0. rewrite a1.
left. trivial.
right.
unfold not. intro. inversion H1. apply (b H5).
right. unfold not. intro. inversion H1. apply (b H4).
right. unfold not. intro. inversion H1. apply (b H3).
Qed.

(** Decidability of equality for paths **)
Lemma Path_eqdec: forall (p q : Path), { p = q }+{ p <> q }.

Proof.
decide equality.
apply (ParamSym_eq).
unfold VarSym in v.
unfold VarSym in v0.
auto with arith.
apply (Value_eqdec).
Qed.

(** Decidability of equality for types **)
Lemma Type_eqdec : forall (T U : type), { T = U }+{ T <> U }.

Proof.
decide equality.
apply TypeSym_eq.
apply (Path_eqdec).
apply ClassSym_eq.
Qed.

(*****************************************************************)
(*****************************************************************)
(* Chaining self and parameter substitutions *)
(*****************************************************************)
(*****************************************************************)

Definition paramValueList_to_paramList (xvs: ParamValueList): (list ParamSym) :=
  (map (fun a => match a with (x, v) => x end) xvs)
.

Lemma lemmaA9_aux_0:
  forall (x: ParamSym) (xvs: ParamValueList),
    ~(In x (paramValueList_to_paramList xvs)) ->
    (map_paramSubst_in_path (Path_param x) xvs) = (Path_param x)
.

Proof.
intros x.
induction xvs.
simpl. intro. trivial.
intros. destruct a.
simpl in H. simpl.
assert (~(p = x)). tauto.
assert (~(In x (paramValueList_to_paramList xvs))). tauto.
clear H.
destruct (ParamSym_eq p x). destruct (H0 e).
apply IHxvs. trivial.
Qed.

Lemma lemmaA9_0:
  forall (q: Path) (p: Path) (xvs: ParamValueList),
    (forall (x: ParamSym), (In x (paramValueList_to_paramList xvs)) -> ~(param_in_path x q)) ->
    (map_paramSubst_in_path (thisSubst_in_path q p) xvs) =
    (thisSubst_in_path q (map_paramSubst_in_path p xvs)) 
.

Proof.
intros. generalize q H. clear q H. 
destruct q.
(* This *)
intros. simpl. trivial.
(* Param *)
intros. simpl.
assert (~(In p0 (paramValueList_to_paramList xvs))).
intro.
apply (H p0). trivial.
unfold param_in_path. simpl. tauto.
apply lemmaA9_aux_0. trivial.
(* Var *)
intros. simpl.
rewrite polySubst_in_pathVar. trivial.
(* Value *)
intros. simpl.
rewrite polySubst_in_pathValue. trivial.
Qed.

Lemma lemmaA9_2:
  forall (T: type) (p: Path) (xvs: ParamValueList),
    (forall (x: ParamSym), (In x (paramValueList_to_paramList xvs)) -> ~(param_in_type x T)) ->
    (map_paramSubst_in_type (thisSubst_in_type T p) xvs) =
    (thisSubst_in_type T (map_paramSubst_in_path p xvs)) 
.

Proof.
intros. generalize T H. clear T H.
induction T.
(* Virtual *)
simpl. intros.
rewrite polySubst_in_typeVirtual.
apply f_equal2 with (f := type_virtual).
apply lemmaA9_0.
intros. intro.
apply (H x). trivial.
unfold param_in_type. simpl. trivial. trivial.
(* Class *)
simpl. intros.
rewrite polySubst_in_typeClass. trivial.
Qed.

(* not used *)
(* Lemma lemmaA9_1: *)
(*   forall (t: Term) (p: Path) (xvs: ParamValueList), *)
(*     (forall (x: ParamSym), (In x (paramValueList_to_paramList xvs)) -> ~(param_in_term x t)) -> *)
(*     (map_paramSubst_in_term (thisSubst_in_term t p) xvs) = *)
(*     (thisSubst_in_term t (map_paramSubst_in_path p xvs))  *)
(* . *)

(* Proof. *)
(* Admitted. *)

(*****************************************************************)
(*****************************************************************)
(* Chaining self and variable substitutions *)
(*****************************************************************)
(*****************************************************************)

Lemma lemmaA10_0:
  forall (q: Path) (p: Path) (k: nat) (v: Value),
    (fv_in_path q) = empty_varset ->
    (varSubst_in_path (thisSubst_in_path q p) k v) =
    (thisSubst_in_path q (varSubst_in_path p k v)) 
.

Proof.
intros.
generalize q H. clear q H.
induction q.
(* This *)
simpl. intros. trivial.
(* Param *)
simpl. intro. trivial.
(* Var *)
intro.
cut False. tauto.
simpl in H.
rewrite match_comp_nat2_ge in H.
discriminate H. auto with arith.
(* Value *)
simpl. intros. trivial.
Qed.

Lemma lemmaA10_2:
  forall (T: type) (p: Path) (k: nat) (v: Value),
    (fv_in_type T) = empty_varset ->
    (varSubst_in_type (thisSubst_in_type T p) k v) =
    (thisSubst_in_type T (varSubst_in_path p k v)) 
.

Proof.
intros.
generalize T H. clear T H.
induction T.
(* Virtual *)
simpl. intro.
apply f_equal2 with (f := type_virtual).
apply lemmaA10_0. trivial. trivial.
(* Class *)
simpl. intros. trivial.
Qed.

(* not used *)
(* Lemma lemmaA10_1: *)
(*   forall (t: Term) (p: Path) (k: nat) (v: Value), *)
(*     (fv_in_term t) = empty_varset -> *)
(*     (varSubst_in_term (thisSubst_in_term t p) k v) = *)
(*     (thisSubst_in_term t (varSubst_in_path p k v))  *)
(* . *)

(* Proof. *)
(* Admitted. *)

(*****************************************************************)
(*****************************************************************)
(* Additional lemmas *)
(*****************************************************************)
(*****************************************************************)

Lemma new_lemma1_0:
  forall (p: Path) (q: Path),
    ~(this_in_path p) ->
    (thisSubst_in_path p q) = p
.

Proof.
destruct p.
intros. simpl in H. tauto.
intros. simpl. trivial.
intros. simpl. trivial.
intros. simpl. trivial.
Qed.

Lemma new_lemma1_2:
  forall (T: type) (q: Path),
    ~(this_in_type T) ->
    (thisSubst_in_type T q) = T
.

Proof.
destruct T.
intros. simpl. apply f_equal2 with (f := type_virtual).
simpl in H. apply new_lemma1_0. trivial. trivial.
intros. simpl. trivial.
Qed.

(* not used *)
(* Lemma new_lemma1_1: *)
(*   forall (t: Term) (q: Path), *)
(*     ~(this_in_term t) -> *)
(*     (thisSubst_in_term t q) = t *)
(* . *)

(* Proof. *)
(* Qed. *)

(***********************************************************)
(* About free parameters *)
(***********************************************************)

Lemma new_lemma2_0:
  forall (q: Path),
    (fp_in_path q) = empty_paramset ->
    forall (p: Path),
      (fp_in_path (thisSubst_in_path p q)) = (fp_in_path p)
.

Proof.
intros.
induction p.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
Qed.

Lemma new_lemma2_2:
  forall (q: Path),
    (fp_in_path q) = empty_paramset ->
    forall (T: type),
      (fp_in_type (thisSubst_in_type T q)) = (fp_in_type T)
.

Proof.
intros.
induction T.
simpl. apply new_lemma2_0.  trivial.
simpl. trivial.
Qed.

Lemma fp_and_paramSubst_in_path:
  forall (p: Path) (xvs: ParamValueList),
    (fp_in_path p) = empty_paramset ->
    (map_paramSubst_in_path p xvs) = p
.

Proof.
intros.
destruct p.
rewrite polySubst_in_pathThis. trivial.
simpl in H. discriminate H.
rewrite polySubst_in_pathVar. trivial.
rewrite polySubst_in_pathValue. trivial.
Qed.

Lemma fp_and_paramSubst_in_type:
  forall (T: type) (xvs: ParamValueList),
    (fp_in_type T) = empty_paramset ->
    (map_paramSubst_in_type T xvs) = T
.

Proof.
intros.
destruct T.
rewrite polySubst_in_typeVirtual.
apply f_equal2 with (f := type_virtual).
apply fp_and_paramSubst_in_path.
simpl in H. trivial. trivial.
rewrite polySubst_in_typeClass. trivial.
Qed.

(******************************************************)

Lemma fv_and_paramSubst_in_path:
  forall (p: Path) (xvs: ParamValueList),
    (fv_in_path (map_paramSubst_in_path p xvs)) = (fv_in_path p)
.

Proof.
induction p.
simpl. intros. rewrite polySubst_in_pathThis. simpl. trivial.
simpl. intros.
destruct (polySubst_in_pathParam_corollary xvs p).
rewrite H. simpl. trivial.
destruct H. rewrite H. simpl. trivial.
intros. rewrite polySubst_in_pathVar. trivial.
intros. rewrite polySubst_in_pathValue. trivial.
Qed.

Lemma fv_and_paramSubst_in_type:
  forall (T: type) (xvs: ParamValueList),
    (fv_in_type (map_paramSubst_in_type T xvs)) = (fv_in_type T)
.

Proof.
intros.
induction T.
rewrite polySubst_in_typeVirtual. simpl.
fold (fv_in_path (map_paramSubst_in_path p xvs)).
fold (fv_in_path p).
apply fv_and_paramSubst_in_path.
simpl. rewrite polySubst_in_typeClass. simpl. trivial.
Qed.

End SetProgram.
