(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Proofs_Progress.v *)
(* Description: Property of progress (Appendix 5) *)

(** * Proof of the Progress Property *)

Require Import Calculus.
Require Import List.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_AdmissibilityOfTransitivity.v *)
Require Proofs_AdmissibilityOfTransitivity.
Module My_Proofs_AdmissibilityOfTransitivity := Proofs_AdmissibilityOfTransitivity.SetProgram(My_Program).

Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)
(*****************************************************************)
(*****************************************************************)

Lemma lemmaA29_aux_1:
  forall (f: FieldSym) (fvs: FieldValueList),
    (In f (fields_of_fieldValueList fvs)) ->
    exists v: Value,
      (FieldValueList_contains fvs f v)
.

Proof.
intros.
induction fvs.
simpl in H. tauto.
simpl in H.
destruct H.
exists v.
rewrite H.
apply FieldValueList_contains_head.
destruct (IHfvs H).
exists x.
apply FieldValueList_contains_tail. trivial.
Qed.

(*****************************************************************)

Definition params_evaluated (xts: ParamTermList): Prop :=
  exists xvs: (list (ParamSym * Value)),
    (paramValueList_to_ParamTermList xvs) = xts
.  

Definition fields_evaluated (fts: FieldTermList): Prop :=
  exists fvs: FieldValueList,
    (fieldValueList_to_fieldTermList fvs) = fts
.  

(*****************************************************************)

Definition lemmaA29_aux_2_prop1 (t: Term): Prop :=
    (term_evaluated t) \/ ~(term_evaluated t)
.

Definition lemmaA29_aux_2_prop2 (fts: FieldTermList): Prop :=
    (fields_evaluated fts) \/ ~(fields_evaluated fts)
.

Definition lemmaA29_aux_2_prop3 (xts: ParamTermList): Prop :=
    (params_evaluated xts) \/ ~(params_evaluated xts)
.

Lemma lemmaA29_aux_2:
  forall (t: Term),
    (term_evaluated t) \/ ~(term_evaluated t)
.

Proof.
cut (forall (t: Term),
  (lemmaA29_aux_2_prop1 t)
).
intros.
apply H.
apply (Term_prop
  lemmaA29_aux_2_prop1
  lemmaA29_aux_2_prop2
  (fun (xts: ParamTermList) => True)
).

(* Various simple cases *)
right. intro. destruct H. destruct x. simpl in H. discriminate H.
right. intro. destruct H. destruct x. simpl in H. discriminate H.
right. intro. destruct H. destruct x. simpl in H. discriminate H.
right. intro. destruct H. destruct x. simpl in H. discriminate H.
Focus 2. intros. clear H.
right. intro. destruct H. destruct x. simpl in H. discriminate H.
Focus 2. intros.  clear H H0.
right. intro. destruct H. destruct x. simpl in H. discriminate H.

(* Case New *)
intros.
rename c into C. rename f into fts.
unfold lemmaA29_aux_2_prop1.
unfold term_evaluated.
unfold lemmaA29_aux_2_prop2 in H.
destruct H.
destruct H. rename x into fvs.
left. exists (Value_mk C fvs).
simpl. rewrite H. trivial.
right. intro. destruct H0. destruct x.
rename f into fvs.
simpl in H0.
injection H0. intros.
eapply H. exists fvs. trivial.

(* Case nil (fields) *)
unfold lemmaA29_aux_2_prop2.
left.
unfold fields_evaluated.
exists FieldValue_nil.
simpl. trivial.

(* Case cons (fields) *)
intros f t IH1 fts IH2.
unfold lemmaA29_aux_2_prop2.
destruct IH1.
destruct IH2.
left.
destruct H.
destruct H0.
exists (FieldValue_cons f x x0).
simpl. rewrite H. rewrite H0. trivial.
right.
intro.
apply H0. destruct H1. destruct x.
simpl in H1. discriminate.
simpl in H1. injection H1. intros.
exists x. trivial.
right. intro. apply H. destruct H0. destruct x.
simpl in H0. discriminate.
simpl in H0. injection H0. intros.
exists v. trivial.

tauto. tauto.
Qed.

Lemma lemmaA29_aux_3:
  forall (xts: ParamTermList),
    (params_evaluated xts) \/ ~(params_evaluated xts)
.

Proof.
intros.
induction xts.
(* Case nil (fields) *)
left.
exists (nil (A := (ParamSym * Value)%type)).
simpl. trivial.

(* Case cons (fields) *)
rename p into x.
destruct (lemmaA29_aux_2 t).
destruct IHxts.
left.
destruct H.
destruct H0.
exists (cons (x,x0) x1).
simpl. rewrite H. rewrite H0. trivial.
right.
intro.
apply H0. destruct H1. destruct x0.
simpl in H1. discriminate.
destruct p.
simpl in H1. injection H1. intros.
exists x0. trivial.
right. intro. apply H. destruct H0. destruct x0.
simpl in H0. discriminate. destruct p.
simpl in H0. injection H0. intros.
exists v. trivial.
Qed.

(*****************************************************************)

Definition lemmaA29_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop :=
  G = rootEnv -> 
  ~(term_evaluated t) ->
  exists u: Term, (Red_term t u)
.

Definition lemmaA29_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList) (H: (ParamTermListTyping G p m xts)): Prop :=
  G = rootEnv ->
  ~(params_evaluated xts) ->
  exists xts': ParamTermList, (Red_paramTermList xts xts')
.

Definition lemmaA29_prop3 (G: Env) (C: ClassSym) (fts: FieldTermList) (H: (FieldTermListTyping G C fts)): Prop :=
  G = rootEnv ->
  ~(fields_evaluated fts) ->
  exists fts': FieldTermList, (Red_fieldTermList fts fts')
.

(*****************************************************************)

(** Progress **)
Lemma lemmaA29:
  forall (t: Term) (T: type),
    (Typing rootEnv t T) ->
    ~(term_evaluated t) ->
    exists u: Term, (Red_term t u)
.

Proof.
cut (
  forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
    (lemmaA29_prop1 G t T H)
).
intros.
unfold lemmaA29_prop1 in H.
eapply H. apply H0. trivial. trivial.

intros G t T H.
apply (Typing_prop
  lemmaA29_prop1
  lemmaA29_prop2
  lemmaA29_prop3
).
(* Case T-This *)
unfold lemmaA29_prop1.
intros.
rewrite H0 in e.
discriminate e.

(* Case T-Param *)
unfold lemmaA29_prop1.
intros.
destruct xTs.
destruct i.
simpl in i.
destruct i.
rewrite H0 in e. discriminate e.
rewrite H0 in e. discriminate e.

(* Case T-Var *)
unfold lemmaA29_prop1.
intros.
destruct a.
induction H3.
rewrite H0 in e. discriminate e.
rewrite H0 in e. discriminate e.

(* Case T-Select *)
unfold lemmaA29_prop1.
intros.
rewrite H0 in g.
destruct (lemmaA7' p (type_class C) g).
rename x into v.
rewrite H2 in g.
destruct v. rename c into B. rename f0 into fvs.
inversion_clear g.
assert (Subclass B C). eapply lemmaA21. eapply lemmaA28_2. apply H6.
inversion_clear H5. clear H9 H10.
unfold IsComplete_fields in H8.
assert (In f (fields_of_fieldValueList fvs)).
apply (H8 C f H7 e). clear H8.
destruct (lemmaA29_aux_1 f fvs H5).
exists (value_to_term x).
rewrite H2.
eapply R_Select. apply H8.

(* Case T-Call *)
unfold lemmaA29_prop1.
intros.
rewrite H1 in g.
destruct (lemmaA7' p (type_class C) g).
rename x into v.
rewrite H3 in g.
destruct (lemmaA29_aux_3 xts).
(* R-Call *)
destruct H4.
rename x into xvs.
destruct v. rename c into B. rename f into fvs.
inversion_clear g.
assert (Subclass B C). eapply lemmaA21.
apply lemmaA28_2. apply H8.
inversion_clear H7. clear H10 H12.
unfold IsComplete_methods in H11.
destruct (H11 C m H9 e). clear H11. destruct H7. destruct H7.
rename x into A. rename x0 into t'.
exists (
  (map_paramSubst_in_term (thisSubst_in_term t' (Path_value (Value_mk B fvs))) xvs)
).
rewrite H3.
apply R_Call with (C := B) (C' := A) (fws := fvs).
trivial. trivial. trivial. trivial.
(* R-Arg *)
unfold lemmaA29_prop2 in H0.
destruct (H0 H1 H4). clear H0.
rewrite H3.
rename x into xts'.
exists (Term_call (Path_value v) m xts').
eapply R_Arg. trivial.

(* Case T-New *)
unfold lemmaA29_prop1.
intros.
unfold lemmaA29_prop3 in H0.
assert (~(fields_evaluated fts)). intro. apply H2.
destruct H3.
exists (Value_mk C x). simpl. rewrite H3. trivial.
destruct (H0 H1 H3). 
rename x into fts'.
exists (Term_new C fts').
eapply R_Field. trivial.

(* Case T-Let *)
unfold lemmaA29_prop1.
intros.
destruct (lemmaA29_aux_2 t0).
(* R-Let *)
destruct H4. rename x into v.
exists (varSubst_in_term u 0 v).
apply R_Let. trivial.
(* R-Local *)
destruct (H0 H2 H4).
exists (Term_loc T0 x u).
apply R_Local. trivial.

(* Case T-Nil (params) *)
unfold lemmaA29_prop2.
intros.
cut False. tauto.
apply H1.
exists (nil (A := (ParamSym * Value))). simpl. trivial.

(* Case T-Cons (params) *)
unfold lemmaA29_prop2.
intros.
destruct (lemmaA29_aux_2 t0).
(* R-Tail (params) *)
assert (~ (params_evaluated xts)). intro. apply H3.
destruct H4. destruct H5.
exists ((x,x0) :: x1). simpl. rewrite H4. rewrite H5. trivial.
destruct (H1 H2 H5).
exists (ParamTerm_cons x t0 x0).
apply (R_ParamTermList_tail). trivial. trivial.
(* R-Head (params) *)
unfold lemmaA29_prop1 in H0.
destruct (H0 H2 H4).
exists (ParamTerm_cons x x0 xts).
apply (R_ParamTermList_head). trivial.

(* Case T-Nil (fields) *)
unfold lemmaA29_prop3.
intros.
cut False. tauto.
apply H1.
exists (FieldValue_nil). simpl. trivial.

(* Case T-Cons (fields) *)
unfold lemmaA29_prop3.
intros.
destruct (lemmaA29_aux_2 t0).
(* R-Tail (fields) *)
assert (~ (fields_evaluated fts)). intro. apply H3.
destruct H4. destruct H5.
exists (FieldValue_cons f x x0). simpl. rewrite H4. rewrite H5. trivial.
destruct (H1 H2 H5).
exists (FieldTerm_cons f t0 x).
apply (R_FieldTermList_tail). trivial. trivial.
(* R-Head (params) *)
unfold lemmaA29_prop1 in H0.
destruct (H0 H2 H4).
exists (FieldTerm_cons f x fts).
apply (R_FieldTermList_head). trivial.
Qed.

End SetProgram.
