(* Author: Vincent Cremet *)
(* Date: August 2006 *) 
(* File: Proofs_SubstitutionLemmas.v *)
(* Description: Substitution lemmas (Appendix 7) *)

(** * Proofs about Substitution for the Current Instance *)

Require Import Calculus.
Require Import Arith.
Require Import MatchCompNat.
Require Import Util.
Require Import List.
Require Import ListSet.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_DeBruijnSIndices.v *)
Require Proofs_DeBruijnSIndices.
Module My_Proofs_DeBruijnSIndices := Proofs_DeBruijnSIndices.SetProgram(My_Program).

Import My_Proofs_DeBruijnSIndices.
Import My_Proofs_Progress.
Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)
(* Well-formed contexts *)
(*****************************************************************)

(** Predicate: [P] is true for all elements of the list [xs] **)
Fixpoint list_forall (A: Set) (xs: (list A)) (P: A -> Prop) {struct xs}: Prop :=
  match xs with
    | nil      => True
    | x :: xs' => (P x) /\ (list_forall A xs' P)
  end
.

Lemma listForall_is_forall:
  forall (A: Set) (xs: (list A)) (P: A -> Prop),
    (list_forall A xs P) -> (forall (x: A), (In x xs) -> (P x))
.

Proof.
intros A xs P.
induction xs.
simpl. tauto.
simpl.
intros.
destruct H.
destruct H0.
rewrite <- H0. trivial.
apply IHxs. trivial. trivial.
Qed.

Lemma forall_is_listForall:
  forall (A: Set) (xs: (list A)) (P: A -> Prop),
    (forall (x: A), (In x xs) -> (P x)) -> (list_forall A xs P)
.

Proof.
intros A xs P.
induction xs.
simpl. tauto.
simpl.
intros.
split. apply H. left. trivial.
apply IHxs.
intros.
apply H. right. trivial.
Qed.

(*****************************************************************)

(** Predicate: for all i in [0,|Ts|-1], for all k in fv(Ts_i), k < i **)

Fixpoint local_types_well_formed_aux (Ts: TypeList) (i: nat) {struct Ts}: Prop :=
  match Ts with
    | nil      => True
    | T :: Ts' =>
      (forall (k: VarSym), (var_in_type k T) -> (k < i)) /\
      (local_types_well_formed_aux Ts' (S i))
  end
.

Definition local_types_well_formed (Ts: TypeList): Prop :=
  (local_types_well_formed_aux Ts 0)
.

(** Well-formed contexts **)
Inductive context_is_well_formed: Env -> Prop :=

  | context_is_well_formed_intro:
    forall (O: Owner) (xTs: ParamTypeList) (Ts: TypeList),
    (list_forall _ xTs (fun xT => let (x, T) := xT in (fp_in_type T) = empty_paramset)) ->
    (local_types_well_formed Ts) ->
    (context_is_well_formed (Env_mk O xTs Ts))
.

(**************************************************************)

Definition lemmaA40_aux_4_prop1 (v: Value): Prop :=
  forall (G: Env) (T: type),
    (Gen_PathTyping G (Path_value v) T) ->
    (Typing G (value_to_term v) T)
.

Definition lemmaA40_aux_4_prop2 (fvs: FieldValueList): Prop :=
  forall (G: Env) (C: ClassSym),
    (Gen_FieldValueListTyping G C fvs) ->
    (FieldTermListTyping G C (fieldValueList_to_fieldTermList fvs))
.

Lemma lemmaA40_aux_4:
  forall (G: Env) (v: Value) (T: type),
    (Gen_PathTyping G (Path_value v) T) ->
    (Typing G (value_to_term v) T)
.

Proof.
cut (forall (v: Value), (lemmaA40_aux_4_prop1 v)).
intros. apply H. trivial.

apply (Value_prop
  lemmaA40_aux_4_prop1
  lemmaA40_aux_4_prop2
).

intros. unfold lemmaA40_aux_4_prop1. intros.
rename c into C; rename f into fvs.
simpl.
inversion_clear H0.
apply T_New.
Goal (
  forall (fvs: FieldValueList),
    (fields_of_fieldTermList (fieldValueList_to_fieldTermList fvs)) =
    (fields_of_fieldValueList fvs)
).
induction fvs.
(* Value *)
simpl. trivial.
simpl. rewrite IHfvs. trivial.
Save temp_lemma.
rewrite temp_lemma. trivial.
apply H. trivial.
rewrite temp_lemma. trivial. trivial.
(* Nil (field-value) *)
red. intros. simpl.
apply FieldTermListTyping_nil.
(* Cons (field-value) *)
red. intros. simpl.
rename f0 into fvs.
inversion_clear H1.
eapply FieldTermListTyping_cons.
apply H. apply H2. trivial. apply H4. trivial.
apply H0. trivial.
Qed.

(**************************************************************)

Definition lemmaA40_aux_8_prop1 (v: Value): Prop :=
  forall (G: Env) (T: type),
    (Typing G (value_to_term v) T) ->
    (Gen_PathTyping G (Path_value v) T)
.

Definition lemmaA40_aux_8_prop2 (fvs: FieldValueList): Prop :=
  forall (G: Env) (C: ClassSym),
    (FieldTermListTyping G C (fieldValueList_to_fieldTermList fvs)) ->
    (Gen_FieldValueListTyping G C fvs)
.

Lemma lemmaA40_aux_8:
  forall (G: Env) (v: Value) (T: type),
    (Typing G (value_to_term v) T) ->
    (Gen_PathTyping G (Path_value v) T)
.

Proof.
cut (forall (v: Value), (lemmaA40_aux_8_prop1 v)).
intros. apply H. trivial.

apply (Value_prop
  lemmaA40_aux_8_prop1
  lemmaA40_aux_8_prop2
).

intros. unfold lemmaA40_aux_8_prop1. intros.
rename c into C; rename f into fvs.
simpl in H0.
inversion_clear H0.
apply Gen_P_Value.
Goal (
  forall (fvs: FieldValueList),
    (fields_of_fieldTermList (fieldValueList_to_fieldTermList fvs)) =
    (fields_of_fieldValueList fvs)
).
induction fvs.
(* Value *)
simpl. trivial.
simpl. rewrite IHfvs. trivial.
Save temp_lemma1.
rewrite temp_lemma1 in H1. trivial.
apply H. trivial.
rewrite temp_lemma1 in H3. trivial. trivial.
(* Nil (field-value) *)
red. intros. simpl.
apply Gen_FieldValueListTyping_nil.
(* Cons (field-value) *)
red. intros. simpl.
rename f0 into fvs.
inversion_clear H1.
eapply Gen_FieldValueListTyping_cons.
apply H. apply H2. trivial. apply H4. trivial.
apply H0. trivial.
Qed.

(**************************************************************)

Lemma lemmaA40_aux_9:
  forall (G: Env) (p: Path) (T: type),
    (Gen_PathTyping G p T) ->
    (Typing G (path_to_term p) T)
.

Proof.
intros.
induction H.
(* P-This *)
eapply T_This. apply H. trivial.
(* P-Param *)
simpl.
eapply T_Param. apply H. apply H0. trivial.
(* P-Var *)
simpl.
eapply T_Var. apply H. apply H0. trivial.
(* P-Value *)
unfold path_to_term.
apply lemmaA40_aux_4.
apply Gen_P_Value. trivial. trivial. trivial. trivial.
Qed.

Lemma lemmaA40_aux_10:
  forall (G: Env) (p: Path) (T: type),
    (Typing G (path_to_term p) T) ->
    (Gen_PathTyping G p T)
.

Proof.
cut (
  forall (G: Env) (t: Term) (T: type),
    (Typing G t T) ->
    forall (p: Path),
      t = (path_to_term p) ->
      (Gen_PathTyping G p T)
).
intros. eapply H. apply H0. trivial.

intros G t T H.
induction H.
(* T-This *)
intros.
assert (p = Path_this). apply path_to_term_this. rewrite H1. trivial.
rewrite H2.
eapply Gen_P_This. apply H. trivial.
(* T-Param *)
intros.
assert (p = (Path_param x)). apply path_to_term_param. rewrite H2. trivial.
rewrite H3.
eapply Gen_P_Param. apply H. apply H0. trivial.
(* T-Var *)
intros.
assert (p = (Path_var n)). apply path_to_term_var. rewrite H2. trivial.
rewrite H3.
eapply Gen_P_Var. apply H. apply H0. trivial.
(* T-Select *)
intros.
cut False. tauto. eapply path_to_term_select. symmetry. apply H3.
(* T-Call *)
intros.
cut False. tauto. eapply path_to_term_call. symmetry. apply H5.
(* T-New *)
intros.
assert ((path_to_term p) = (Term_new C fts)). symmetry. trivial. clear H3.
destruct (path_to_term_new p C fts H4). destruct H3.
rewrite H3.
apply lemmaA40_aux_8. destruct x.
simpl. simpl in H5. injection H5; intros. clear H5.
apply T_New. rewrite H6. trivial. rewrite H6. rewrite H7. trivial.
rewrite H6. rewrite H7. trivial. rewrite H7. trivial.
(* Let *)
intros.
cut False. tauto. eapply path_to_term_let. symmetry. apply H4.
Qed.

(*****************************************************************)
(* Weakening *)
(*****************************************************************)

Definition lemmaA38_0_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (xTs: ParamTypeList) (Ts Us: TypeList),
    G = (Env_mk Owner_root nil Ts) ->
    (Gen_PathTyping (Env_mk Owner_root xTs (app Us Ts)) p T)
.

Definition lemmaA38_0_prop2 (G: Env) (C: ClassSym) (fvs: FieldValueList) (H: (Gen_FieldValueListTyping G C fvs)): Prop :=
  forall (xTs: ParamTypeList) (Ts Us: TypeList),
    G = (Env_mk Owner_root nil Ts) ->
    (Gen_FieldValueListTyping (Env_mk Owner_root xTs (app Us Ts)) C fvs)
.

Definition lemmaA38_0_prop3 (G: Env) (T U: type) (H: (Gen_Subtyping G T U)): Prop :=
  forall (xTs: ParamTypeList) (Ts Us: TypeList),
    G = (Env_mk Owner_root nil Ts) ->
    (Gen_Subtyping (Env_mk Owner_root xTs (app Us Ts)) T U)
.

Lemma lemmaA38_0:
  forall (xTs: ParamTypeList) (Ts Us: TypeList) (p: Path) (T: type),
    (Gen_PathTyping (Env_mk Owner_root nil Ts) p T) ->
    (Gen_PathTyping (Env_mk Owner_root xTs (app Us Ts)) p T)
.

Proof.
cut (
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA38_0_prop1 G p T H)
).
intros.
unfold lemmaA38_0_prop1 in H.
eapply H. apply H0. trivial.

apply (Gen_PathTyping_prop
  lemmaA38_0_prop1
  lemmaA38_0_prop3
  lemmaA38_0_prop2
).

(* P-This *)
intros. unfold lemmaA38_0_prop1. intros.
rewrite H0 in e. discriminate e.
(* P-Param *)
intros. unfold lemmaA38_0_prop1. intros.
rewrite H0 in e. inversion e.
rewrite <- H3 in i. destruct i.
(* P-Var*)
intros. unfold lemmaA38_0_prop1. intros.
destruct a.
rewrite e in H0. injection H0; intros. clear H0.
rewrite H4 in e. rewrite H5 in e. rewrite <- H3.
clear H4 xTs. rename xTs0 into xTs.
apply Gen_P_Var with
  (O := Owner_root)
  (xTs := xTs)
  (Ts := (Us ++ Ts))
  (T := T)
.
trivial.
rewrite length_of_append.
split.
omega.
rewrite plus_minus_assoc.
rewrite plus_minus_assoc.
apply typeList_at_append. trivial.
omega. omega.
apply H. trivial.
(* P-Value *)
intros. unfold lemmaA38_0_prop1.
intros.
apply Gen_P_Value with
  (G := (Env_mk Owner_root xTs (Us ++ Ts))).
trivial.
apply H. trivial. trivial.
apply H0. trivial.
(* S-Class *)
intros. unfold lemmaA38_0_prop3.
intros.
apply Gen_S_Class.
(* S-Extends *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Extends. apply e.
apply H. trivial.
(* S-Up *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Up.
apply H. trivial. trivial. apply e0.
apply H0. trivial.
(* S-Down *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Down.
apply H. trivial. trivial. apply e0.
apply H0. trivial.
(* S-Alias-Left *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Alias_Left.
apply H. trivial. trivial. apply e.
apply H0. trivial.
(* S-Alias-Right *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Alias_Right.
apply H. trivial. trivial. apply e.
apply H0. trivial.
(* S-Virtual *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Virtual.
apply H. trivial. trivial.
(* S-Trans *)
intros. unfold lemmaA38_0_prop3.
intros.
eapply Gen_S_Trans.
apply H. trivial. trivial.
apply H0. trivial. trivial.
(* T-Nil (field/value)  *)
intros. unfold lemmaA38_0_prop2.
intros.
apply Gen_FieldValueListTyping_nil.
(* T-Cons (field/value)  *)
intros. unfold lemmaA38_0_prop2.
intros.
eapply Gen_FieldValueListTyping_cons.
apply H. trivial. trivial. apply e. trivial.
apply H0. trivial.
Qed.

Lemma lemmaA38_2:
  forall (xTs: ParamTypeList) (Ts Us: TypeList) (T U: type),
    (Gen_Subtyping (Env_mk Owner_root nil Ts) T U) ->
    (Gen_Subtyping (Env_mk Owner_root xTs (app Us Ts)) T U)
.

Proof.
cut (
  forall (G: Env) (T U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA38_0_prop3 G T U H)
).
intros. unfold lemmaA38_0_prop3 in H. eapply H.
apply H0. trivial.
intros G T U H.
unfold lemmaA38_0_prop3.
induction H.

(* S-Class *)
intros. 
apply Gen_S_Class.
(* S-Extends *)
intros.
eapply Gen_S_Extends. apply H.
apply IHGen_Subtyping. trivial.
(* S-Up *)
intros.
eapply Gen_S_Up with (C := C).
apply lemmaA38_0.
rewrite H3 in H. trivial. trivial. apply H1.
apply IHGen_Subtyping. trivial.
(* S-Down *)
intros.
eapply Gen_S_Down with (C := C).
apply lemmaA38_0.
rewrite H3 in H. trivial. trivial. apply H1.
apply IHGen_Subtyping. trivial.
(* S-Alias-Left *)
intros.
eapply Gen_S_Alias_Left with (C := C).
apply lemmaA38_0.
rewrite H2 in H. trivial. apply H0.
apply IHGen_Subtyping. trivial.
(* S-Alias-Right *)
intros.
eapply Gen_S_Alias_Right with (C := C).
apply lemmaA38_0.
rewrite H2 in H. trivial. apply H0.
apply IHGen_Subtyping. trivial.
(* S-Virtual *)
intros.
apply Gen_S_Virtual with (C := C).
apply lemmaA38_0.
rewrite H1 in H. trivial. trivial.
(* S-Trans *)
intros.
rewrite H2 in H.
eapply Gen_S_Trans.
apply IHGen_Subtyping1. trivial.
apply IHGen_Subtyping2. trivial. trivial.
Qed.

(* Lemma lemmaA38_1: *)
(*   forall (xTs: ParamTypeList) (Ts Us: TypeList) (t: Term) (T: type), *)
(*     (Typing (Env_mk Owner_root nil Ts) t T) -> *)
(*     (Typing (Env_mk Owner_root xTs (app Us Ts)) t T) *)
(* . *)

(* Proof. *)
(* Admitted. *)

Lemma lemmaA38_0_corollary:
  forall (xTs: ParamTypeList) (Ts: TypeList) (p: Path) (T: type),
    (Gen_PathTyping rootEnv p T) ->
    (Gen_PathTyping (Env_mk Owner_root xTs Ts) p T)
.

Proof.
intros.
replace (Ts) with (Ts ++ nil).
apply lemmaA38_0. trivial.
rewrite <- app_nil_end. trivial.
Qed.

(*****************************************************************)
(* Substitution lemmas -- This *)
(*****************************************************************)

(** Substitution for this in a list of types **)
Definition thisSubst_in_typeList (Ts: TypeList) (p: Path): TypeList :=
  (map (fun T => (thisSubst_in_type T p)) Ts)
.

(** Substitution for this in a list of bindings (Param * Type) **)
Definition thisSubst_in_paramTypeList (xTs: ParamTypeList) (p: Path): ParamTypeList :=
  (map (fun xT => match xT with (x, T) => (x, (thisSubst_in_type T p)) end) xTs)
.
(*****************************************************************)

Lemma lemmaA1_for_paths:
  forall (G: Env) (p: Path) (T U: type),
    (Gen_PathTyping G p T) ->
    (Gen_Subtyping G T U) ->
    (Gen_PathTyping G p U)
.

Proof.
intros.
apply lemmaA40_aux_10.
apply lemmaA1 with (T := T).
apply lemmaA40_aux_9. trivial.
trivial.
Qed.

(*****************************************************************)

Definition lemmaA39_this_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Gen_PathTyping (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_path p (Path_value v)) (thisSubst_in_type T (Path_value v))
      )
.

Definition lemmaA39_this_prop2 (G: Env) (T U: type) (H: (Gen_Subtyping G T U)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Gen_Subtyping (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_type T (Path_value v)) (thisSubst_in_type U (Path_value v))
      )
.

Definition lemmaA39_this_prop3 (G: Env) (A: ClassSym) (fvs: FieldValueList) (H: (Gen_FieldValueListTyping G A fvs)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Gen_FieldValueListTyping (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        A fvs
      )
.

Lemma lemmaA39_this_main:
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA39_this_prop1 G p T H)
.

Proof.
apply (Gen_PathTyping_prop
  lemmaA39_this_prop1
  lemmaA39_this_prop2
  lemmaA39_this_prop3
).
(* P-This *)
intros. unfold lemmaA39_this_prop1. intros.
rewrite e in H0. injection H0; intros. clear H0.
rewrite <- H3. rewrite <- H4 in H1. rewrite <- H2. clear H2 H3 H4.
apply lemmaA1_for_paths with
  (T := (thisSubst_in_type (type_class C) (Path_value v))).
simpl.
apply lemmaA38_0_corollary. trivial.
unfold lemmaA39_this_prop2 in H. eapply H. apply e.
trivial.
(* P-Param *)
intros. unfold lemmaA39_this_prop1. intros.
rewrite e in H0. injection H0; intros. clear H0.
rewrite <- H2. rewrite <- H3. rewrite H4 in e. clear H2 H3 H4 O xTs0 Us.
simpl.
eapply Gen_P_Param with
  (O := Owner_root)
  (xTs := (thisSubst_in_paramTypeList xTs (Path_value v)))
  (Ts := (thisSubst_in_typeList Ts (Path_value v)))
  (T := (thisSubst_in_type T (Path_value v)))
. trivial.
unfold thisSubst_in_paramTypeList.
pose (f := (fun xT : ParamSym * type =>
  let (x0, T0) := xT in
    (x0, thisSubst_in_type T0 (Path_value v)))
).
fold (f (x,T)).
fold f.
apply map_intro. trivial.
unfold lemmaA39_this_prop2 in H. eapply H.
apply e. trivial.
(* P-Var *)
intros. unfold lemmaA39_this_prop1. intros.
rewrite e in H0. injection H0; intros. clear H0.
rewrite <- H2. rewrite <- H3. rewrite H4 in e. clear H2 H3 H4 O xTs0 Us.
simpl.
eapply Gen_P_Var with
  (O := Owner_root)
  (xTs := (thisSubst_in_paramTypeList xTs (Path_value v)))
  (Ts := (thisSubst_in_typeList Ts (Path_value v)))
  (T := (thisSubst_in_type T (Path_value v)))
. trivial.
unfold thisSubst_in_typeList.
pose (f :=
  (fun T0 : type => thisSubst_in_type T0 (Path_value v))
).
fold f. fold (f T).
rewrite length_of_map.
split. tauto.
apply typeList_at_map. tauto.
rewrite lemmaA33_2.
unfold lemmaA39_this_prop2 in H. eapply H.
apply e. trivial.
(* P-Value *)
intros. unfold lemmaA39_this_prop1. intros.
rename C0 into B.
simpl.
eapply Gen_P_Value.
trivial.
unfold lemmaA39_this_prop3 in H. eapply H. apply H1. trivial.
trivial.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop2 in H0. eapply H0.
apply H1. trivial.
(* S-Class *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
apply Gen_S_Class.
(* S-Extends *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Extends. apply e.
fold (thisSubst_in_type (type_class C') (Path_value v)).
unfold lemmaA39_this_prop2 in H. eapply H.
apply H0. trivial.
(* S-Up *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Up with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply H. apply H1. trivial. trivial. apply e0.
rewrite <- lemmaA8_2.
eapply H0. apply H1. trivial.
(* S-Down *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Down with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply H. apply H1. trivial. trivial. apply e0.
rewrite <- lemmaA8_2.
eapply H0. apply H1. trivial.
(* S-Alias-Left *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Alias_Left with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply H. apply H1. trivial. apply e.
rewrite <- lemmaA8_2.
eapply H0. apply H1. trivial.
(* S-Alias_Right *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Alias_Right with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply H. apply H1. trivial. apply e.
rewrite <- lemmaA8_2.
eapply H0. apply H1. trivial.
(* S-Virtual *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Virtual with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply H. apply H0. trivial. trivial.
(* S-Trans *)
intros. unfold lemmaA39_this_prop2. intros.
simpl.
eapply Gen_S_Trans with
  (S := (thisSubst_in_type S (Path_value v)))
.
eapply H. apply H1. trivial.
eapply H0. apply H1. trivial.
tauto.
(* T-Nil (field/value) *)
intros. unfold lemmaA39_this_prop3. intros.
apply Gen_FieldValueListTyping_nil.
(* T-Cons (field/value) *)
intros. unfold lemmaA39_this_prop3. intros.
assert (~(this_in_type U)).
unfold type_is_closed in t. tauto.
eapply Gen_FieldValueListTyping_cons with
  (U := (thisSubst_in_type U (Path_value v0)))
.
fold (thisSubst_in_path (Path_value v) (Path_value v0)).
eapply H. apply H1. trivial.
rewrite new_lemma1_2. trivial. trivial. apply e.
rewrite new_lemma1_2. trivial. trivial.
eapply H0. apply H1. trivial.
Qed.

(** Subtitution lemmas -- path **)
Lemma lemmaA39_this_0:
    forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList) (p: Path) (T: type),
    (Gen_PathTyping (Env_mk (Owner_class C) xTs Us) p T) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Gen_PathTyping (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_path p (Path_value v)) (thisSubst_in_type T (Path_value v))
      )
.

Proof.
pose (H := lemmaA39_this_main).
intros.
unfold lemmaA39_this_prop1 in H.
eapply H with (C := C). apply H0. trivial. trivial.
Qed.

(** Subtitution lemmas -- type **)
Lemma lemmaA39_this_2:
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList) (T U: type),
    (Gen_Subtyping (Env_mk (Owner_class C) xTs Us) T U) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Gen_Subtyping (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_type T (Path_value v)) (thisSubst_in_type U (Path_value v))
      )
.

Proof.
cut (
  forall (G: Env) (T U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA39_this_prop2 G T U H)
).
intros.
unfold lemmaA39_this_prop2 in H. eapply H with (C := C). apply H0.
trivial. trivial.
pose (H := lemmaA39_this_main).
intros.
unfold lemmaA39_this_prop2.
induction H0.

(* S-Class *)
intros.
simpl.
apply Gen_S_Class.
(* S-Extends *)
intros.
simpl.
eapply Gen_S_Extends. apply H0.
fold (thisSubst_in_type (type_class C') (Path_value v)).
eapply IHGen_Subtyping.
apply H2. trivial.
(* S-Up *)
intros.
simpl.
eapply Gen_S_Up with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop1 in H.
eapply H.
apply H0. eapply H4. trivial. trivial. apply H2.
rewrite <- lemmaA8_2.
eapply IHGen_Subtyping. apply H4. trivial.
(* S-Down *)
intros.
simpl.
eapply Gen_S_Down with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop1 in H.
eapply H.
apply H0. eapply H4. trivial. trivial. apply H2.
rewrite <- lemmaA8_2.
eapply IHGen_Subtyping. apply H4. trivial.
(* S-Alias-Left *)
intros.
simpl.
eapply Gen_S_Alias_Left with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop1 in H.
eapply H. apply H0. apply H3. trivial. apply H1.
rewrite <- lemmaA8_2.
eapply IHGen_Subtyping. apply H3. trivial.
(* S-Alias_Right *)
intros.
simpl.
eapply Gen_S_Alias_Right with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop1 in H.
eapply H. apply H0. apply H3. trivial. apply H1.
rewrite <- lemmaA8_2.
eapply IHGen_Subtyping. apply H3. trivial.
(* S-Virtual *)
intros.
simpl.
eapply Gen_S_Virtual with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
unfold lemmaA39_this_prop1 in H.
eapply H. apply H0. apply H2. trivial. trivial.
(* S-Trans *)
intros.
eapply Gen_S_Trans with
  (S := (thisSubst_in_type S (Path_value v)))
.
eapply IHGen_Subtyping1. apply H1. trivial.
eapply IHGen_Subtyping2. apply H1. trivial.
tauto.
Qed.

(*****************************************************************)

Definition lemmaA39_this1_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Typing
        (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v))
        (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_term t (Path_value v)) (thisSubst_in_type T (Path_value v))
      )
.

Definition lemmaA39_this1_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList)
  (H: (ParamTermListTyping G p m xts)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (ParamTermListTyping
        (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_path p (Path_value v))
        m
        (thisSubst_in_paramTermList xts (Path_value v))
      )
.

Definition lemmaA39_this1_prop3 (G: Env) (A: ClassSym) (fts: FieldTermList)
  (H: (FieldTermListTyping G A fts)): Prop :=
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList),
    G = (Env_mk (Owner_class C) xTs Us) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (FieldTermListTyping
        (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        A
        (thisSubst_in_fieldTermList fts (Path_value v))
      )
.

Lemma lemmaA39_this1_aux_1:
  forall (xts: ParamTermList) (p: Path),
    (params_of_paramTermList (thisSubst_in_paramTermList xts p)) =
    (params_of_paramTermList xts)
.

Proof.
intros. induction xts.
simpl. trivial.
simpl. rewrite IHxts. trivial.
Qed.

Lemma lemmaA39_this1_aux_2:
  forall (fts: FieldTermList) (p: Path),
    (fields_of_fieldTermList (thisSubst_in_fieldTermList fts p)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros. induction fts.
simpl. trivial.
simpl. rewrite IHfts. trivial.
Qed.

(** Subtitution lemmas -- term **)
Lemma lemmaA39_this_1:
  forall (C: ClassSym) (xTs: ParamTypeList) (Us: TypeList) (t: Term) (T: type),
    (Typing (Env_mk (Owner_class C) xTs Us) t T) ->
    forall (v: Value),
      (Gen_PathTyping rootEnv (Path_value v) (type_class C)) ->
      (Typing
        (Env_mk Owner_root (thisSubst_in_paramTypeList xTs (Path_value v)) (thisSubst_in_typeList Us (Path_value v)))
        (thisSubst_in_term t (Path_value v))
        (thisSubst_in_type T (Path_value v))
      )
.

Proof.
cut (
  forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
    (lemmaA39_this1_prop1 G t T H)
).
intros.
unfold lemmaA39_this1_prop1 in H. eapply H with (C := C). apply H0.
trivial. trivial.
apply (Typing_prop
  lemmaA39_this1_prop1
  lemmaA39_this1_prop2
  lemmaA39_this1_prop3
).

(* T-This *)
intros. unfold lemmaA39_this1_prop1. intros.
simpl.
apply lemmaA40_aux_4.
rewrite e in H. injection H; intros. clear H.
rewrite <- H1. rewrite <- H2. rewrite <- H3 in H0. clear H1 H2 H3.
apply lemmaA1_for_paths with
  (T := (thisSubst_in_type (type_class C) (Path_value v))).
simpl.
apply lemmaA38_0_corollary.
trivial.
eapply lemmaA39_this_2 with (C := C). rewrite <- e.
trivial. trivial.
(* T-Param *)
intros. unfold lemmaA39_this1_prop1. intros.
simpl.
rewrite e in H. injection H; intros. clear H.
rewrite <- H1. rewrite <- H2. rewrite H3 in e. clear H1 H2 H3 O xTs0 Us.
eapply T_Param with
  (O := Owner_root)
  (xTs := (thisSubst_in_paramTypeList xTs (Path_value v)))
  (Ts := (thisSubst_in_typeList Ts (Path_value v)))
  (T := (thisSubst_in_type T (Path_value v)))
. trivial.
unfold thisSubst_in_paramTypeList.
pose (f := (fun xT : ParamSym * type =>
  let (x0, T0) := xT in
    (x0, thisSubst_in_type T0 (Path_value v)))
).
fold (f (x,T)).
fold f.
apply map_intro. trivial.
eapply lemmaA39_this_2 with (C := C).
rewrite <- e. trivial.
trivial.
(* T-Var *)
intros. unfold lemmaA39_this1_prop1. intros.
rewrite e in H. injection H; intros. clear H.
rewrite <- H1. rewrite <- H2. rewrite H3 in e. clear H1 H2 H3 O xTs0 Us.
simpl.
eapply T_Var with
  (O := Owner_root)
  (xTs := (thisSubst_in_paramTypeList xTs (Path_value v)))
  (Ts := (thisSubst_in_typeList Ts (Path_value v)))
  (T := (thisSubst_in_type T (Path_value v)))
. trivial.
unfold thisSubst_in_typeList.
pose (f :=
  (fun T0 : type => thisSubst_in_type T0 (Path_value v))
).
fold f. fold (f T).
rewrite length_of_map.
split. tauto.
apply typeList_at_map. tauto.
rewrite lemmaA33_2.
eapply lemmaA39_this_2 with (C := C).
rewrite <- e. trivial.
trivial.
(* T-Select *)
intros. unfold lemmaA39_this1_prop1. intros.
simpl.
eapply T_Select with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply lemmaA39_this_0 with
  (C := C0)
. rewrite <- H. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA8_2.
eapply lemmaA39_this_2 with
  (C := C0)
. rewrite <- H. trivial. trivial.
(* T-Call *)
intros. unfold lemmaA39_this1_prop1. intros.
simpl.
eapply T_Call with
  (C := C)
.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply lemmaA39_this_0 with
  (C := C0)
. rewrite <- H0. trivial. trivial. trivial. apply e0.
rewrite lemmaA39_this1_aux_1. trivial.
eapply H. apply H0. trivial.
rewrite <- lemmaA8_2.
eapply lemmaA39_this_2 with
  (C := C0)
. rewrite <- H0. trivial. trivial.
(* T-New *)
intros. unfold lemmaA39_this1_prop1. intros.
rename C0 into B.
simpl.
eapply T_New.
rewrite lemmaA39_this1_aux_2. trivial.
eapply H. apply H0. trivial.
rewrite lemmaA39_this1_aux_2. trivial.
fold (thisSubst_in_type (type_class C) (Path_value v)).
eapply lemmaA39_this_2 with (C := B).
rewrite <- H0. trivial. trivial.
(* T-Let *)
intros. unfold lemmaA39_this1_prop1. intros.
simpl.
rewrite e in H1. injection H1; intros. clear H1.
rewrite H5 in e.
rewrite <- H3. rewrite <- H4. clear H3 H4 Us xTs0.
eapply T_Let with
  (O := Owner_root)
  (xTs := (thisSubst_in_paramTypeList xTs (Path_value v)))
  (Ts :=  (thisSubst_in_typeList Ts (Path_value v)))
  (U := (thisSubst_in_type U (Path_value v)))
. trivial.
eapply H. apply e. trivial.
replace
  (thisSubst_in_type T (Path_value v) :: nil) 
  with
  (thisSubst_in_typeList (T :: nil) (Path_value v))
.
unfold thisSubst_in_typeList.
rewrite <- map_of_append.
fold (thisSubst_in_typeList (Ts ++ T :: nil) (Path_value v)).
eapply H0 with (C := C). rewrite H5. trivial.
trivial. simpl. trivial.
(* 0 notin fv(U[v]) *)
assert (~(varset_In 0 (fv_in_type U))).
intro. apply n.
apply fv_and_var_in_type2. trivial.
assert (~(varset_In 0 (fv_in_type (thisSubst_in_type U (Path_value v))))).
rewrite fv_and_thisSubst_in_type. trivial.
intro. apply H3.
apply fv_and_var_in_type1. trivial.
(* *)
rewrite lemmaA36_33_2.
eapply lemmaA39_this_2 with (C := C).
rewrite <- e. trivial. trivial.
(* T-Nil (param/term) *)
intros. unfold lemmaA39_this1_prop2. intros.
simpl.
apply ParamTermListTyping_nil.
(* T-Cons (param/term) *)
intros. unfold lemmaA39_this1_prop2. intros.
simpl.
eapply ParamTermListTyping_cons. apply e.
rewrite <- lemmaA8_2.
eapply H. apply H1. trivial.
eapply H0. apply H1. trivial.
(* T-Nil (field/term) *)
intros. unfold lemmaA39_this1_prop3. intros.
simpl.
apply FieldTermListTyping_nil.
(* T-Cons (field/term) *)
intros. unfold lemmaA39_this1_prop3. intros.
simpl.
assert (~(this_in_type U)).
unfold type_is_closed in t1. tauto.
eapply FieldTermListTyping_cons with
  (U := (thisSubst_in_type U (Path_value v)))
.
eapply H. apply H1. trivial.
rewrite new_lemma1_2. trivial. trivial.
apply e.
rewrite new_lemma1_2. trivial. trivial.
eapply H0. apply H1. trivial.
Qed.

End SetProgram.
