(* Author: Vincent Cremet *)
(* Date: August 2006 *) 
(* File: Proofs_SubstitutionLemmas2.v *)
(* Description: Substitution lemmas (Appendix 7) *)

(** * Proofs about Substitution for Parameters *)

Require Import Calculus.
Require Import Arith.
Require Import MatchCompNat.
Require Import Util.
Require Import List.
Require Import ListSet.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_SubstitutionLemmas.v *)
Require Proofs_SubstitutionLemmas.
Module My_Proofs_SubstitutionLemmas := Proofs_SubstitutionLemmas.SetProgram(My_Program).

Import My_Proofs_SubstitutionLemmas.
Import My_Proofs_DeBruijnSIndices.
Import My_Proofs_Progress.
Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(*****************************************************************)
(* Substitution lemmas -- Parameters *)
(*****************************************************************)

(** Substitution for this in a list of types **)
Definition map_paramSubst_in_typeList (Ts: TypeList) (xvs: ParamValueList): TypeList :=
  (map (fun T => (map_paramSubst_in_type T xvs)) Ts)
.

(** Predicate: list [ys] contains, at least, all elements of [xs] **)
Definition List_sublist (A: Set) (xs ys: (list A)): Prop :=
  forall (x: A), (In x xs) -> (In x ys)
.

(** Predicate: bindings of [xvs] conforms to all corresponding bindings in [xTs] **)
Definition args_are_correct (xvs: ParamValueList) (xTs: ParamTypeList) :=
  forall (x: ParamSym) (v: Value) (T: type),
    (In (x, v) xvs) ->
    (In (x, T) xTs) ->
    (Gen_PathTyping rootEnv (Path_value v) T)
.

Definition params_of_paramValueList (xvs: ParamValueList): (list ParamSym) :=
  (map (fun (xv: (ParamSym * Value)) => let (x, v) := xv in x) xvs)
. 

Definition params_of_paramTypeList (xTs: ParamTypeList): (list ParamSym) :=
  (map (fun (xT: (ParamSym * type)) => let (x, T) := xT in x) xTs)
. 

(** Predicate: a list of arguments [xvs] conforms to a list of typed parameters [xTs] **)
Definition args_conform_to_params (xvs: ParamValueList) (xTs: ParamTypeList): Prop :=
  (List_sublist _ (params_of_paramTypeList xTs) (params_of_paramValueList xvs)) /\
  (args_are_correct xvs xTs)
.

Definition lemmaA39_param_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Gen_PathTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_path p xvs) (map_paramSubst_in_type T xvs)
      )
.

Definition lemmaA39_param_prop2 (G: Env) (T U: type) (H: (Gen_Subtyping G T U)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Gen_Subtyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_type T xvs) (map_paramSubst_in_type U xvs)
      )
.

Definition lemmaA39_param_prop3 (G: Env) (C: ClassSym) (fvs: FieldValueList)
  (H: (Gen_FieldValueListTyping G C fvs)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Gen_FieldValueListTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        C fvs
      )
.

Lemma lemmaA39_param_main:
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA39_param_prop1 G p T H)
.

Proof.
apply (Gen_PathTyping_prop
  lemmaA39_param_prop1
  lemmaA39_param_prop2
  lemmaA39_param_prop3
).

(* Case (P-This): *)
intros. unfold lemmaA39_param_prop1. intros.
rewrite e in H0. discriminate H0.
(* Case (P-Param): *)
intros. unfold lemmaA39_param_prop1. intros.
rewrite e in H0. injection H0; intros; clear H0.
rewrite <- H3. clear Us H3. rewrite <- H4 in H2. clear xTs0 H4.
rewrite H5 in e. clear O H5.
assert (V : (args_conform_to_params xvs xTs)). trivial.
unfold args_conform_to_params in H2.
destruct H2.
unfold List_sublist in H0.
unfold args_are_correct in H2.
(* (x, v) belongs to xvs *)
assert (In x (params_of_paramValueList xvs)).
apply H0.
unfold params_of_paramTypeList.
pose (f := (fun xT : ParamSym * type => let (x0, _) := xT in x0)).
fold f. fold (f (x, T)).
apply map_intro. trivial. clear H0.
destruct (polySubst_in_pathParam xvs x).
cut False. tauto.
unfold params_of_paramValueList in H3.
pose (f := (fun xv : ParamSym * Value => let (x, _) := xv in x)).
fold f in H3.
destruct (map_elim (ParamSym * Value) ParamSym f xvs x H3). rename x0 into y.
destruct H4. destruct y. simpl in H5. rewrite <- H5 in H4. clear p H5.
clear H3 f.
destruct H0. apply (H0 v). trivial.
destruct H0. destruct H0. rename x0 into v.
rewrite H4.
(* v: T *)
assert (Gen_PathTyping rootEnv (Path_value v) T).
apply H2 with (x := x). trivial. trivial. clear H2.
(* subsumption *)
apply lemmaA1_for_paths with (T := T).
(* Ts[xvs] |- v: T *)
eapply lemmaA38_0_corollary. trivial.
(* T = T[xvs] *)
assert ((map_paramSubst_in_type T xvs) = T).
  (* fv(T) = empty_paramset *)
destruct H1. clear H2. injection e; intros; clear e.
rewrite H6 in H1.
pose (f := (fun xT : ParamSym * type =>
          let (_, T) := xT in fp_in_type T = empty_paramset)).
fold f in H1.
assert (forall (xT: (ParamSym * type)), (In xT xTs) -> (f xT)).
apply listForall_is_forall. trivial. clear H1.
assert ((fp_in_type T) = empty_paramset).
apply (H8 (x, T)). trivial.
apply fp_and_paramSubst_in_type. trivial.
(* Ts[xvs] |- T <: S[xvs] *)
rewrite <- H2.
eapply H. apply e. trivial. trivial.
(* Case (P-Var) *)
intros. unfold lemmaA39_param_prop1. intros.
rewrite polySubst_in_pathVar.
rewrite e in H0. injection H0; intros; clear H0.
rewrite <- H3. clear Us H3. rewrite <- H4 in H2. clear xTs0 H4.
rewrite H5 in e. clear O H5.
destruct a.
eapply Gen_P_Var with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_paramSubst_in_typeList Ts xvs))
  (T := (map_paramSubst_in_type T xvs))
. trivial.
split.
unfold map_paramSubst_in_typeList.
rewrite length_of_map. trivial.
unfold map_paramSubst_in_typeList.
pose (f := (fun T0 : type => map_paramSubst_in_type T0 xvs)).
fold f. fold (f T).
rewrite length_of_map.
apply typeList_at_map. trivial.
rewrite lemmaA34_2.
eapply H. apply e. trivial. trivial.
(* Case (P-Value) *)
intros. unfold lemmaA39_param_prop1. intros.
rewrite polySubst_in_pathValue.
eapply Gen_P_Value. trivial.
eapply H. apply H1. trivial. trivial. trivial.
rewrite <- (polySubst_in_typeClass xvs C).
eapply H0. apply H1. trivial. trivial.
(* S-Class *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeClass.
apply Gen_S_Class.
(* S-Extends *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeClass.
apply Gen_S_Extends with (C' := C'). trivial.
rewrite <- (polySubst_in_typeClass xvs C').
eapply H. apply H0. trivial. trivial.
(* S-Up *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Up with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply H. apply H1. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA9_2.
eapply H0. apply H1. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeUpperBounds_are_well_formed. apply e. trivial.
(* S-Down *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Down with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply H. apply H1. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA9_2.
eapply H0. apply H1. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeLowerBounds_are_well_formed. apply e. trivial.
(* S-Alias-Left *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Alias_Left with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply H. apply H1. trivial. trivial. apply e.
rewrite <- lemmaA9_2.
eapply H0. apply H1. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeValues_are_well_formed. apply e.
(* S-Alias-Right *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Alias_Right with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply H. apply H1. trivial. trivial. apply e.
rewrite <- lemmaA9_2.
eapply H0. apply H1. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeValues_are_well_formed. apply e.
(* S-Virtual *)
intros. unfold lemmaA39_param_prop2. intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Virtual with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply H. apply H0. trivial. trivial. trivial.
(* S-Trans *)
intros. unfold lemmaA39_param_prop2. intros.
eapply Gen_S_Trans with (S := (map_paramSubst_in_type S xvs)).
eapply H. apply H1. trivial. trivial.
eapply H0. apply H1. trivial. trivial.
tauto.
(* Nil (field/value) *)
intros. unfold lemmaA39_param_prop3. intros.
apply Gen_FieldValueListTyping_nil.
(* Cons (field/value) *)
intros. unfold lemmaA39_param_prop3. intros.
eapply Gen_FieldValueListTyping_cons with
  (U := U)
.
rewrite <- (polySubst_in_pathValue xvs v).
unfold type_is_closed in t.
rewrite <- (fp_and_paramSubst_in_type U xvs).
eapply H. apply H1. trivial. trivial.
tauto. trivial. apply e. trivial.
eapply H0. apply H1. trivial. trivial.
Qed.

(** Subtitution lemmas -- path **)
Lemma lemmaA39_param_0:
  forall (G: Env) (xTs: ParamTypeList) (Us: TypeList) (p: Path) (T: type),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    (Gen_PathTyping G p T) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Gen_PathTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_path p xvs) (map_paramSubst_in_type T xvs)
      )
.

Proof.
pose (H := lemmaA39_param_main).
intros.
unfold lemmaA39_param_prop1 in H.
eapply H. apply H2. apply H0. trivial. trivial.
Qed.

(** Subtitution lemmas -- type **)
Lemma lemmaA39_param_2:
  forall (G: Env) (xTs: ParamTypeList) (Us: TypeList) (T U: type),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    (Gen_Subtyping G T U) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Gen_Subtyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_type T xvs) (map_paramSubst_in_type U xvs)
      )
.

Proof.
cut (
  forall (G: Env) (T U: type) (H: (Gen_Subtyping G T U)),
    (lemmaA39_param_prop2 G T U H)
).
intros.
unfold lemmaA39_param_prop2 in H.
eapply H. apply H2. apply H0. trivial. trivial.
intros.
unfold lemmaA39_param_prop2.
induction H.

(* S-Class *)
intros.
rewrite polySubst_in_typeClass.
apply Gen_S_Class.
(* S-Extends *)
intros.
rewrite polySubst_in_typeClass.
apply Gen_S_Extends with (C' := C'). trivial.
rewrite <- (polySubst_in_typeClass xvs C').
eapply IHGen_Subtyping. apply H1. trivial. trivial.
(* S-Up *)
intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Up with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H3. trivial. trivial. trivial. trivial. apply H1.
rewrite <- lemmaA9_2.
eapply IHGen_Subtyping. apply H3. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeUpperBounds_are_well_formed. apply H0. trivial.
(* S-Down *)
intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Down with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H3. trivial. trivial. trivial. trivial. apply H1.
rewrite <- lemmaA9_2.
eapply IHGen_Subtyping. apply H3. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeLowerBounds_are_well_formed. apply H0. trivial.
(* S-Alias-Left *)
intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Alias_Left with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H2. trivial. trivial. trivial. apply H0.
rewrite <- lemmaA9_2.
eapply IHGen_Subtyping. apply H2. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeValues_are_well_formed. apply H0.
(* S-Alias-Right *)
intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Alias_Right with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H2. trivial. trivial. trivial. apply H0.
rewrite <- lemmaA9_2.
eapply IHGen_Subtyping. apply H2. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply typeValues_are_well_formed. apply H0.
(* S-Virtual *)
intros.
rewrite polySubst_in_typeVirtual.
eapply Gen_S_Virtual with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H1. trivial. trivial. trivial. trivial.
(* S-Trans *)
intros.
eapply Gen_S_Trans with (S := (map_paramSubst_in_type S xvs)).
eapply IHGen_Subtyping1. apply H2. trivial. trivial.
eapply IHGen_Subtyping2. apply H2. trivial. trivial.
tauto.
Qed.

(*****************************************************************)

Lemma lemmaA39_param_aux_1:
  forall (xts: ParamTermList) (xvs: ParamValueList),
    (params_of_paramTermList (map_paramSubst_in_paramTermList xts xvs)) =
    (params_of_paramTermList xts)
.

Proof.
intros. induction xts.
simpl. rewrite polySubst_in_paramTermNil. simpl. trivial.
simpl. rewrite polySubst_in_paramTermCons. simpl.
rewrite IHxts. trivial.
Qed.

Lemma lemmaA39_param_aux_2:
  forall (fts: FieldTermList) (xvs: ParamValueList),
    (fields_of_fieldTermList (map_paramSubst_in_fieldTermList fts xvs)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros. induction fts.
simpl. rewrite polySubst_in_fieldTermNil. simpl. trivial.
simpl. rewrite polySubst_in_fieldTermCons. simpl.
rewrite IHfts. trivial.
Qed.

Lemma lemmaA39_param_aux_3:
  forall (Ts: TypeList) (i: nat),
    (local_types_well_formed_aux Ts i) ->
    (local_types_well_formed_aux Ts (S i))
.

Proof.
induction Ts.
simpl. intros. tauto.
simpl. intros.
destruct H.
split.
intros.
cut (k < i). omega. apply H. trivial.
apply IHTs. trivial.
Qed.

Lemma lemmaA39_param_aux_4:
  forall (Ts: TypeList) (T: type) (i: nat),
    (local_types_well_formed_aux Ts i) ->
    (forall (k: VarSym), (varset_In k (fv_in_type T)) -> k < (length Ts) + i) -> 
    (local_types_well_formed_aux (Ts ++ T :: nil) i)
.

Proof.
induction Ts.
simpl. intros.
split.
intros. apply H0.
apply fv_and_var_in_type1. trivial. tauto.
simpl. intros.
destruct H.
split.
trivial.
apply IHTs. trivial.
intros.
rewrite <- plus_Snm_nSm. simpl.
apply H0. trivial.
Qed.

Lemma lemmaA39_param_aux_4_corollary:
  forall (Ts: TypeList) (T: type),
    (local_types_well_formed Ts) ->
    (forall (k: VarSym), (varset_In k (fv_in_type T)) -> k < (length Ts)) -> 
    (local_types_well_formed (Ts ++ T :: nil))
.

Proof.
intros.
unfold local_types_well_formed.
apply lemmaA39_param_aux_4.
trivial.
intros. rewrite plus_comm. simpl.
apply H0. trivial.
Qed.

(*****************************************************************)

Definition lemmaA39_param1_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Typing (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_term t xvs) (map_paramSubst_in_type T xvs)
      )
.

Definition lemmaA39_param1_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList) (H: (ParamTermListTyping G p m xts)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (ParamTermListTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_path p xvs)
        m
        (map_paramSubst_in_paramTermList xts xvs)
      )
.

Definition lemmaA39_param1_prop3 (G: Env) (C: ClassSym) (fts: FieldTermList) (H: (FieldTermListTyping G C fts)): Prop :=
  forall (xTs: ParamTypeList) (Us: TypeList),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (FieldTermListTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        C (map_paramSubst_in_fieldTermList fts xvs)
      )
.

(** Subtitution lemmas -- term **)
Lemma lemmaA39_param_1:
  forall (G: Env) (xTs: ParamTypeList) (Us: TypeList) (t: Term) (T: type),
    (G = (Env_mk Owner_root xTs Us)) ->
    (context_is_well_formed G) ->
    (Typing G t T) ->
    forall (xvs: ParamValueList),
      (args_conform_to_params xvs xTs) ->
      (Typing (Env_mk Owner_root nil (map_paramSubst_in_typeList Us xvs))
        (map_paramSubst_in_term t xvs) (map_paramSubst_in_type T xvs)
      )
.

Proof.
cut (forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
  (lemmaA39_param1_prop1 G t T H)
).
intros.
unfold lemmaA39_param1_prop1 in H.
eapply H. apply H2. apply H0. trivial. trivial.
apply (Typing_prop
  lemmaA39_param1_prop1
  lemmaA39_param1_prop2
  lemmaA39_param1_prop3
).

(* Case (T-This): *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite e in H. discriminate H.
(* Case (T-Param): *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite e in H. injection H; intros; clear H.
rewrite <- H2. clear Us H2. rewrite <- H3 in H1. clear xTs0 H3.
rewrite H4 in e. clear O H4.
assert (V : (args_conform_to_params xvs xTs)). trivial.
unfold args_conform_to_params in H1.
destruct H1.
unfold List_sublist in H.
unfold args_are_correct in H1.
(* (x, v) belongs to xvs *)
assert (In x (params_of_paramValueList xvs)).
apply H.
unfold params_of_paramTypeList.
pose (f := (fun xT : ParamSym * type => let (x0, _) := xT in x0)).
fold f. fold (f (x, T)).
apply map_intro. trivial. clear H.
destruct (polySubst_in_termParam xvs x).
cut False. tauto.
unfold params_of_paramValueList in H2.
pose (f := (fun xv : ParamSym * Value => let (x, _) := xv in x)).
fold f in H2.
destruct (map_elim (ParamSym * Value) ParamSym f xvs x H2). rename x0 into y.
destruct H3. destruct y. simpl in H4. rewrite <- H4 in H3. clear p H4.
clear H2 f.
destruct H. apply (H v). trivial.
destruct H. destruct H. rename x0 into v.
rewrite H3.
cut (
  (Gen_PathTyping (Env_mk Owner_root nil (map_paramSubst_in_typeList Ts xvs))
    (Path_value v) (map_paramSubst_in_type S xvs))
).
intros.
apply lemmaA40_aux_4. trivial.
(* v: T *)
assert (Gen_PathTyping rootEnv (Path_value v) T).
apply H1 with (x := x). trivial. trivial. clear H1.
(* subsumption *)
apply lemmaA1_for_paths with (T := T).
(* Ts[xvs] |- v: T *)
eapply lemmaA38_0_corollary. trivial.
(* T = T[xvs] *)
assert ((map_paramSubst_in_type T xvs) = T).
  (* fv(T) = empty_paramset *)
destruct H0. clear H1. injection e; intros; clear e.
rewrite H5 in H0.
pose (f := (fun xT : ParamSym * type =>
          let (_, T) := xT in fp_in_type T = empty_paramset)).
fold f in H0.
assert (forall (xT: (ParamSym * type)), (In xT xTs) -> (f xT)).
apply listForall_is_forall. trivial. clear H0.
assert ((fp_in_type T) = empty_paramset).
apply (H7 (x, T)). trivial.
apply fp_and_paramSubst_in_type. trivial.
(* Ts[xvs] |- T <: S[xvs] *)
rewrite <- H1.
eapply lemmaA39_param_2. apply e. trivial. trivial. trivial.
(* Case (T-Var) *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite polySubst_in_termVar.
rewrite e in H. injection H; intros; clear H.
rewrite <- H2. clear Us H2. rewrite <- H3 in H1. clear xTs0 H3.
rewrite H4 in e. clear O H4.
destruct a.
eapply T_Var with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_paramSubst_in_typeList Ts xvs))
  (T := (map_paramSubst_in_type T xvs))
. trivial.
split.
unfold map_paramSubst_in_typeList.
rewrite length_of_map. trivial.
unfold map_paramSubst_in_typeList.
pose (f := (fun T0 : type => map_paramSubst_in_type T0 xvs)).
fold f. fold (f T).
rewrite length_of_map.
apply typeList_at_map. trivial.
rewrite lemmaA34_2.
eapply lemmaA39_param_2. apply e. trivial. trivial. trivial.
(* Case (T-Select) *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite polySubst_in_termSelect.
eapply T_Select with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H. trivial. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA9_2.
eapply lemmaA39_param_2. apply H. trivial. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply fieldTypes_are_well_formed. apply e. trivial.
(* Case (T-Call) *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite polySubst_in_termCall.
eapply T_Call with (C := C).
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_0. apply H0. trivial. trivial. trivial. trivial. apply e0.
rewrite lemmaA39_param_aux_1. trivial.
eapply H. apply H0. trivial. trivial.
rewrite <- lemmaA9_2.
eapply lemmaA39_param_2. apply H0. trivial. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with (C := C) (x := x).
eapply methodTypes_are_well_formed. apply e. trivial.
(* Case (T-New) *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite polySubst_in_termNew.
eapply T_New.
rewrite lemmaA39_param_aux_2. trivial.
eapply H. apply H0. trivial. trivial.
rewrite lemmaA39_param_aux_2. trivial.
rewrite <- (polySubst_in_typeClass xvs C).
eapply lemmaA39_param_2.
apply H0. trivial. trivial. trivial.
(* Case (T-Let) *)
intros. unfold lemmaA39_param1_prop1. intros.
rewrite polySubst_in_termLet.
rewrite e in H1. injection H1; intros; clear H1.
rewrite <- H5 in H3. clear H5 xTs0.
rewrite <- H4. clear H4 Us.
eapply T_Let with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_paramSubst_in_typeList Ts xvs))
  (U := (map_paramSubst_in_type U xvs))
. trivial.
(* t *)
eapply H. rewrite <- H6. apply e. trivial. trivial.
unfold map_paramSubst_in_typeList.
pose (f := (fun T0 : type => map_paramSubst_in_type T0 xvs)).
fold f. fold (f T). fold (map f (T :: nil)).
rewrite <- map_of_append.
unfold f.
fold (map_paramSubst_in_typeList (Ts ++ (T :: nil)) xvs).
clear f.
unfold lemmaA39_param1_prop1 in H0.
rewrite H6 in H0.
eapply H0 with (xTs0 := xTs). trivial.
(* WF_ctx(root; xTs; Ts, T) *)
rewrite e in H2.
inversion_clear H2.
apply context_is_well_formed_intro.
trivial.
apply lemmaA39_param_aux_4_corollary.
trivial.
intros.
unfold lemmaA39_param1_prop1 in H.
rewrite e in t0.
eapply lemmaA4_1. apply t0.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
trivial.
intro.
cut (varset_In 0 (fv_in_type (map_paramSubst_in_type U xvs))).
intro. 
rewrite fv_and_paramSubst_in_type in H4.
apply n.
apply fv_and_var_in_type2. trivial.
apply fv_and_var_in_type1. trivial.
rewrite lemmaA36_34_2.
eapply lemmaA39_param_2. rewrite <- H6. apply e. trivial.
trivial. trivial.
(* T-Nil (param/term) *)
intros. unfold lemmaA39_param1_prop2. intros.
rewrite polySubst_in_paramTermNil.
apply ParamTermListTyping_nil.
(* T-Cons (param/term) *)
intros. unfold lemmaA39_param1_prop2. intros.
rewrite polySubst_in_paramTermCons.
eapply ParamTermListTyping_cons. apply e.
rewrite <- lemmaA9_2.
eapply H. apply H1. trivial. trivial.
intros.
unfold param_in_type.
apply lemmaA3_3_corollary with
  (C := (methodOwner (paramOwner x)))
  (x := x0).
eapply param_types_are_well_formed with
  (m := (paramOwner x))
.
apply e. trivial. trivial.
eapply H0. apply H1. trivial. trivial.
(* T-Nil (field/term) *)
intros. unfold lemmaA39_param1_prop3. intros.
rewrite polySubst_in_fieldTermNil.
apply FieldTermListTyping_nil.
(* T-Cons (field/term) *)
intros. unfold lemmaA39_param1_prop3. intros.
rewrite polySubst_in_fieldTermCons.
eapply FieldTermListTyping_cons with
  (U := (map_paramSubst_in_type U xvs))
.
eapply H. apply H1. trivial. trivial.
rewrite fp_and_paramSubst_in_type. trivial.
unfold type_is_closed in t1. tauto.
apply e.
rewrite fp_and_paramSubst_in_type. trivial.
unfold type_is_closed in t1. tauto.
eapply H0. apply H1. trivial. trivial.
Qed.

End SetProgram.
