(* Author: Vincent Cremet *)
(* Date: August 2006 *) 
(* File: Proofs_SubstitutionLemmas2.v *)
(* Description: Substitution lemmas (Appendix 7) *)

(** * Proofs about Substitution for a Local Variable *)

Require Import Calculus.
Require Import Arith.
Require Import MatchCompNat.
Require Import Util.
Require Import List.
Require Import ListSet.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_SubstitutionLemmas.v *)
Require Proofs_SubstitutionLemmas2.
Module My_Proofs_SubstitutionLemmas2 := Proofs_SubstitutionLemmas2.SetProgram(My_Program).

Import My_Proofs_SubstitutionLemmas2.
Import My_Proofs_SubstitutionLemmas.
Import My_Proofs_DeBruijnSIndices.
Import My_Proofs_Progress.
Import My_Proofs_AdmissibilityOfTransitivity.
Import My_Proofs_Subclassing.
Import My_Proofs_TypeOrdering.
Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(****************************************************************)

(*****************************************************************)
(* Substitution lemmas -- Var *)
(*****************************************************************)

(** Substitution of the value [v] for the variable [k] in a list of types [Ts] **)
Fixpoint map_varSubst_in_typeList (Ts: TypeList) (k: nat) (v: Value) {struct Ts}: TypeList :=
  match Ts with
    | nil      => nil
    | T :: Ts' => (varSubst_in_type T k v) :: (map_varSubst_in_typeList Ts' (S k) v) 
  end
.

(*****************************************************************)

Lemma zero_fv_and_varSubst_0:
  forall (p: Path) (k: nat) (v: Value),
    ~(varset_In 0 (fv_in_path p)) ->
    ~(varset_In 0 (fv_in_path (varSubst_in_path p (S k) v)))
.

Proof.
intros.
induction p.
(* This *)
simpl. tauto.
(* Param *)
simpl. tauto.
(* Var *)
simpl. rename v0 into i.
destruct (lt_eq_lt_dec i (S k)).
destruct s.
(* i < k + 1*)
rewrite match_comp_nat3_lt.
simpl.
rewrite match_comp_nat2_ge.
simpl. intro. destruct H0.
assert (i <= 0). omega. assert (i = 0). auto with arith.
apply H. rewrite H2. simpl.
rewrite match_comp_nat2_ge. simpl. tauto.
omega. tauto. omega. trivial.
(* i = k + 1 *)
rewrite match_comp_nat3_eq.
simpl. tauto. trivial.
(* i > k + 1 *)
rewrite match_comp_nat3_gt.
simpl. rewrite match_comp_nat2_ge.
simpl. intro. destruct H0.
assert ((pred i) - 0 > 0). omega.
rewrite H0 in H1.
cut (~(0 > 0)). intro. tauto.
auto with arith. tauto. auto with arith. omega.
(* Value *)
simpl. tauto.
Qed.

Lemma zero_fv_and_varSubst_2:
  forall (T: type) (k: nat) (v: Value),
    ~(varset_In 0 (fv_in_type T)) ->
    ~(varset_In 0 (fv_in_type (varSubst_in_type T (S k) v)))
.

Proof.
intros.
induction T.
(* Virtual type *)
simpl. simpl in H.
fold (fv_in_path (varSubst_in_path p (S k) v)).
apply zero_fv_and_varSubst_0. trivial.
(* Class type *)
simpl. tauto.
Qed.

Lemma lemmaA39_var_aux_1:
  forall (Ts: TypeList) (k: nat) (v: Value),
    (length (map_varSubst_in_typeList Ts k v)) = (length Ts)
.

Proof.
intros. generalize k. clear k.
induction Ts.
simpl. trivial.
simpl. intros. rewrite IHTs. trivial.
Qed.

Lemma lemmaA39_var_aux_2:
  forall (Ts: TypeList) (n: nat) (T: type) (v: Value),
    (TypeList_at Ts n T) ->
    (TypeList_at (map_varSubst_in_typeList Ts 0 v) n (varSubst_in_type T n v))
.

Proof.
cut (
  forall (Ts: TypeList) (n: nat) (T: type),
    (TypeList_at Ts n T) ->
    forall (k: nat) (v: Value),
      (TypeList_at (map_varSubst_in_typeList Ts k v) n (varSubst_in_type T (k + n) v))
).
intros.
pattern n at 2.
rewrite <- plus_0_l.
eapply H. trivial.

intros Ts n T H.
induction H.
simpl.
intros. rewrite plus_0_r.
apply TypeList_at_head.
simpl. intros.
apply TypeList_at_tail.
rewrite <- plus_Snm_nSm.
apply IHTypeList_at.
Qed.

Lemma lemmaA39_var_aux_3:
  forall (T: type),
  (forall (k : VarSym), (var_in_type k T) -> (k < 0)) ->
  (fv_in_type T = empty_varset)
.

Proof.
intros.
unfold empty_varset.
apply empty_is_empty_2 with (A := VarSym). apply VarSym_eqdec.
intros. intro.
cut (x < 0). fold (~ x < 0). auto with arith.
apply H.
apply fv_and_var_in_type2. trivial.
Qed.

Lemma lemmaA39_var_aux_4:
  forall (T: type),
    (forall (n: nat), ~(varset_In n (fv_in_type T))) ->
    (fv_in_type T = empty_varset)
.

Proof.
intros.
unfold empty_varset.
apply empty_is_empty_2 with (A := VarSym). apply VarSym_eqdec.
trivial.
Qed.

Lemma lemmaA39_var_aux_5:
  forall (xts: ParamTermList) (k: nat) (v: Value),
    (params_of_paramTermList (varSubst_in_paramTermList xts k v)) =
    (params_of_paramTermList xts)
.

Proof.
intros.
induction xts.
simpl. trivial.
simpl. rewrite IHxts. trivial.
Qed.

Lemma lemmaA39_var_aux_6:
  forall (fts: FieldTermList) (k: nat) (v: Value),
    (fields_of_fieldTermList (varSubst_in_fieldTermList fts k v)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros.
induction fts.
simpl. trivial.
simpl. rewrite IHfts. trivial.
Qed.

Lemma lemmaA39_var_aux_7:
  forall (Ts Us: TypeList) (v: Value),
    (map_varSubst_in_typeList (Ts ++ Us) 0 v) =
    (map_varSubst_in_typeList Ts 0 v) ++ (map_varSubst_in_typeList Us (length Ts) v)
.

Proof.
cut (
  forall (v: Value) (Ts Us: TypeList) (n: nat),
    (map_varSubst_in_typeList (Ts ++ Us) n v) =
    (map_varSubst_in_typeList Ts n v) ++ (map_varSubst_in_typeList Us (n + (length Ts)) v)
).
intros.
fold (0 + (length Ts)).
eapply H.
intro v.
induction Ts.
simpl. intros. rewrite plus_comm. simpl. trivial.
simpl. intros.
apply cons_destruct with (A := type). trivial.
rewrite <- plus_Snm_nSm.
apply IHTs.
Qed.

Lemma lemmaA39_var_aux_8:
  forall (T: type),
  (fv_in_type T = empty_varset) ->
  (forall (k : VarSym), (var_in_type k T) -> (k < 0))
.

Proof.
intros.
cut (varset_In k (fv_in_type T)).
intros. rewrite H in H1. simpl in H1. destruct H1.
apply fv_and_var_in_type1. trivial.
Qed.

(*****************************************************************)

Definition lemmaA39_var_prop1 (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)): Prop :=
  forall (U: type) (Us: TypeList), 
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Gen_PathTyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_path p N v) (varSubst_in_type T N v)
      )
.

Definition lemmaA39_var_prop2 (G: Env) (S1 S2: type) (H: (Gen_Subtyping G S1 S2)): Prop :=
  forall (U: type) (Us: TypeList),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Gen_Subtyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_type S1 N v) (varSubst_in_type S2 N v)
      )
.

Definition lemmaA39_var_prop3 (G: Env) (C: ClassSym) (fvs: FieldValueList)
  (H: (Gen_FieldValueListTyping G C fvs)): Prop :=
  forall (U: type) (Us: TypeList),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Gen_FieldValueListTyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        C fvs
      )
.

(*****************************************************************)

Lemma lemmaA39_var_main:
  forall (G: Env) (p: Path) (T: type) (H: (Gen_PathTyping G p T)),
    (lemmaA39_var_prop1 G p T H)
.

Proof.
apply (Gen_PathTyping_prop
  lemmaA39_var_prop1
  lemmaA39_var_prop2
  lemmaA39_var_prop3
).

(* P-This *)
intros. unfold lemmaA39_var_prop1. intros.
rewrite e in H0. discriminate H0.

(* P-Param *)
intros. unfold lemmaA39_var_prop1. intros.
simpl.
rewrite e in H0.
injection H0; intros; clear H0.
rewrite H5 in i. simpl in i.
tauto.

(* P-Var *)
intros. unfold lemmaA39_var_prop1. intros.
rename S into S'.
simpl.
rewrite e in H0. injection H0; intros; clear H0.
rewrite H5 in e. clear H5 xTs.
rewrite H6 in e. clear H6 O.
destruct a.
rewrite H4 in H0. simpl in H0.
rewrite H4 in e.
(* *)
assert (Gen_Subtyping
  (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
  (varSubst_in_type (lift_in_type T (n + 1)) N v)
  (varSubst_in_type S' N v)
).
eapply H. apply e. trivial. trivial. trivial.
(* Case analysis on n < N *)
destruct (lt_eq_lt_dec n N). destruct s.
(* n < N *)
rewrite match_comp_nat3_lt.
eapply Gen_P_Var with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_varSubst_in_typeList Us 0 v))
  (T := (varSubst_in_type T (N - n - 1) v))
. trivial.
split.
rewrite lemmaA39_var_aux_1. omega.
rewrite lemmaA39_var_aux_1.
(* Us(|Us| - n - 1) = T *)
assert (TypeList_at Us ((length Us) - n - 1) T).
assert ((pred ((length Ts) - n - 1)) = (length Us - n - 1)).
rewrite H4.
simpl (length (U :: Us)).
rewrite <- minus_Sn_m.
rewrite <- pred_of_minus. simpl.
rewrite pred_of_minus. trivial. omega.
rewrite <- H7. clear H7.
apply typeList_at_cons with (T := U).
rewrite H4.
simpl (length (U :: Us)).
omega.
rewrite <- H4. trivial.
(* *)
rewrite <- H2.
apply lemmaA39_var_aux_2.
rewrite H2. trivial.
(* *)
rewrite lemmaA35_2.
(* N - n - 1 + (n + 1) = N *)
assert ((N - n - 1 + (n + 1)) = N).
assert ((n + 1) <= N). omega.
rewrite myarith_2.
rewrite myarith_1. trivial.
trivial. trivial.
rewrite H7. clear H7.
eapply H. apply e. trivial. trivial. trivial.
trivial.
(* n = N *)
rewrite match_comp_nat3_eq.
(* apply subsumption *)
apply lemmaA1_for_paths with (T := U).
(* apply weakening *)
apply lemmaA38_0_corollary. trivial.
(* *)
cut (U= (varSubst_in_type (lift_in_type T (n + 1)) N v)). intros.
rewrite H7.
eapply H6.
(* U = T *)
assert (U = T).
rewrite H4 in H5.
simpl (length (U :: Us)) in H5.
rewrite <- H2 in H5.
rewrite <- e0 in H5.
assert ((S n - n - 1) = 0).
rewrite myarith_2. rewrite plus_comm. simpl.
rewrite <- minus_n_n. trivial.
omega.
rewrite H7 in H5. clear H7.
symmetry.
eapply typeList_at_head. apply H5.
rewrite <- H7.
(* *)
rewrite e in H1.
inversion_clear H1. clear H8.
unfold local_types_well_formed in H9.
simpl in H9. destruct H9. clear H8.
rewrite lemmaA30_2.
rewrite lemmaA37_2. trivial.
unfold type_is_var_closed.
apply lemmaA39_var_aux_3. trivial.
unfold type_is_var_closed.
apply lemmaA39_var_aux_3. trivial.
trivial.
rewrite <- H2 in H0.
assert (n < n). omega.
assert (~(n < n)). auto with arith.
destruct (H8 H7).

(* P-Value *)
intros. unfold lemmaA39_var_prop1. intros.
simpl.
eapply Gen_P_Value. trivial.
eapply H. apply H1. trivial. apply H3. trivial.
trivial.
fold (varSubst_in_type (type_class C) N v). 
eapply H0. apply H1. trivial. trivial. trivial.

(* S-Class *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
apply Gen_S_Class.

(* S-Extends *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Extends. apply e.
fold (varSubst_in_type (type_class C') N v). 
eapply H. apply H0. trivial. trivial. trivial.

(* S-Up *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Up with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply H. apply H1. trivial. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA10_2.
eapply H0. apply H1. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeUpperBounds_are_well_formed. apply e. apply e0.

(* S-Down *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Down with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply H. apply H1. trivial. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA10_2.
eapply H0. apply H1. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeLowerBounds_are_well_formed. apply e. apply e0.

(* S-Alias-Left *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Alias_Left with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply H. apply H1. trivial. trivial. trivial. apply e.
rewrite <- lemmaA10_2.
eapply H0. apply H1. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeValues_are_well_formed. apply e.

(* S-Alias-Right *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Alias_Right with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply H. apply H1. trivial. trivial. trivial. apply e.
rewrite <- lemmaA10_2.
eapply H0. apply H1. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeValues_are_well_formed. apply e.

(* S-Virtual *)
intros. unfold lemmaA39_var_prop2. intros.
simpl.
eapply Gen_S_Virtual with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply H. apply H0. trivial. trivial. trivial. apply e.

(* S-Trans *)
intros. unfold lemmaA39_var_prop2. intros.
eapply Gen_S_Trans with (S :=  (varSubst_in_type S N v)).
eapply H. apply H1. trivial. trivial. trivial.
eapply H0. apply H1. trivial. trivial. trivial.
tauto.

(* T-Nil (field/value) *)
intros. unfold lemmaA39_var_prop3. intros.
apply Gen_FieldValueListTyping_nil.

(* T-Cons (field/value) *)
intros. unfold lemmaA39_var_prop3. intros.
eapply Gen_FieldValueListTyping_cons with
  (U := (varSubst_in_type U N v0))
.
fold (varSubst_in_path (Path_value v) N v0).
eapply H. apply H1. trivial. trivial. trivial.
rewrite lemmaA37_2. trivial.
unfold type_is_closed in t.
unfold type_is_var_closed. tauto.
apply e.
rewrite lemmaA37_2. trivial.
unfold type_is_closed in t.
unfold type_is_var_closed. tauto.
eapply H0. apply H1. trivial. apply H3. trivial.
Qed.

(*****************************************************************)

(** Subtitution lemmas -- Var -- path **)
Lemma lemmaA39_var_0:
  forall (G: Env) (U: type) (Us: TypeList) (p: Path) (T: type),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    (Gen_PathTyping G p T) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Gen_PathTyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_path p N v) (varSubst_in_type T N v)
      )
.

Proof.
pose (H := lemmaA39_var_main).
intros.
unfold lemmaA39_var_prop1 in H.
eapply H.
apply H2. apply H0. trivial. trivial. trivial.
Qed.

(** Subtitution lemmas -- Var -- type **)
Lemma lemmaA39_var_2:
  forall (G: Env) (U: type) (Us: TypeList) (S1 S2: type),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    (Gen_Subtyping G S1 S2) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Gen_Subtyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_type S1 N v) (varSubst_in_type S2 N v)
      )
.

Proof.
cut (
  forall (G: Env) (S1 S2: type) (H: (Gen_Subtyping G S1 S2)),
    (lemmaA39_var_prop2 G S1 S2 H)
).
intros.
unfold lemmaA39_var_prop2 in H.
eapply H. apply H2. apply H0. trivial. trivial. trivial.

intros.
unfold lemmaA39_var_prop2.
induction H.

(* S-Class *)
intros.
simpl.
apply Gen_S_Class.

(* S-Extends *)
intros.
simpl.
eapply Gen_S_Extends. apply H.
fold (varSubst_in_type (type_class C') N v).
eapply IHGen_Subtyping. apply H1. trivial. trivial. trivial.

(* S-Up *)
intros.
simpl.
eapply Gen_S_Up with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H3. trivial. trivial. trivial. trivial. trivial. apply H1.
rewrite <- lemmaA10_2.
eapply IHGen_Subtyping. apply H3. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeUpperBounds_are_well_formed. apply H0. trivial.

(* S-Down *)
intros.
simpl.
eapply Gen_S_Down with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H3. trivial. trivial. trivial. trivial. trivial. apply H1.
rewrite <- lemmaA10_2.
eapply IHGen_Subtyping. apply H3. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeLowerBounds_are_well_formed. apply H0. trivial.

(* S-Alias-Left *)
intros.
simpl.
eapply Gen_S_Alias_Left with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H2. trivial. trivial. trivial. trivial. apply H0.
rewrite <- lemmaA10_2.
eapply IHGen_Subtyping. apply H2. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeValues_are_well_formed. apply H0.

(* S-Alias-Right *)
intros.
simpl.
eapply Gen_S_Alias_Right with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H2. trivial. trivial. trivial. trivial. apply H0.
rewrite <- lemmaA10_2.
eapply IHGen_Subtyping. apply H2. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply typeValues_are_well_formed. apply H0.

(* S-Virtual *)
intros.
simpl.
eapply Gen_S_Virtual with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H1. trivial. trivial. trivial. trivial. trivial.

(* S-Trans *)
intros.
eapply Gen_S_Trans with (S :=  (varSubst_in_type S N v)).
eapply IHGen_Subtyping1. apply H2. trivial. trivial. trivial.
eapply IHGen_Subtyping2. apply H2. trivial. trivial. trivial.
tauto.
Qed.

(*****************************************************************)

Definition lemmaA39_var1_prop1 (G: Env) (t: Term) (T: type) (H: (Typing G t T)): Prop :=
  forall (U: type) (Us: TypeList),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Typing (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_term t N v) (varSubst_in_type T N v)
      )
.

Definition lemmaA39_var1_prop2 (G: Env) (p: Path) (m: MethodSym) (xts: ParamTermList) (H: (ParamTermListTyping G p m xts)): Prop :=
  forall (U: type) (Us: TypeList),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (ParamTermListTyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_path p N v)
        m
        (varSubst_in_paramTermList xts N v)
      )
.

Definition lemmaA39_var1_prop3 (G: Env) (C: ClassSym) (fts: FieldTermList) (H: (FieldTermListTyping G C fts)): Prop :=
  forall (U: type) (Us: TypeList),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (FieldTermListTyping (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        C (varSubst_in_fieldTermList fts N v)
      )
.

(** Subtitution lemmas -- Var -- term **)
Lemma lemmaA39_var_1:
  forall (G: Env) (U: type) (Us: TypeList) (t: Term) (T: type),
    (G = (Env_mk Owner_root nil (U :: Us))) ->
    (context_is_well_formed G) ->
    (Typing G t T) ->
    forall (v: Value) (N: nat),
      N = (length Us) ->
      (Gen_PathTyping rootEnv (Path_value v) U) ->
      (Typing (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
        (varSubst_in_term t N v) (varSubst_in_type T N v)
      )
.

Proof.
cut (forall (G: Env) (t: Term) (T: type) (H: (Typing G t T)),
  (lemmaA39_var1_prop1 G t T H)
).
intros.
unfold lemmaA39_var1_prop1 in H.
eapply H. apply H2. apply H0. trivial. trivial. trivial.
apply (Typing_prop
  lemmaA39_var1_prop1
  lemmaA39_var1_prop2
  lemmaA39_var1_prop3
).

(* T-This *)
intros. unfold lemmaA39_var1_prop1. intros.
rewrite e in H. discriminate H.

(* T-Param *)
intros. unfold lemmaA39_var1_prop1. intros.
simpl.
rewrite e in H.
injection H; intros; clear H.
rewrite H4 in i. simpl in i.
tauto.

(* T-Var *)
intros. unfold lemmaA39_var1_prop1. intros.
rename S into S'.
simpl.
rewrite e in H. injection H; intros; clear H.
rewrite H4 in e. clear H4 xTs.
rewrite H5 in e. clear H5 O.
destruct a.
rewrite H3 in H. simpl in H.
rewrite H3 in e.
(* *)
assert (Gen_Subtyping
  (Env_mk Owner_root nil (map_varSubst_in_typeList Us 0 v))
  (varSubst_in_type (lift_in_type T (n + 1)) N v)
  (varSubst_in_type S' N v)
).
eapply lemmaA39_var_2. apply e. trivial. trivial. trivial. trivial.
(* Case analysis on n < N *)
destruct (lt_eq_lt_dec n N). destruct s.
(* n < N *)
rewrite match_comp_nat3_lt.
eapply T_Var with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_varSubst_in_typeList Us 0 v))
  (T := (varSubst_in_type T (N - n - 1) v))
. trivial.
split.
rewrite lemmaA39_var_aux_1. omega.
rewrite lemmaA39_var_aux_1.
(* Us(|Us| - n - 1) = T *)
assert (TypeList_at Us ((length Us) - n - 1) T).
assert ((pred ((length Ts) - n - 1)) = (length Us - n - 1)).
rewrite H3.
simpl (length (U :: Us)).
rewrite <- minus_Sn_m.
rewrite <- pred_of_minus. simpl.
rewrite pred_of_minus. trivial. omega.
rewrite <- H6. clear H6.
apply typeList_at_cons with (T := U).
rewrite H3.
simpl (length (U :: Us)).
omega.
rewrite <- H3. trivial.
(* *)
rewrite <- H1.
apply lemmaA39_var_aux_2.
rewrite H1. trivial.
(* *)
rewrite lemmaA35_2.
(* N - n - 1 + (n + 1) = N *)
assert ((N - n - 1 + (n + 1)) = N).
assert ((n + 1) <= N). omega.
rewrite myarith_2.
rewrite myarith_1. trivial.
trivial. trivial.
rewrite H6. clear H6.
eapply lemmaA39_var_2. apply e. trivial. trivial. trivial. trivial. trivial.
(* n = N *)
rewrite match_comp_nat3_eq.
(* apply subsumption *)
apply lemmaA40_aux_4.
apply lemmaA1_for_paths with (T := U).
(* apply weakening *)
apply lemmaA38_0_corollary. trivial.
(* *)
cut (U= (varSubst_in_type (lift_in_type T (n + 1)) N v)). intros.
rewrite H6.
eapply H5.
(* U = T *)
assert (U = T).
rewrite H3 in H4.
simpl (length (U :: Us)) in H4.
rewrite <- H1 in H4.
rewrite <- e0 in H4.
assert ((S n - n - 1) = 0).
rewrite myarith_2. rewrite plus_comm. simpl.
rewrite <- minus_n_n. trivial.
omega.
rewrite H6 in H4. clear H6.
symmetry.
eapply typeList_at_head. apply H4.
rewrite <- H6.
(* *)
rewrite e in H0.
inversion_clear H0. clear H7.
unfold local_types_well_formed in H8.
simpl in H8. destruct H8. clear H7.
rewrite lemmaA30_2.
rewrite lemmaA37_2. trivial.
unfold type_is_var_closed.
apply lemmaA39_var_aux_3. trivial.
unfold type_is_var_closed.
apply lemmaA39_var_aux_3. trivial.
trivial.
(* n > N *)
rewrite <- H1 in H.
assert (n < n). omega.
assert (~(n < n)). auto with arith.
destruct (H7 H6).

(* T-Select *)
intros. unfold lemmaA39_var1_prop1. intros.
simpl.
eapply T_Select with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H. trivial. trivial. trivial. trivial. trivial. apply e0.
rewrite <- lemmaA10_2.
eapply lemmaA39_var_2. apply H. trivial. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply fieldTypes_are_well_formed. apply e. trivial.

(* T-Call *)
intros. unfold lemmaA39_var1_prop1. intros.
simpl.
eapply T_Call with (C := C).
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_0. apply H0. trivial. trivial. trivial. trivial. trivial. apply e0.
rewrite lemmaA39_var_aux_5. trivial.
eapply H. apply H0. trivial. trivial. trivial.
(* *)
rewrite <- lemmaA10_2.
eapply lemmaA39_var_2. apply H0. trivial. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply methodTypes_are_well_formed. apply e. trivial.

(* T-New *)
intros. unfold lemmaA39_var1_prop1. intros.
simpl.
eapply T_New.
rewrite lemmaA39_var_aux_6. trivial.
eapply H. apply H0. trivial. trivial. trivial.
rewrite lemmaA39_var_aux_6. trivial.
fold (varSubst_in_type (type_class C) N v).
eapply lemmaA39_var_2. apply H0. trivial. trivial. trivial. trivial. 

(* T-Let *)
intros. unfold lemmaA39_var1_prop1. intros.
simpl. rename S into S'.
rewrite H1 in e. injection e; intros; clear e.
eapply T_Let with
  (O := Owner_root)
  (xTs := (nil (A := (ParamSym * type))))
  (Ts := (map_varSubst_in_typeList Us 0 v))
  (U :=  (varSubst_in_type U (S N) v))
. trivial.
eapply H. apply H1. trivial. trivial. trivial.
assert (
  ((map_varSubst_in_typeList Us 0 v) ++ ((varSubst_in_type T N v) :: nil)) =
  (map_varSubst_in_typeList (Us ++ (T :: nil)) 0 v)
).
rewrite lemmaA39_var_aux_7.
simpl. rewrite <- H3. trivial.
rewrite H8. clear H8.
unfold lemmaA39_var1_prop1 in H0.
eapply H0 with (U0 := U0).
rewrite <- H7. rewrite <- H6. rewrite <- H5.
simpl. trivial.
(* WF_ctx(root; nil; (Ts ++ T :: nil)) *)
rewrite <- H7. rewrite <- H6. rewrite <- H5. simpl.
rewrite H1 in H2. inversion_clear H2.
clear H8.
apply context_is_well_formed_intro.
simpl. trivial.
unfold local_types_well_formed. simpl.
split.
apply lemmaA39_var_aux_8.
apply lemmaA39_var_aux_4.
intro.
eapply lemmaA4_0_corollary. apply H4.
unfold local_types_well_formed in H9.
simpl in H9. destruct H9. clear H2.
eapply lemmaA39_param_aux_4. trivial.
(* *)
clear H0.
clear H.
rewrite H1 in t0.
rewrite plus_comm.
intros.
assert ((length (U0 :: Us)) = 1 + (length Us)). simpl. trivial.
rewrite <- H0.
eapply lemmaA4_1. apply t0.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial.
(* *)
rewrite length_of_append. simpl. rewrite plus_comm. simpl.
rewrite H3. trivial. trivial.
(* *)
intro.
apply (zero_fv_and_varSubst_2 U N v).
intro. apply n.
apply fv_and_var_in_type2. trivial.
apply fv_and_var_in_type1. trivial.
(* *)
rewrite lemmaA36_35_2.
eapply lemmaA39_var_2. apply H1. trivial. trivial. trivial. trivial.
intro. apply n. apply fv_and_var_in_type2. trivial.

(* T-Nil (param/term) *)
intros. unfold lemmaA39_var1_prop2. intros.
simpl.
apply ParamTermListTyping_nil.

(* T-Cons (param/term) *)
intros. unfold lemmaA39_var1_prop2. intros.
simpl.
eapply ParamTermListTyping_cons. apply e.
rewrite <- lemmaA10_2.
eapply H. apply H1. trivial. trivial. trivial.
apply lemmaA39_var_aux_4.
intros.
eapply lemmaA4_3_corollary.
eapply param_types_are_well_formed with
  (C := (methodOwner (paramOwner x)))
  (m := (paramOwner x))
.
apply e. trivial. trivial.
eapply H0. apply H1. trivial. trivial. trivial.

(* T-Nil (field/term) *)
intros. unfold lemmaA39_var1_prop3. intros.
simpl.
apply FieldTermListTyping_nil.

(* T-Cons (field/term) *)
intros. unfold lemmaA39_var1_prop3. intros.
simpl.
eapply FieldTermListTyping_cons with
  (U := (varSubst_in_type U N v))
.
eapply H. apply H1. trivial. trivial. trivial.
rewrite lemmaA37_2. trivial.
unfold type_is_closed in t1.
unfold type_is_var_closed. tauto.
apply e.
rewrite lemmaA37_2. trivial.
unfold type_is_closed in t1.
unfold type_is_var_closed. tauto.
eapply H0. apply H1. trivial. trivial. trivial.
Qed.

(** Subtitution lemmas -- Var -- term | Corollary **)
Lemma lemmaA39_var_1_corollary:
  forall (v: Value) (U: type) (t: Term) (T: type),
    (Typing (Env_mk Owner_root nil (U :: nil)) t T) ->
    (Gen_PathTyping rootEnv (Path_value v) U) ->
    (Typing rootEnv (varSubst_in_term t 0 v) (varSubst_in_type T 0 v))
.

Proof.
intros.
unfold rootEnv.
fold (map_varSubst_in_typeList nil 0 v).
eapply lemmaA39_var_1 with
  (G := (Env_mk Owner_root nil (U :: nil)))
  (U := U)
. trivial.
apply context_is_well_formed_intro.
simpl. trivial.
unfold local_types_well_formed. simpl.
split.
intros.
assert (varset_In k (fv_in_type U)).
apply fv_and_var_in_type1. trivial. clear H1.
eapply lemmaA4_0 with
  (Us := (nil (A := type)))
.
unfold rootEnv in H0. apply H0.
unfold varset_In. unfold varset_union.
apply set_union_intro2. trivial. trivial.
trivial. simpl. trivial.
trivial.
Qed.

End SetProgram.
