(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Proofs_TypeOrdering.v *)
(* Description: Properties of the relation over types (Appendix 2) *)

(** * Proofs about Type Ordering *)

Require Import Calculus.
Require Import WellFoundation.

Module SetProgram(My_Program: Program).

Import My_Program.

(* Import definitions of Proofs_Miscellaneous.v *)
Require Proofs_Miscellaneous.
Module My_Proofs_Miscellaneous := Proofs_Miscellaneous.SetProgram(My_Program).

Import My_Proofs_Miscellaneous.
Import My_WellFormedness.
Import My_Semantics.
Import My_Typing.
Import My_TypeOrdering.
Import My_DeBruijn.
Import My_Substitutions.
Import My_Subclassing.

(** Interpretation and substitution of a value **)
Lemma lemmaA12:
  forall (T: type) (v: Value),
  (symbol_of_type (thisSubst_in_type T (Path_value v))) = (symbol_of_type T)
.

Proof.
intros.
induction T.
simpl. trivial.
simpl. trivial.
Qed.

(** Interpretation and substitution of a path **)
Lemma lemmaA12':
  forall (T: type) (p: Path),
  (symbol_of_type (thisSubst_in_type T p)) = (symbol_of_type T)
.

Proof.
intros.
induction T.
simpl. trivial.
simpl. trivial.
Qed.

(** typeRel is a function extension of symRel **)
Lemma typeRel_is_fun_extension:
    typeRel = (fun_extension type Sym symbol_of_type symRel)
.

Proof.
unfold fun_extension.
unfold typeRel. trivial.
Qed.

(** Type ordering and substitution of a value **)
Lemma lemmaA14:
  forall (T U: type) (v: Value),
    (typeRel T U) ->
    (typeRel (thisSubst_in_type T (Path_value v)) (thisSubst_in_type U (Path_value v))).

Proof.
intros.
red in H.
red.
rewrite lemmaA12.
rewrite lemmaA12.
trivial.
Qed.

(** Type ordering is well-founded **)
Lemma lemmaA17: (well_founded typeRel).

Proof.
rewrite typeRel_is_fun_extension.
apply well_foundation4.
exact symRel_is_well_founded_proof.
Qed.

(** Facts about type ordering - 1 **)
Lemma lemmaA15_1: forall (T: type) (C: ClassSym),
  (typeRel T (type_class C)) ->
  exists C': ClassSym, T = (type_class C')
.

Proof.
intros.
induction T.
rename t into L.
unfold typeRel in H.
simpl in H.
assert (symRel (inl _ C) (inr _ L)).
apply (classes_smaller_than_types_proof).
assert False.
eapply (well_foundation2 Sym symRel).
exact symRel_is_well_founded_proof.
apply H. trivial.
induction H1.
exists c. trivial.
Qed.

(** Facts about type ordering - 2 **)
Lemma lemmaA15_2: forall (C: ClassSym) (p: Path) (L: TypeSym),
  (typeRel (type_class C) (type_virtual p L))
.

Proof.
intros.
unfold typeRel.
unfold symbol_of_type.
apply (classes_smaller_than_types_proof).
Qed.

(** Declarations and type ordering **)
Lemma lemmaA13_1:
  forall (L: TypeSym) (T: type) (p: Path),
    (typeUpperBound L) = T ->
    (typeRel (thisSubst_in_type T p) (type_virtual p L))
.

Proof.
(* Upper-bound *)
intros.
pose (C := (typeOwner L)).
assert (WF_class C).
apply (classes_are_well_formed_proof C).
assert (WF_typeDef C L).
inversion H0.
unfold typeDefs_are_well_formed in H3.
apply H3. trivial.
unfold typeRel.
simpl.
inversion H1.
clear H5. clear H3. clear H6. clear H7.
rewrite H in H2.
rewrite <- H2 in H4.
rewrite lemmaA12'. trivial.
Qed.

(** Declarations and type ordering **)
Lemma lemmaA13_1':
  forall (L: TypeSym) (U: type) (p: Path),
    (typeLowerBound L) = (Some U) ->
    (typeRel (thisSubst_in_type U p) (type_virtual p L))
.

Proof.
(* Lower-bound *)
intros.
pose (C := (typeOwner L)).
assert (WF_class C).
apply (classes_are_well_formed_proof C).
assert (WF_typeDef C L).
inversion H0.
unfold typeDefs_are_well_formed in H3.
apply H3. trivial.
unfold typeRel.
simpl.
inversion H1.
clear H4. clear H3. clear H6. clear H7. clear H2.
induction (H5 U H). clear H2. clear H5.
rewrite lemmaA12'. trivial.
Qed.

(** Declarations and type ordering **)
Lemma lemmaA13_2:
  forall (L: TypeSym) (C: ClassSym) (T: type) (p: Path),
    (typeValue L C) = (Some T) ->
    (typeRel (thisSubst_in_type T p) (type_virtual p L))
.

Proof.
intros.
assert (WF_class C).
apply (classes_are_well_formed_proof C).
assert (WF_typeVal C L T).
inversion H0.
unfold typeVals_are_well_formed in H4.
apply H4. trivial.
unfold typeRel.
simpl.
inversion H1.
clear H2 H4 H5 H6 H7 H8 H9 H10 H11.
rewrite lemmaA12'. trivial.
Qed.

(** Declarations and type ordering **)
Lemma lemmaA13_3:
  forall (C C': ClassSym),
    (classSuper C) = (Some C') ->
    (typeRel (type_class C') (type_class C))
.

Proof.
intros.
assert (WF_class C).
apply (classes_are_well_formed_proof C).
unfold typeRel.
simpl.
inversion H0.
eapply H1. trivial.
Qed.

(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)
(******************************************************************)


(* Multiset abstract data type *)

Require Import Multiset.
Require Import MultisetCore.
Require Import MultisetList.
Require Import RelExtras.
Require Import Setoid.
Require Import MultisetOrder.
Require Import MultisetTheory.

Module Type_Eqset: Eqset
  with Definition A := type
  with Definition eqA := fun (T U: type) => T = U
.

  Definition A : Set := type.
  Definition eqA : A -> A -> Prop := fun (T U: type) => T = U.
  Lemma sid_theoryA: Setoid_Theory A eqA.
  Proof.
    split.
    unfold eqA; trivial.
    unfold eqA. auto.
    unfold eqA. intros.
    rewrite H. trivial.
  Qed.

End Type_Eqset.

Module Type_Eqset_dec: Eqset_dec
  with Module ESM := Type_Eqset
  with Definition eqA_dec := Type_eqdec.

  Module ESM := Type_Eqset.
  Import ESM.

  Definition eqA_dec: forall (x y: A), { (eqA x y) } + { ~(eqA x y) } := Type_eqdec.

End Type_Eqset_dec.

Module Type_Multiset: FiniteMultisetCore
  with Module Sid := Type_Eqset_dec
:= FiniteMultisetList(Type_Eqset_dec).

Import Type_Multiset.

Module Type_MultisetOrder := MultisetOrder(Type_Multiset).

Import Type_MultisetOrder.
Import MSet.

(** Multiset extension of typeRel **)
Definition typeRelMul: Multiset -> Multiset -> Prop := MultisetLt (transp _ typeRel).

Lemma lemmaA18_aux_1: forall (A: Set) (R: A -> A -> Prop) (x: A),
  (Acc R x) ->
  (Acc (fun (x y: A) => (R x y)) x).

Proof.
intros.
induction H.
apply Acc_intro.
intros.
apply H0. trivial.
Qed.

Lemma lemmaA18_aux_2: forall (A: Set) (R: A -> A -> Prop),
  (well_founded R) ->
  (well_founded (fun (x y: A) => (R x y))).

Proof.
unfold well_founded. intros.
apply lemmaA18_aux_1. apply H.
Qed.

(** The multiset extension of type ordering is well-founded **)
Lemma lemmaA18:
  well_founded typeRelMul.

Proof.
unfold typeRelMul.
apply mord_wf.
unfold eqA.
unfold transp.
intros. rewrite <- H. rewrite <- H0. trivial.
unfold transp.
apply lemmaA18_aux_2.
apply lemmaA17.
Qed.

(**************************************************************)

(** Some useful multiset inequalities **)

Lemma typeRelMul1:
  forall (T T' S U: type),
    (typeRel T T') ->
    (typeRelMul
      ({{T}}  + {{S}} + {{U}})
      ({{T'}} + {{S}} + {{U}})
    )
.

Proof.
intros.
unfold typeRelMul.
unfold MultisetLt.
unfold transp. unfold MultisetGt.
apply t_step.
apply (MSetRed (transp _ typeRel)) with (X:= {{S}} + {{U}}) (a := T') (Y := {{T}}).
try_solve_meq.
try_solve_meq.
intros.
unfold transp.
assert (y = T).
apply singleton_member_eqA. trivial.
rewrite H1. trivial.
Qed.

Lemma typeRelMul3:
  forall (T S S' U: type),
    (typeRel S S') ->
    (typeRelMul
      ({{T}} + {{S}}  + {{U}})
      ({{T}} + {{S'}} + {{U}})
    )
.

Proof.
intros.
unfold typeRelMul.
unfold MultisetLt.
unfold transp. unfold MultisetGt.
apply t_step.
apply (MSetRed (transp _ typeRel)) with (X:= {{T}} + {{U}}) (a := S') (Y := {{S}}).
try_solve_meq.
try_solve_meq.
intros.
unfold transp.
assert (y = S).
apply singleton_member_eqA. trivial.
rewrite H1. trivial.
Qed.

Lemma typeRelMul2:
  forall (T S S1 S2 U: type),
    (typeRel S1 S) ->
    (typeRel S2 S) ->
    (typeRelMul
      ({{T}} + {{S1}} + {{S2}})
      ({{T}} + {{S}} + {{U}})
    )
.

Proof.
intros.
unfold typeRelMul.
unfold MultisetLt.
unfold transp.
unfold MultisetGt.
apply t_trans with (y := ({{T}} + {{S}})).

(* *)
apply t_step.
apply (MSetRed (transp _ typeRel)) with (X:= ({{T}} + {{S}})) (a := U) (Y := empty).
try_solve_meq.
try_solve_meq.
intros.
assert (empty <>mul empty).
apply (member_notempty H1).
assert (empty =mul= empty).
try_solve_meq.
induction (H2 H3).

(* *)
apply t_step.
apply (MSetRed (transp _ typeRel)) with (X:= ({{T}})) (a := S) (Y := {{S1}} + {{S2}}).
try_solve_meq.
try_solve_meq.
intros.
unfold transp.
assert (y in {{S1,S2}}).
unfold pair.
unfold insert. trivial.
induction (member_pair H2).
unfold eqA in H3.
rewrite H3. trivial.
unfold eqA in H3.
rewrite H3. trivial.
Qed.

Lemma typeRelMul4:
  forall (T S U U': type),
    (typeRel U U') ->
    (typeRelMul
      ({{T}} + {{S}} + {{U}})
      ({{T}} + {{S}} + {{U'}})
    )
.

Proof.
intros.
unfold typeRelMul.
unfold MultisetLt.
unfold transp. unfold MultisetGt.
apply t_step.
apply (MSetRed (transp _ typeRel)) with (X:= {{T}} + {{S}}) (a := U') (Y := {{U}}).
try_solve_meq.
try_solve_meq.
intros.
unfold transp.
assert (y = U).
apply singleton_member_eqA. trivial.
rewrite H1. trivial.
Qed.

End SetProgram.
