(* Author: Vincent Cremet *)
(* Date: July 2006 *) 
(* File: Util.v *)
(* Description: Definitions and theorems of general use *)

(** * Useful Properties of Basic Datatypes *)

Require Import Arith.
Require Import List.
Require Import ListSet.
Require Import Omega.

(*****************************************************************)

Lemma length_of_append:
  forall (A: Set) (xs ys: (list A)),
    (length (xs ++ ys)) = (length xs) + (length ys)
.

Proof.
intros A.
induction xs.
intros. simpl. trivial.
intros. simpl. rewrite IHxs. trivial.
Qed.

(*****************************************************************)
(* Elimination of list map *)

Lemma map_elim:
  forall (A B: Set) (f: A -> B) (xs: (list A)) (y: B),
    (In y (map f xs)) ->
    exists x: A, (In x xs) /\ y = (f x)
.

Proof.
intros.
generalize xs H. clear xs H.
induction xs.
simpl. tauto.
simpl. intros.
destruct H.
exists a.
split. left. trivial.
symmetry. trivial.
destruct (IHxs H). destruct H0.
exists x.
split. right. trivial. trivial.
Qed.

(*****************************************************************)
(* Introduction of list map *)

Lemma map_intro:
  forall (A B: Set) (f: A -> B) (x: A) (xs: (list A)),
    (In x xs) ->
    (In (f x) (map f xs))
.

Proof.
intros.
generalize xs H. clear xs H.
induction xs.
simpl. tauto.
simpl. intros.
destruct H.
left. rewrite H. trivial.
right. apply IHxs. trivial.
Qed.

(*****************************************************************)
(* Length of map *)

Lemma length_of_map:
  forall (A B: Set) (f: A -> B) (xs: (list A)),
    (length (map f xs)) = (length xs)
.

Proof.
intros.
induction xs.
simpl. trivial.
simpl. rewrite IHxs. trivial.
Qed.

(*****************************************************************)
(* Length of map *)

Lemma map_of_append:
  forall (A B: Set) (f: A -> B) (xs ys: (list A)),
    (map f (xs ++ ys)) = (map f xs) ++ (map f ys)
.

Proof.
intros A B f xs.
induction xs.
simpl. trivial.
simpl. intros.
rewrite IHxs. trivial.
Qed.

(*****************************************************************)

Lemma cons_destruct:
    forall (A: Set) (x y: A) (xs ys: (list A)),
      (x = y) -> (xs = ys) -> (x :: xs) = (y :: ys)
.

Proof.
intros.
rewrite H. rewrite H0. trivial.
Qed.

(*****************************************************************)
(* Arithmetic properties *)

Lemma plus_minus_assoc:
  forall (n m p: nat),
    (p <= m) ->
    n + m - p = n + (m - p)
.

Proof.
intros. generalize n p H. clear n p H.
induction m.
intros.
simpl.
destruct n.
simpl. trivial.
rewrite plus_0_r.
assert (p = 0). auto with arith. rewrite H0.
rewrite <- minus_n_O. trivial.
intros.
destruct p.
simpl.
rewrite <- minus_n_O. trivial.
simpl.
rewrite <- plus_Snm_nSm. simpl.
apply IHm. auto with arith.
Qed.

Lemma myarith_1:
  forall (n m: nat),
    (m <= n) ->
    (n - m + m) = n
.

Proof.
intros.
rewrite plus_comm.
rewrite <- plus_minus_assoc.
apply minus_plus.
trivial.
Qed.

Lemma myarith_2:
  forall (n m p: nat),
    ((m + p) <= n) ->
    (n - m - p) = (n - (m + p))
.

Proof.
intros. generalize n p H. clear n p H.
induction m.
intros.
simpl. rewrite <- minus_n_O. trivial.
simpl. intros.
destruct n.
simpl. trivial.
simpl. apply IHm.
omega.
Qed.

(*******************************************************************)
(* Properties of listsets *)

Lemma empty_is_empty_1:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (x: A),
      ~(set_In x (empty_set A))
.

Proof.
intros.
unfold set_In. simpl. tauto.
Qed.

Lemma empty_is_empty_2:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s: (set A)),
      (forall (x: A), ~(set_In x s)) ->
      (s = (empty_set A))
.

Proof.
intros A eqdec s.
destruct s.
trivial.
intros.
cut False. tauto.
unfold not in H.
apply H with (x := a).
simpl. tauto.
Qed.

Lemma empty_union_left:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s1 s2: (set A)),
    (set_union eqdec s1 s2) = (empty_set A) ->
    s1 = (empty_set A)
.

Proof.
intros.
apply empty_is_empty_2. trivial.
unfold not.
intros.
assert (set_In x (set_union eqdec s1 s2)).
apply set_union_intro1. trivial.
rewrite H in H1.
apply empty_is_empty_1 with (x := x). trivial. trivial.
Qed.

Lemma empty_union_right:
  forall (A: Set) (eqdec: forall (x y: A), { x = y } + { x <> y }),
    forall (s1 s2: (set A)),
    (set_union eqdec s1 s2) = (empty_set A) ->
    s2 = (empty_set A)
.

Proof.
intros.
apply empty_is_empty_2. trivial.
unfold not.
intros.
assert (set_In x (set_union eqdec s1 s2)).
apply set_union_intro2. trivial.
rewrite H in H1.
apply empty_is_empty_1 with (x := x). trivial. trivial.
Qed.

(*****************************************************************)

